#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <malloc.h>
#include <time.h>
#include <float.h>

#include <rbr_f2c.h>

#define MAXLINE 200

#define RBR_a3to1_acids3 "ALA ARG ASN ASP CYS GLN GLU GLY HIS ILE LEU LYS MET PHE PRO SER THR TRP TYR VAL ASX GLX UNK CYH HET"
#define RBR_a3to1_acids1 "ARNDCQEGHILKMFPSTWYVBZXch"

/*(A=>106, C=>135, D=>163, E=>194, F=>197, G=>84, H=>184, I=>169,
           K=>205, L=>164, M=>188, N=>157, P=>136, Q=>198, R=>248, S=>130,
           T=>142, V=>142, W=>227, Y=>222); */
#define MAX_ACC "106 135 163 194 197 084 184 169 205 164 188 157 136 198 248 130 142 142 227 222 160 180 196" 
#define ACC_NAM "ACDEFGHIKLMNPQRSTVWYBXZ"
#define ABUND "8.19 1.64 5.79 5.99 3.98 7.96 2.33 5.42 6.04 8.39 2.03 4.66 4.59 3.71 4.62 6.33 6.15 7.00 1.54 3.65"
#define ABUND_NAM "ACDEFGHIKLMNPQRSTVWY"

#define VERY_SMALL -FLT_MAX
#define VERY_BIG FLT_MAX
#define MAX_ID_LEN 30

const MAX_SEQ_LEN = 5000;
const PRECIS = 1000;


/*added 280503 to allow different atom-superimpositions like permutations in PHE.. */
struct permut_struct {
  int n; /*number of permutations*/
  int a; /*number of atoms*/
  int **p; /*permutations themselves: p[permut][atom] = 0 (1, 2, ...)
	  *this will be used to redirect atom equivalencing. For example
	  *0,1,2,3 is a 1:1 redirect; 1,2,3,0 is a rotation that puts the second atom on the first.... */
};


struct sc{
  float score;
  int n;
  int *members;
  struct sc *parentA;
  struct sc *parentB;
  char lab;
};

struct brookn{
  int n;
  char cid;
  char in;
};

struct domain_loc{		/* This structure allows rather complex domains to be described */
   char filename[1000];
   char filename2[1000];
   char id[30];
   int nobj;			/* The number of objects considered within the named file */
   int *type;			/* The type that each object is:
					0 ==> an error
					1 ==> All of the residues in the file
					2 ==> A particular chain
					3 ==> A particular named region (eg. A 25 _ to B 10 _ ) */
   struct brookn *start;	/* There will be a start and end for each 'object' */
   struct brookn *end;
   int *reverse;		/* if 1, then reverse invert the object N to C */
   int **coords;
   int **coords2;
   int **coords3;
   char *aa;
   char *align;
   char *oldalign;
   char *sec;
   struct brookn *numb;
   int *length;			/* length of object */
   int ncoords;
   int ncoords2;
   int *use;
   float **R;			/* Initial transformation */
   float *V;
   float **r;			/* current transformation, when STAMP is done, we must update(r,v,R,V) to get the final transformation */
   float *v;
   float value;
};

struct surf {
  int num;
  char cid;
  char in;
  char aa[4];
  float aac;
  float rac;
};

struct atom {
  int num;
  char type[4];
  int aa_num;
  char aa[4];
  char cid;
  char in;
  long int co[3];
  int temp;
  float aac;
  float rac;
};

struct point {
  int type;
  int number;
  struct atom **atoms;
  long int co[3];
};

struct store_struc {
  int *p1; /*array of indices for points[0]*/
  int *p2; /*array of indices for points[1]*/
  int n; /*number of equivalenced points*/
  int domain; /* number of domain for points[1] */
  double P; /* E-value */
  double Ps; /* standard E-value without corrections */
  float rms;
  float R[3][3];
  float V[3];
};

struct table_struct {
  char aa[4];
  char atom;
  char sp_atom[4];
  int type;
  float aac;
  float rac;
};

struct pseudo_struct {
  char aa[4];
  int natoms;
  char atom[15][4];
  int type;
  float aac;
  float rac;
};

struct parameter {
  float pf;               /* fac. parameter for statistics */
  float pe;               /* exp. parameter for statistics */
  int mode;               /* residue based (0) OR atoms based (1) */
  long int max_tolerance; /* distance tolerance during search */
  long int max_size;      /* max. diameter of match */
  int max_depth;          /* max. search depth (= max. depth of recursion = max. number of points in match */
  int max_hits;           /* max. number of matches during search (0 for off) */
  int max_rep;            /* max. number of matches reported (0 for all) */
  int max_rep_each;            /* max. number of matches reported from each search (0 for all) */
  float abs_surf;         /* min. surface accessibility (absolute) */
  float rel_surf;         /* min. surface accessibility (relative) */
  float max_rms;          /* max. RMSD for match */
  float max_P;            /* max. E-value for match */
  float max_Ps;           /* max. standard E-value (E700) for match */
  int min_atoms;          /* min. number of points for match DURING SEARCH*/
  float match_rel;        /* match match_rel % of atomset */
  int match_abs;          /* match always AT LEAST match_abs res. */
  /*  char match_all; */        /* match all atoms of one atomset (0/1/2) */
  char def_file[1000];    /* filename for type definitions (only if rms = 1)*/
  int file_type[2];       /* 0/1 = dom/pdb */
  char atom_file[2][1000];/* filename for dom file or pdb file */
  FILE *out;              /* filehandle for output */
  int out_format;         /* specifies output format */
  int verbose;            /* produce output, often used in Rob's routines */
  int found;              /* number of already found motifs */
  char heteroatoms;       /* load heteroatom entries (HETATM) from pdb-file */
  int group_filter;       /* filter similar motifs (0 / 1 / 2) */
  float min_group_sim;    /* min. similarity for motif-filter */
  int cluster_type;       /* cluster-type for motif filter */
  int res_fit;            /* RMSD calculation on points (0), residues (1) or weighted fit (2) */
  char view_permut;       /* do not exclude permutations (same residues, different equivalencing) of one motif */
  char view_subs;         /* do not remove submotifs entirely contained in another one -> show all */
  char AAA;               /* All Against All (0/1) */
  char linear;            /* linear search (residues have to be in the same order -> no redirection via j_array, faster */
  int long_range;         /* require at least one res. per motif to be long_range apart in sequence */
};


/* function prototypes */
char *getenv(const char *name);

void getdomain_error(char *buff);

int skiptononspace(char *string, int pointer);

int RBR_getdomain(FILE *IN, struct domain_loc *domains, int *ndomain, int maxdomain, int *gottrans, char *env, int DSSP, FILE *OUTPUT);

int RBR_domdefine(struct domain_loc *domain, int *gottrans, char *env, int DSSP, FILE *INPUT, FILE *OUTPUT);

char *RBR_getfile(char *code_safe, char *dirfile, int code_length, FILE *OUTPUT) ;

int RBR_count_domain(FILE *IN);

char RBR_a3to1(char *a3);

int RBR_getatoms(struct domain_loc *domain, struct atom **atoms, int *natoms, char *pdb_files, int verbose, int file_type, struct parameter *par);

int gettable(struct permut_struct **permut, struct table_struct **table, struct pseudo_struct **pseudo, char *file, int *ntable_def, int *npseudo_def, char *def_files, float **abund);

void a1to3(char aa, char new[4]);

/******************************************************************
make_entity: return pointer to malloc'd space for a struct sc
	structure for a single entity. i.
	base_val: The value form minimum distance, or maximum similarity.
-------------------------------------------------------------------*/
struct sc *make_entity(int i,float base_val)
{
	struct sc *ret_val;
	ret_val = (struct sc *) malloc(sizeof(struct sc));
	ret_val->score = base_val;
	ret_val->n = 1;
	ret_val->members = (int *) malloc(sizeof(int));
	ret_val->members[0] = i;
	ret_val->parentA = NULL;
	ret_val->parentB = NULL;
	ret_val->lab = 1;
	return ret_val;
}
	
/******************************************************************
val_A_B: Given the upper diagonal array, return the value stored
         in the notional arr[i][j].
         Where 0 <= i < n-1
         and   i < j < n

Note:  This could/should be turned into a macro - without the bounds
       check.
------------------------------------------------------------------*/

float val_A_B(
	float **arr,   
    	int n,          
    	int i,          
    	int j)          

{
  /*    extern FILE *std_err;*/
    if(!(i >= 0 && i < n-1 && i < j && j < n)){
      /*    	GJerror("Invalid request in val_A_B");*/
    	fprintf(stderr,"val_A_B : i,j,n %d %d %d \n",i,j,n);
    	exit(-1);
    }

    return arr[i][j-i-1];
}

/*******************************************************************
val_Grp_B: Given upper diagonal array,
	find min, max, or mean of comparison between the
	entities in grp and j
	choice = 0 =  find min value of grp-grp comparison
		 1 =  find max ...
		 2 =  find mean ...
		 
Note:  there is some duplication in the code to avoid testing choice
       on every step in the loop
--------------------------------------------------------------------*/
float val_Grp_B(
	float **arr,	/* upper diagonal array */
	int n,		/* side of array */
	int *grpA,	/* array containing entities of group */
	int ng,		/* number of members in grpA */
	int j,		/* entity to compare to grpA */
	int choice	/* see above */
)
{
	int k;
	float val;
	float rval;

	if(choice == 0){
		/* min case */
		val = VERY_BIG;
		for(k=0;k<ng;++k){
			if(grpA[k] < j){
				rval = val_A_B(arr,n,grpA[k],j);
			}else{
				rval = val_A_B(arr,n,j,grpA[k]);
			}
			if(rval < val) val = rval;
		}
	}else if (choice == 1){
		val = VERY_SMALL;
		for(k=0;k<ng;++k){
			if(grpA[k] < j){
				rval = val_A_B(arr,n,grpA[k],j);
			}else{
				rval = val_A_B(arr,n,j,grpA[k]);
			}
			if(rval > val) val = rval;
		}
	}else if (choice == 2){
		val = 0.0;
		for(k=0;k<ng;++k){
			if(grpA[k] < j){
				rval = val_A_B(arr,n,grpA[k],j);
			}else{
				rval = val_A_B(arr,n,j,grpA[k]);
			}
			val += rval;
		}
		val /= ng;
	}
	return val;
}
/*******************************************************************
val_Grp_Grp: Given upper diagonal array,
	find min, max, or mean of comparison between the
	entities in grpA and grpB
	choice = 0 =  find min value of grp-grp comparison
		 1 =  find max ...
		 2 =  find mean ...
		 
Note:  there is some duplication in the code to avoid testing choice
       on every step in the loop
--------------------------------------------------------------------*/
float val_Grp_Grp(
	float **arr,	/* upper diagonal array */
	int n,		/* side of array */
	int *grpA,	/* array containing entities of group */
	int ngA,	/* number of members in grpA */
	int *grpB,	/* array containing entities of group B*/
	int ngB,	/* number of members in grpB */
	int choice	/* see above */
)
{
	int k,l;
	float val;
	float rval;

	if(choice == 0){
		/* min case */
		val = VERY_BIG;
		for(k=0;k<ngA;++k){
			for(l=0;l<ngB;++l){
				if(grpA[k] < grpB[l]){
					rval = val_A_B(arr,n,grpA[k],grpB[l]);
				}else{
					rval = val_A_B(arr,n,grpB[l],grpA[k]);
				}
				if(rval < val) val = rval;
			}
		}
	}else if (choice == 1){
		val = VERY_SMALL;
		for(k=0;k<ngA;++k){
			for(l=0;l<ngB;++l){
				if(grpA[k] < grpB[l]){
					rval = val_A_B(arr,n,grpA[k],grpB[l]);
				}else{
					rval = val_A_B(arr,n,grpB[l],grpA[k]);
				}
				if(rval > val) val = rval;
			}
		}
	}else if (choice == 2){
		val = 0.0;
		for(k=0;k<ngA;++k){
			for(l=0;l<ngB;++l){
				if(grpA[k] < grpB[l]){
					rval = val_A_B(arr,n,grpA[k],grpB[l]);
				}else{
					rval = val_A_B(arr,n,grpB[l],grpA[k]);
				}
				val += rval;
			}
		}
		val /= (ngA * ngB);
	}
	return val;
}
/*****************************************************************
remove_unclust:

Take the array containing integers showing members that have
yet to be clustered and remove an element.  nunclust is returned
reduced by one.
-----------------------------------------------------------------*/
void remove_unclust(int *unclust,int *nunclust,int val)
{
    int i,j,found;
    i = 0;
    found = 0;
    
    while(i < *nunclust){
        if(unclust[i] == val){
            ++i;
            found = 1;
        }
        if(found){
            j = i - 1;
        }else{
            j = i;
        }
        unclust[j] = unclust[i];
        ++i;
    }
    --(*nunclust);
}
/*****************************************************************
sub_notparent:  given the notparent list substitutes A by B
-----------------------------------------------------------------*/
void sub_notparent(int *np,int *n,int A,int B)
{
  int i;
  for(i=0;i<*n;++i){
    if(np[i] == A){
      np[i] = B;
      return;
    }
  }
}
/*****************************************************************
remove_notparent:  given the notparent list removes A from the list
                   
-----------------------------------------------------------------*/
void remove_notparent(int *np,int *n,int A)
{
   remove_unclust(np,n,A);
}
/*****************************************************************
add_notparent:  given the notparent list adds A to the list
                   
-----------------------------------------------------------------*/
void add_notparent(int *np,int *n,int A)
{
   np[*n] = A;
   ++(*n);
}

/***************************************************************************/

struct sc *GJnoc(float **arr, int n, int type){

  float base_val;
  int nclust;
  int i,j,k,l;
  int i1loc,j1loc;
  int i2loc,j2loc;
  int i3loc,j3loc;
  float U1val,U2val,U3val,rval;
	
  struct sc *clust;
	struct sc **free_list; /* list of pointers to sc structures that are 
				  allocated by make_entity these are the ones 
				  that are not part of the main clust array, 
				  but must be 
				  freed before returning from this routine. */
	int nfree_list;

        int *unclust;      /* list of unclustered entities */
	int nunclust;      /* number of unclustered entities */

	int *notparent;    /* list of clusters that have no children */
	int nnotparent;    /* number of clusters that have no children */

	/*	int *inx; */         /* array of indexes - inx[i] is the re-ordered position */

	/*	char *postfile; */   /* name of PostScript output file */

	struct gjnoc *ret_val;  /* cluster structure to return for STAMP etc.*/

	/* options */
	int option;     /* use to feed the Grp_Grp function */

	int single_order;

	FILE *fp;

	single_order = 0;

	nfree_list = 1;
	
	base_val = VERY_BIG;

	/* holds index of unclustered entities */

	unclust = (int *)malloc(sizeof(int) * n);
	for(i=0;i<n;++i){
		unclust[i] = i;
	}
	nunclust = n;

	notparent = (int *)malloc(sizeof(int) * n);
	for(i=0;i<n;++i){
		notparent[i] = 0;
	}
	nnotparent = 0;
	
	/* set up the clust array to hold results */
	/* there are always n-1 clusters formed - we will also store the nth (all entities)*/
	clust = (struct sc *)malloc(sizeof(struct sc) * n);
	
	/* allocate space for the free list pointers - for now just length 1 */
	free_list = (struct sc **)malloc(sizeof(struct sc *) * nfree_list);

	nclust = 0;
	U1val = VERY_SMALL;
	U2val = VERY_SMALL;
	U3val = VERY_SMALL;

	  if(type == 1){
	    option = 1;
	  }else if(type == 2){
	    option = 0;
	  }else if(type == 3){
	    option = 2;
	  }

/* implement max similarity single linkage first */

	while(nclust < (n-1)){
		  U1val = VERY_SMALL;
		  U2val = VERY_SMALL;
		  U3val = VERY_SMALL;
		if(nunclust > 0){
		        if(!single_order || (single_order && nclust == 0)){
			  /* we've not clustered all entities into groups */
			  /* find max of what is left */
			  for(i=0;i<nunclust-1;++i){
				for(j=(i+1);j<nunclust;++j){
					rval = val_A_B(arr,n,unclust[i],unclust[j]);

					  if(rval > U1val){
						U1val = rval;
						i1loc = unclust[i];
						j1loc = unclust[j];
					  }
				}
			  } /* ... for(i=0... */ 
			} /* ...if(!single_order...*/
		}
		if(nnotparent > 0 && nunclust > 0){
			/* we have some clusters */
                        /* so compare them to the unclustered entitites */
			for(i=0;i<nunclust;++i){
				for(j=0;j<nnotparent;++j){
				        k = notparent[j];
					rval = val_Grp_B(arr,n,clust[k].members,clust[k].n,unclust[i],option);
				
					  if(rval >= U2val){
					        /* need >= because want biggest previous cluster */
						U2val = rval;
						i2loc = unclust[i];
						j2loc = k;
					  }
				
				}
			}
		}
		if(nnotparent > 1){
                        /* compare clusters to clusters if there are more than 1*/
                        for(i=0;i<nnotparent-1;++i){
			    k = notparent[i];
                            for(j=(i+1);j<nnotparent;++j){
			          l = notparent[j];
				  rval = val_Grp_Grp(arr,n,clust[k].members,clust[k].n,
                                                         clust[l].members,clust[l].n,
                                                         option);
				
				    if(rval >= U3val){
					U3val = rval;
					i3loc = k;
					j3loc = l;
				    } /* if(rval >= U3val... */
				
			      }  /* if j=(i+1) ... */
			} /* for(i=0... */ 
		}
		/* find which search gave the biggest (or smallest) value */
    		if(U3val >= U1val && U3val >= U2val){
    		    /* joining of  two clusters */
    		    clust[nclust].score = U3val;
    		    clust[nclust].n = clust[i3loc].n + clust[j3loc].n;
    		    clust[nclust].members = (int *)malloc(sizeof(int) * clust[nclust].n);
    		    for(i=0;i<clust[i3loc].n;++i){
    		        clust[nclust].members[i] = clust[i3loc].members[i];
    		    }
    		    for(i=0;i<clust[j3loc].n;++i){
    		        clust[nclust].members[i+clust[i3loc].n] = clust[j3loc].members[i];
    		    }
    		    clust[nclust].parentA = &clust[i3loc];
    		    clust[nclust].parentB = &clust[j3loc];
		    clust[nclust].lab = 1;
		    remove_notparent(notparent,&nnotparent,i3loc);
		    sub_notparent(notparent,&nnotparent,j3loc,nclust);
    		}else if(U2val >= U1val && U2val >= U3val){
    		    /* joining of  a single entity to an existing cluster */
    		    clust[nclust].score = U2val;
    		    clust[nclust].n = 1 + clust[j2loc].n;
    		    clust[nclust].members = (int *)malloc(sizeof(int) * clust[nclust].n);
    		    /* copy previous cluster list into this one - not strictly neccessary */
    		    /* but done to simplify cluster comparisons above*/
    		    for(i=1;i<clust[nclust].n;++i){
    		        clust[nclust].members[i] = clust[j2loc].members[i-1];
    		    }
    		    /* stick solitary cluster member on the beginning */
    		    clust[nclust].members[0] = i2loc;
    		    /* put in parent pointers */
    		    clust[nclust].parentA = make_entity(i2loc,base_val);

		    free_list[nfree_list - 1] = clust[nclust].parentA;
		    ++nfree_list;
		    free_list = (struct sc **)realloc(free_list, sizeof(struct sc *) * nfree_list);

    		    clust[nclust].parentB = &clust[j2loc];
		    clust[nclust].lab = 1;
		    sub_notparent(notparent,&nnotparent,j2loc,nclust);
    		    remove_unclust(unclust,&nunclust,i2loc);
		}else if(U1val >= U2val && U1val >= U3val){
		    /* unclustered pair were biggest */
		    clust[nclust].score = U1val;
    		    clust[nclust].n = 2;
		    clust[nclust].members = (int *)malloc(sizeof(int) * clust[nclust].n);
		    clust[nclust].members[0] = i1loc;
		    clust[nclust].members[1] = j1loc;
		    clust[nclust].parentA = make_entity(i1loc,base_val);
		    free_list[nfree_list - 1] = clust[nclust].parentA;
		    ++nfree_list;
		    free_list = (struct sc **)realloc(free_list, sizeof(struct sc *) * nfree_list);

    		    clust[nclust].parentB = make_entity(j1loc,base_val);
		    free_list[nfree_list - 1] = clust[nclust].parentB;
		    ++nfree_list;
		    free_list = (struct sc **)realloc(free_list, sizeof(struct sc *) * nfree_list);

    		    clust[nclust].lab = 1;
		    add_notparent(notparent,&nnotparent,nclust);
    		    remove_unclust(unclust,&nunclust,i1loc);
		    remove_unclust(unclust,&nunclust,j1loc);
		}
		++nclust;
	}
	/* now we should have all clusters stored in the clust array */
	/* set up the inx array */
	/*	inx = (int *)malloc(sizeof(int) * n);
	for(i=0;i<clust[nclust-1].n;++i){
	    inx[clust[nclust-1].members[i]]=i;
	}
	*/
	  /* output all clusters from smallest to largest */
    	    /* free all memory  - 
	       freeing clust is not straightforward - this is why we have
	       free_list of pointers*/
	    free(unclust);
	    free(notparent);
	    /*	    free(inx);*/
	    for(i=0;i<(nfree_list-1);++i) {
	      free(free_list[i]->members);
	      free(free_list[i]);
	    }
	    free(free_list);

	    return clust;
	  
}



/* qkfit.f -- translated by f2c (version of 11 May 1990  14:38:39).
   You must link the resulting object file with the libraries:
	-lF77 -lI77 -lm -lc   (in that order)
*/

/* for birdie
#include </g/russell1/alstark/3D_Comp/rbr_f2c.h>


#include </russell/rg1/alstark/3D_Comp/rbr_f2c.h>
*/
/* Table of constant values */

static integer c__3 = 3;
static integer c__0 = 0;

/* Subroutine */ int RBR_qkfit(umat, rtsum, r, entry_)
doublereal *umat, *rtsum, *r;
integer *entry_;
{
    /* Initialized data */

    static doublereal eps = 1e-5;
    static doublereal pi = 3.14159265358979;
    static doublereal one = 1.;
    static doublereal two = 2.;
    static doublereal three = 3.;
    static doublereal half = .5;
    static doublereal third = .333333333;
    static doublereal forthr = 1.333333333;
    static struct {
	doublereal fill_1[1];
	doublereal e_2[2];
	doublereal fill_3[2];
	doublereal e_4;
	doublereal fill_5[3];
	} equiv_6 = { {0}, 0., 0., {0}, 0. };


    /* System generated locals */
    integer i_1;
    static doublereal equiv_7[9];

    /* Builtin functions */
    double sqrt(), d_sign(), atan(), cos();

    /* Local variables */
    static doublereal diff;
    static integer isig;
    static doublereal detu, root[3];
#define a ((doublereal *)&equiv_6)
#define b (equiv_7)
    static integer i, j, k;
    static doublereal s, t;
    extern /* Subroutine */ int eigen_();
    static doublereal digav, theta, argsq, b1, b2;
    extern /* Subroutine */ int esort_();
    static doublereal cos3th, cc, b13, dd, b23;
    static integer ia;
    static doublereal b33, qq, rt;
#define usqmat ((doublereal *)&equiv_6)
#define aam ((doublereal *)&equiv_6)
#define bam ((doublereal *)&equiv_6 + 4)
#define cam ((doublereal *)&equiv_6 + 8)
#define fam ((doublereal *)&equiv_6 + 7)
#define gam ((doublereal *)&equiv_6 + 6)
#define ham ((doublereal *)&equiv_6 + 3)
    static doublereal du11, du21, du31;
#define utr (equiv_7)





    /* Parameter adjustments */
    r -= 4;
    umat -= 4;

    /* Function Body */
    isig = 1;
    if (*entry_ == 1) {
	goto L200;
    }
/*     CALC DET OF UMAT */

    du11 = umat[8] * umat[12] - umat[11] * umat[9];
    du21 = umat[11] * umat[6] - umat[5] * umat[12];
    du31 = umat[5] * umat[9] - umat[8] * umat[6];
    detu = umat[4] * du11 + umat[7] * du21 + umat[10] * du31;
    if (detu < 0.) {
	isig = -1;
    }

/*     FORM USQMAT AS POSITIVE SEMI DEFINITE MATRIX */

    for (j = 1; j <= 3; ++j) {
	i_1 = j;
	for (i = 1; i <= i_1; ++i) {
	    usqmat[i + j * 3 - 4] = umat[i * 3 + 1] * umat[j * 3 + 1] + umat[
		    i * 3 + 2] * umat[j * 3 + 2] + umat[i * 3 + 3] * umat[j * 
		    3 + 3];
/* L105: */
	}
/* L110: */
    }
/* %    WRITE(6,999) USQMAT */

/*     REDUCE AVG OF DIAGONAL TERMS TO ZERO */

    digav = (*aam + *bam + *cam) * third;
/* %    WRITE(6,999) DIGAV */
    *aam -= digav;
    *bam -= digav;
    *cam -= digav;

/*     SETUP COEFFS OF SECULAR EQUATION OF MATRIX WITH TRACE ZERO */

    cc = *fam * *fam + *gam * *gam + *ham * *ham - *aam * *bam - *bam * *cam 
	    - *cam * *aam;
    dd = *aam * *bam * *cam + two * (*fam * *gam * *ham) - *aam * *fam * *fam 
	    - *bam * *gam * *gam - *cam * *ham * *ham;

/*     THE SECULAR EQN IS Y**3-CC*Y-DD=0  AND DD IS DET(USQMAT) */
/*     REDUCE THIS TO THE FORM COS**3-(3/4)COS- */
/*     (1/4)COS3THETA = 0 */
/*     WITH SOLUTIONS COSTHETA.  SO Y=QQ*COSTHETA */

    if (cc <= eps) {
	goto L115;
    }
    qq = sqrtf(forthr * cc);
    cos3th = three * dd / (cc * qq);
    if (abs(cos3th) > one) {
/*	cos3th = d_sign(&one, &cos3th); */
	cos3th = (cos3th > 0 ? one:-one); 
    }

/*     FUNCTION ARCOS */

    if (cos3th != 0.) {
	goto L1200;
    }
/* L1100: */
    theta = (float)1.570796327;
    goto L1400;
L1200:
    argsq = cos3th * cos3th;
    theta = atan(sqrtf((float)1. - argsq) / cos3th);
    if (cos3th < 0.) {
	theta = pi - theta;
    }
L1400:

/*     ROOTS IN ORDER OF SIZE GO 1,2,3 1 LARGEST */

    theta *= third;
    root[0] = qq * cos(theta);
    diff = half * sqrtf(three * (qq * qq - root[0] * root[0]));
    root[1] = -root[0] * half + diff;
    root[2] = -root[0] * half - diff;
    goto L120;
L115:

/*     SPECIAL FOR TRIPLY DEGENERATE */

    root[0] = (float)0.;
    root[1] = (float)0.;
    root[2] = (float)0.;
L120:
/*     ADD ON DIGAV AND TAKE SQRT */
    for (j = 1; j <= 3; ++j) {
	rt = root[j - 1] + digav;
	if (rt < eps) {
	    rt = (float)0.;
	}
	root[j - 1] = sqrtf(rt);
/* L125: */
    }
/* %    WRITE(6,999) ROOT */
/*     IF DETU IS <0 CHANGE SIGN OF ROOT(3) */
    if (isig == -1) {
	root[2] = -root[2];
    }
    *rtsum = root[0] + root[1] + root[2];
/* %    WRITE(6,999) RTSUM */
    return 0;

/*     THIS IS THE FANCY PART */

L200:

/*     FORM USQ = (UT).U    (IN UPPER TRIANGULAR SYMMETRIC STORAGE MODE) 
*/

    for (i = 1; i <= 3; ++i) {
	for (j = i; j <= 3; ++j) {
	    t = (float)0.;
	    for (k = 1; k <= 3; ++k) {
		t += umat[k + i * 3] * umat[k + j * 3];
/* L210: */
	    }
	    ia = i + (j * j - j) / 2;
	    utr[ia - 1] = t;
/* L220: */
	}
    }
/* %    WRITE(6,999) UTR */

/*     CALCULATE EIGENVALUES AND VECTORS */

    eigen_(utr, a, &c__3, &c__0);
    esort_(utr, a, &c__3, &c__0);
/* %    WRITE(6,999) UTR */

    root[0] = utr[0];
    root[1] = utr[2];
    root[2] = utr[5];
/* %    WRITE(6,999) ROOT */
/* %    WRITE(6,999) A */

/*     SET A3 = A1 CROSS A2 */
/*     ROOTS ARE IN ORDER R(1) >= R(2) >= R(3) >= 0 */

    a[6] = a[1] * a[5] - a[2] * a[4];
    a[7] = a[2] * a[3] - a[0] * a[5];
    a[8] = a[0] * a[4] - a[1] * a[3];
/* %    WRITE(6,999) A */

/*     VECTOR SET B=U.A */

    for (i = 1; i <= 3; ++i) {
	for (j = 1; j <= 3; ++j) {
	    t = (float)0.;
	    for (k = 1; k <= 3; ++k) {
/* L230: */
		t += umat[j + k * 3] * a[k + i * 3 - 4];
	    }
	    b[j + i * 3 - 4] = t;
/* L240: */
	}
    }

/*      NORMALIZE B1 AND B2 AND CALCULATE B3 = B1 CROSS B2 */

    b1 = sqrtf(b[0] * b[0] + b[1] * b[1] + b[2] * b[2]);
    b2 = sqrtf(b[3] * b[3] + b[4] * b[4] + b[5] * b[5]);
    for (i = 1; i <= 3; ++i) {
	b[i - 1] /= b1;
/* L250: */
	b[i + 2] /= b2;
    }

/*      CHECK FOR LEFT HANDED ROTATION */

    b13 = b[1] * b[5] - b[2] * b[4];
    b23 = b[2] * b[3] - b[0] * b[5];
    b33 = b[0] * b[4] - b[1] * b[3];

    s = b13 * b[6] + b23 * b[7] + b33 * b[8];
    if (s < 0.) {
	isig = -1;
    }
    b[6] = b13;
    b[7] = b23;
    b[8] = b33;
/* %    WRITE(6,999) B */

/*     CALCULATE ROTATION MATRIX R */

    for (i = 1; i <= 3; ++i) {
	for (j = 1; j <= 3; ++j) {
	    t = (float)0.;
	    for (k = 1; k <= 3; ++k) {
/* L260: */
		t += b[i + k * 3 - 4] * a[j + k * 3 - 4];
	    }
	    r[i + j * 3] = t;
/* L270: */
	}
    }

/*     RMS ERROR */

    for (i = 1; i <= 3; ++i) {
	if (root[i - 1] < 0.) {
	    root[i - 1] = (float)0.;
	}
	root[i - 1] = sqrtf(root[i - 1]);
/* L280: */
    }

/*     CHANGE SIGN OF EVAL #3 IF LEFT HANDED */

    if (isig < 0) {
	root[2] = -root[2];
    }
    *rtsum = root[2] + root[1] + root[0];
    return 0;
} /* qkfit_ */

#undef utr
#undef ham
#undef gam
#undef fam
#undef cam
#undef bam
#undef aam
#undef usqmat
#undef b
#undef a


/*---- SUBROUTINE TO COMPUTE EIGENVALUES & EIGENVECTORS OF A REAL SYMMETRIC*/
/* ---- MATRIX, STOLEN FROM IBM SSP MANUAL (SEE P165) */
/* ---- DESCRIPTION OF PARAMETERS - */
/* ---- A - ORIGINAL MATRIX STORED COLUMNWISE AS UPPER TRIANGLE ONLY, */
/* ---- I.E. "STORAGE MODE" = 1.  EIGENVALUES ARE WRITTEN INTO DIAGONAL */
/* ---- ELEMENTS OF A  I.E.  A(1)  A(3)  A(6)  FOR A 3*3 MATRIX. */
/* ---- R - RESULTANT MATRIX OF EIGENVECTORS STORED COLUMNWISE IN SAME */
/* ---- ORDER AS EIGENVALUES. */
/* ---- N - ORDER OF MATRICES A & R. */
/* ---- MV = 0 TO COMPUTE EIGENVALUES & EIGENVECTORS. */
/* Subroutine */ int eigen_(a, r, n, mv)
doublereal *a, *r;
integer *n, *mv;
{
    /* System generated locals */
    integer i_1, i_2;
    doublereal d_1;

    /* Builtin functions */
    double sqrt();

    /* Local variables */
    static doublereal cosx, sinx, cosx2, sinx2;
    static integer i, j, l, m;
    static doublereal x, y, range, anorm, sincs, anrmx;
    static integer ia, ij, il, im, ll, lm, iq, mm, lq, mq, ind, ilq, imq, ilr,
	     imr;
    static doublereal thr;

/* ---- DOUBLE PRECISION A,R,ANORM,ANRMX,THR,X,Y,SINX,SINX2,COSX, */
/* 	1COSX2,SINCS,RANGE */
/*---- FOR DOUBLE PRECISION, SQRT IN STATEMENTS 40,68,75&78 MUST BE DSQRT,
*/
/* ---- ABS IN 62 MUST BE DABS AND 1.E-6 IN 5 MUST BE 1.D-12 . */
    /* Parameter adjustments */
    --r;
    --a;

    /* Function Body */
/* L5: */
    range = (float)1e-6;
    if (*mv - 1 != 0) {
	goto L10;
    } else {
	goto L25;
    }
L10:
    iq = -(*n);
    i_1 = *n;
    for (j = 1; j <= i_1; ++j) {
	iq += *n;
	i_2 = *n;
	for (i = 1; i <= i_2; ++i) {
	    ij = iq + i;
	    r[ij] = (float)0.;
	    if (i - j != 0) {
		goto L20;
	    } else {
		goto L15;
	    }
L15:
	    r[ij] = (float)1.;
L20:
	    ;
	}
    }
/* ---- INITIAL AND FINAL NORMS (ANORM & ANRMX) */
L25:
    anorm = (float)0.;
    i_2 = *n;
    for (i = 1; i <= i_2; ++i) {
	i_1 = *n;
	for (j = i; j <= i_1; ++j) {
	    if (i - j != 0) {
		goto L30;
	    } else {
		goto L35;
	    }
L30:
	    ia = i + (j * j - j) / 2;
/* Computing 2nd power */
	    d_1 = a[ia];
	    anorm += d_1 * d_1;
L35:
	    ;
	}
    }
    if (anorm <= 0.) {
	goto L165;
    } else {
	goto L40;
    }
L40:
    anorm = sqrtf(anorm * (float)2.);
    anrmx = anorm * range / *n;
/* ---- INITIALIZE INDICATORS AND COMPUTE THRESHOLD */
    ind = 0;
    thr = anorm;
L45:
    thr /= *n;
L50:
    l = 1;
L55:
    m = l + 1;
/* ---- COMPUTE SIN & COS */
L60:
    mq = (m * m - m) / 2;
    lq = (l * l - l) / 2;
    lm = l + mq;
/* L62: */
    if ((d_1 = a[lm], abs(d_1)) - thr >= 0.) {
	goto L65;
    } else {
	goto L130;
    }
L65:
    ind = 1;
    ll = l + lq;
    mm = m + mq;
    x = (a[ll] - a[mm]) * (float).5;
/* L68: */
/* Computing 2nd power */
    d_1 = a[lm];
    y = -a[lm] / sqrtf(d_1 * d_1 + x * x);
    if (x >= 0.) {
	goto L75;
    } else {
	goto L70;
    }
L70:
    y = -y;
L75:
    sinx = y / sqrtf((sqrtf((float)1. - y * y) + (float)1.) * (float)2.);
/* Computing 2nd power */
    d_1 = sinx;
    sinx2 = d_1 * d_1;
/* L78: */
    cosx = sqrtf((float)1. - sinx2);
/* Computing 2nd power */
    d_1 = cosx;
    cosx2 = d_1 * d_1;
    sincs = sinx * cosx;
/* ---- ROTATE L & M COLUMNS */
    ilq = *n * (l - 1);
    imq = *n * (m - 1);
    i_1 = *n;
    for (i = 1; i <= i_1; ++i) {
	iq = (i * i - i) / 2;
	if (i - l != 0) {
	    goto L80;
	} else {
	    goto L115;
	}
L80:
	if ((i_2 = i - m) < 0) {
	    goto L85;
	} else if (i_2 == 0) {
	    goto L115;
	} else {
	    goto L90;
	}
L85:
	im = i + mq;
	goto L95;
L90:
	im = m + iq;
L95:
	if (i - l >= 0) {
	    goto L105;
	} else {
	    goto L100;
	}
L100:
	il = i + lq;
	goto L110;
L105:
	il = l + iq;
L110:
	x = a[il] * cosx - a[im] * sinx;
	a[im] = a[il] * sinx + a[im] * cosx;
	a[il] = x;
L115:
	if (*mv - 1 != 0) {
	    goto L120;
	} else {
	    goto L125;
	}
L120:
	ilr = ilq + i;
	imr = imq + i;
	x = r[ilr] * cosx - r[imr] * sinx;
	r[imr] = r[ilr] * sinx + r[imr] * cosx;
	r[ilr] = x;
L125:
	;
    }
    x = a[lm] * (float)2. * sincs;
    y = a[ll] * cosx2 + a[mm] * sinx2 - x;
    x = a[ll] * sinx2 + a[mm] * cosx2 + x;
    a[lm] = (a[ll] - a[mm]) * sincs + a[lm] * (cosx2 - sinx2);
    a[ll] = y;
    a[mm] = x;
/* ---- TESTS FOR COMPLETION */
/* ---- TEST FOR M = LAST COLUMN */
L130:
    if (m - *n != 0) {
	goto L135;
    } else {
	goto L140;
    }
L135:
    ++m;
    goto L60;
/* ---- TEST FOR L =PENULTIMATE COLUMN */
L140:
    if (l - (*n - 1) != 0) {
	goto L145;
    } else {
	goto L150;
    }
L145:
    ++l;
    goto L55;
L150:
    if (ind - 1 != 0) {
	goto L160;
    } else {
	goto L155;
    }
L155:
    ind = 0;
    goto L50;
/* ---- COMPARE THRESHOLD WITH FINAL NORM */
L160:
    if (thr - anrmx <= 0.) {
	goto L165;
    } else {
	goto L45;
    }
L165:
    return 0;
/*---- SORT EIGENVALUES AND EIGENVECTORS IN DESCENDING ORDER OF EIGENVALUE
S*/
} /* eigen_ */

/* Subroutine */ int esort_(a, r, n, mv)
doublereal *a, *r;
integer *n, *mv;
{
    /* System generated locals */
    integer i_1, i_2, i_3;

    /* Local variables */
    static integer i, j, k;
    static doublereal x;
    static integer ll, iq, jq, mm, ilr, imr;

    /* Parameter adjustments */
    --r;
    --a;

    /* Function Body */
    iq = -(*n);
    i_1 = *n;
    for (i = 1; i <= i_1; ++i) {
	iq += *n;
	ll = i + (i * i - i) / 2;
	jq = *n * (i - 2);
	i_2 = *n;
	for (j = i; j <= i_2; ++j) {
	    jq += *n;
	    mm = j + (j * j - j) / 2;
	    if (a[ll] - a[mm] >= 0.) {
		goto L185;
	    } else {
		goto L170;
	    }
L170:
	    x = a[ll];
	    a[ll] = a[mm];
	    a[mm] = x;
	    if (*mv - 1 != 0) {
		goto L175;
	    } else {
		goto L185;
	    }
L175:
	    i_3 = *n;
	    for (k = 1; k <= i_3; ++k) {
		ilr = iq + k;
		imr = jq + k;
		x = r[ilr];
		r[ilr] = r[imr];
/* L180: */
		r[imr] = x;
	    }
L185:
	    ;
	}
    }
    return 0;
} /* esort_ */



/* Rob's rms fitting function. It needs qkfit.c. The coords are
   integers. The resulting transformation is in the sense of: 
   atoms1 = R*atoms2 + V */
float RBR_imatfit(atoms1,atoms2,R,V,nats,entry,PRECISION)
long int **atoms1,**atoms2; 
float R[3][3], V[3];
int nats, entry;
int PRECISION;
{
	/* Note that everything that is passed to the routine
	 *  qkfit must be double precision. */
	double cma[3], cmb[3];
	double umat[3][3];
	double xasq, xbsq, xni,t,rtsum,rmse;
	double r[3][3];
	int i,j,k,xn;

	xn=nats;
	xasq=0.0;
	xbsq=0.0;
	xni=1.0/((double)xn);

/* Accumulate uncorrected (for c.m.) sums and squares */

/*        printf("Matfit... %d atoms, entry is %d\n",nats,entry);
	for(j=0; j<nats; ++j) {
           for(k=0; k<3; ++k) printf("%8d ",atoms1[j][k]);
           printf("and ");
           for(k=0; k<3; ++k) printf("%8d ",atoms2[j][k]); 
	   printf("\n");
	}
*/

	for(i=0; i<3; ++i) {
	   cma[i]=cmb[i]=0.0;
	   for(j=0; j<=2; ++j) umat[j][i]=0.0;

           for(j=0; j<nats; ++j) {
	      for(k=0; k<3; ++k) {
	      umat[k][i]+=(double)atoms1[j][i]/(double)PRECISION*(double)atoms2[j][k]/(double)PRECISION;
	      }  /* End of for(k=... */

		
	      t=(double)atoms1[j][i]/(double)PRECISION;
	      cma[i]+=t;
	      xasq+=t*t;
	      t=(double)atoms2[j][i]/(double)PRECISION;
	      cmb[i]+=t;
	      xbsq+=t*t;

	   } /*  End of for(j=... */
	}  /* End of for(i=... */

/* Subtract cm offsets */

	for(i=0; i<=2; ++i) {
	   xasq-=cma[i]*cma[i]*xni;
	   xbsq-=cmb[i]*cmb[i]*xni;
	   for(j=0; j<=2; ++j) 
	       umat[j][i]=(umat[j][i]-cma[i]*cmb[j]*xni)*xni;
	} /* End of for(i=... */

/* Fit it */

	RBR_qkfit(&umat[0][0],&rtsum,&r[0][0],&entry);
	rmse=(xasq+xbsq)*xni - 2.0*rtsum;
	if(rmse<0.0) rmse=0.0;
	rmse=sqrtf(rmse);

/* The matrix obtained from the FORTRAN routine (r) must be transfered to
 *  the 'malloc'ed C matrix (R).  */
	for(i=0; i<=2; ++i)
	   for(j=0; j<=2; ++j) R[i][j]=(float)r[j][i];


/* Calculate offset if entry=1 */

	if(!entry) return rmse;
	for(i=0; i<=2; ++i) {
	   t=0.0;
	   for(j=0; j<=2; ++j)  t+=((double)R[i][j])*cmb[j];
	   V[i]=(float)(cma[i]-t)*xni;
	} /* End of for(i=... */

/*	printf("Matfit..\n");
	for(i=0; i<=2; ++i) {
	   for(j=0; j<=2; ++j) printf("%10.5f ",R[i][j]);
	   printf(" & %10.5f\n",V[i]);
	} 
*/
	return rmse;

} /* End of function */


/* calc_dist() calculates all pairwise distances between all atoms in points. The parameters are *points
   (pointer) to the array of points and npoints (number of entries). The program loops from i=0 to 
   i<npoints (considering all columns) and from j=i to n<npoints (considering the upper half of the matrix including
   the diagonal. The lower half of the matrix is copied from the upper half by swapping the indexes i and j.
   Distances are rounded to integer values.*/
/*
long int **calc_dist(struct point *points, int npoints){
  int i, j;
  long int **dist;
  if ((dist = (long int **)malloc((npoints)*(sizeof(long int*)))) == NULL){
    fprintf(stderr, "ERROR: Memory overflow while calculating distances\n");
    return NULL;
  }
  for(i=0; i<npoints; i++){
    if((dist[i] = (long int *)malloc((npoints*(sizeof(long int))))) == NULL){
      fprintf(stderr, "ERROR: Memory overflow while calculating distances\n");
      return NULL;
    }
  }
  for (i=0; i<(npoints); i++){
    for(j=i; j<(npoints); j++){
      dist[i][j] = (long int) (sqrtf(powf(((points[i]).co[0]-(points[j]).co[0]), 2) + powf(((points[i]).co[1]-(points[j]).co[1]), 2) + powf(((points[i]).co[2]-(points[j]).co[2]), 2))+0.5);
      dist[j][i] = dist[i][j];
    }
  }
  return dist;
}
*/

int printdomain(FILE *TRANS, struct domain_loc *domain, int writetrans) {

	int j,k;
   	fprintf(TRANS,"%s ",domain->filename);
	fprintf(TRANS,"%s {",domain->id);
	for(j=0; j<domain->nobj; ++j) {
	  switch(domain->type[j]){
	  case 1:
	    fprintf(TRANS," ALL ");
	    break;
	  case 2: 
	    if(domain->start[j].cid==' ') domain->start[j].cid='_';
	    fprintf(TRANS," CHAIN %c ",domain->start[j].cid);
	    break;
	  case 3: 
	    if(domain->start[j].cid==' ') domain->start[j].cid='_';
	    if(domain->start[j].in==' ') domain->start[j].in='_';
	    if(domain->end[j].cid==' ') domain->end[j].cid='_';
	    if(domain->end[j].in==' ') domain->end[j].in='_';
	    fprintf(TRANS," %c %d %c to %c %d %c ",
			  domain->start[j].cid,domain->start[j].n,
			  domain->start[j].in,
			  domain->end[j].cid,domain->end[j].n,
		          domain->end[j].in); break;
	    if(domain->start[j].cid=='_') domain->start[j].cid=' ';
	    if(domain->start[j].in=='_') domain->start[j].in=' ';
	    if(domain->end[j].cid=='_') domain->end[j].cid=' ';
	    if(domain->end[j].in=='_') domain->end[j].in=' ';
	    break;
	  } /* end of switch... */
	}
	if(writetrans) {
	   fprintf(TRANS,"\n");
	   for(j=0; j<3; ++j) {
	     for(k=0; k<3; ++k) fprintf(TRANS,"%10.5f ",domain->R[j][k]); 
	     fprintf(TRANS,"     %10.5f ",domain->V[j]);
	     if(j!=2) fprintf(TRANS," \n");
	   }
	} /* end of if(write... */
	fprintf(TRANS," }\n");

	return 0;
}

int linedomain(FILE *TRANS, struct domain_loc *domain){

	int j,k;
   	fprintf(TRANS,"%s ",domain->filename);
	fprintf(TRANS,"%s {",domain->id);
	for(j=0; j<domain->nobj; ++j) {
	  switch(domain->type[j]){
	  case 1:
	    fprintf(TRANS," ALL ");
	    break;
	  case 2: 
	    if(domain->start[j].cid==' ') domain->start[j].cid='_';
	    fprintf(TRANS," CHAIN %c ",domain->start[j].cid);
	    break;
	  case 3: 
	    if(domain->start[j].cid==' ') domain->start[j].cid='_';
	    if(domain->start[j].in==' ') domain->start[j].in='_';
	    if(domain->end[j].cid==' ') domain->end[j].cid='_';
	    if(domain->end[j].in==' ') domain->end[j].in='_';
	    fprintf(TRANS," %c %d %c to %c %d %c ",
			  domain->start[j].cid,domain->start[j].n,
			  domain->start[j].in,
			  domain->end[j].cid,domain->end[j].n,
		          domain->end[j].in); break;
	    if(domain->start[j].cid=='_') domain->start[j].cid=' ';
	    if(domain->start[j].in=='_') domain->start[j].in=' ';
	    if(domain->end[j].cid=='_') domain->end[j].cid=' ';
	    if(domain->end[j].in=='_') domain->end[j].in=' ';
	    break;
	  } /* end of switch... */
	}
	fprintf(TRANS," }");

	return 0;
}


int print_line(struct store_struc *store, int dom1, struct point **points, int *natoms, struct atom **atoms, struct domain_loc *domains, struct parameter *par){
  int i, j;

  for(j=0; ((j<par->found) && ((j<par->max_rep) || (par->max_rep==0))); j++){
    fprintf(par->out, "%g | %g | %f | %s | %s | %i | %i | ",  store[j].P, store[j].Ps, store[j].rms, domains[dom1].id, domains[store[j].domain].id, j+1, store[j].n);
    linedomain(par->out, &(domains[dom1]));
    fprintf(par->out, " | ");
    linedomain(par->out, &(domains[store[j].domain]));
    fprintf(par->out, " | %f | %f | %f | %f", store[j].R[0][0], store[j].R[0][1], store[j].R[0][2], store[j].V[0]);
    fprintf(par->out, " | %f | %f | %f | %f", store[j].R[1][0], store[j].R[1][1], store[j].R[1][2], store[j].V[1]);
    fprintf(par->out, " | %f | %f | %f | %f", store[j].R[2][0], store[j].R[2][1], store[j].R[2][2], store[j].V[2]);
    for(i=0; i<store[j].n; i++){
      fprintf(par->out, " | %s %c %i %c - %s %c %i %c", (*(points[dom1][(store[j].p1[i])].atoms[0])).aa, (*(points[dom1][(store[j].p1[i])].atoms[0])).cid,
	      (*(points[dom1][(store[j].p1[i])].atoms[0])).aa_num, (*(points[dom1][(store[j].p1[i])].atoms[0])).in,
	      (*(points[store[j].domain][(store[j].p2[i])].atoms[0])).aa, (*(points[store[j].domain][(store[j].p2[i])].atoms[0])).cid,
	      (*(points[store[j].domain][(store[j].p2[i])].atoms[0])).aa_num, (*(points[store[j].domain][(store[j].p2[i])].atoms[0])).in);
    }
    fprintf(par->out, "\n");
  }
  return j;
}


int print_domain(struct store_struc *store, int dom1, struct point **points, int *natoms, struct atom **atoms, struct domain_loc *domains, struct parameter *par){
  int i, j;

  printdomain(par->out, &(domains[dom1]), 0);
  for(j=0; ((j<par->found) && ((j<par->max_rep) || (par->max_rep==0))); j++){
    fprintf(par->out, "%% MOTIF %i : %i ATOMS FITTED WITH E = %g (E(700) = %g; RMSD = %f)\n", j+1, store[j].n, store[j].P, store[j].Ps, store[j].rms);
    fprintf(par->out, "%s %s_%i {", domains[store[j].domain].filename, domains[store[j].domain].id, j);
    for(i=0; i<store[j].n; i++){
      fprintf(par->out, " %c %i %c to %c %i %c", (*(points[store[j].domain][(store[j].p2[i])].atoms[0])).cid, (*(points[store[j].domain][(store[j].p2[i])].atoms[0])).aa_num,
	      (*(points[store[j].domain][(store[j].p2[i])].atoms[0])).in, (*(points[store[j].domain][(store[j].p2[i])].atoms[0])).cid,
	      (*(points[store[j].domain][(store[j].p2[i])].atoms[0])).aa_num, (*(points[store[j].domain][(store[j].p2[i])].atoms[0])).in);
    }
    fprintf(par->out, "\n");
    fprintf(par->out, "%10.5f %10.5f %10.5f      %10.5f\n", store[j].R[0][0], store[j].R[0][1], store[j].R[0][2], store[j].V[0]);
    fprintf(par->out, "%10.5f %10.5f %10.5f      %10.5f\n", store[j].R[1][0], store[j].R[1][1], store[j].R[1][2], store[j].V[1]);
    fprintf(par->out, "%10.5f %10.5f %10.5f      %10.5f }\n", store[j].R[2][0], store[j].R[2][1], store[j].R[2][2], store[j].V[2]);
  }
  return j;
}


int print_pdb(struct store_struc *store, int dom1, struct point **points, int *natoms, struct atom **atoms, struct domain_loc *domains, struct parameter *par){
  /* prints all atoms of dom1 and the transformed atoms of all other matches to one file */ 
  int i, j, n, k;
  char chain[50];
  float x,y,z,xn,yn,zn;

  /* possible chain IDs */
  strcpy(chain, "ABCDEFGHIJKLMNOPQRSTUVWXYZ123456789 !@#$%^&*()_+?<>");
 
 /* print Header of PDB file */
  fprintf(par->out, "REMARK PINTS - PATTERNS IN NONHOMOLOGOUS TERTIAERY STRUCTURE\n");
  fprintf(par->out, "REMARK A. STARK & R.B. RUSSELL, 2002\n");
  fprintf(par->out, "REMARK %i MOTIFS HAVE BEEN FOUND\n", par->found);
  fprintf(par->out, "REMARK ONLY 50 DIFF MOTIFS ARE SHOWN IN THIS OUTPUT-FORMAT\n");
  fprintf(par->out, "REMARK CHAIN 0 : ALL ATOMS FROM %s\n", domains[dom1].id);

  /* print ATOMS for first domain dom1 */
  n=0;
  for(i=0; i<natoms[dom1]; i++){
    x = (float)(atoms[dom1][i].co[0]);
    y = (float)(atoms[dom1][i].co[1]);
    z = (float)(atoms[dom1][i].co[2]);
    fprintf(par->out, "ATOM  %5d %4s%4s %1c%4d    %8.3f%8.3f%8.3f  1.00\n", n++, atoms[dom1][i].type, atoms[dom1][i].aa, '0', atoms[dom1][i].aa_num, x/PRECIS, y/PRECIS, z/PRECIS);
  }

  for(j=0; ((j<par->found) && (j<=50) && ((j<par->max_rep) || (par->max_rep==0))); j++){
    fprintf(par->out, "REMARK MOTIF # %i\n", j+1);
    fprintf(par->out, "REMARK CHAIN %c : MOTIF FROM %s\n", chain[j], domains[store[j].domain].id);
    fprintf(par->out, "REMARK %i ATOMS FITTED WITH E = %g (E(700) = %g; RMSD = %f)\n", store[j].n, store[j].P, store[j].Ps, store[j].rms);
    fprintf(par->out, "REMARK RES. EQUIVALENCES FOR MOTIF FROM %s | %s (CHAIN 0)\n", domains[dom1].id, domains[store[j].domain].id);
    /* print list of equiv. residues */
    for(i=0; i<store[j].n; i++){
      fprintf(par->out, "REMARK %s %i %c %c | %s %i %c %c\n", (*(points[dom1][(store[j].p1[i])].atoms[0])).aa, (*(points[dom1][(store[j].p1[i])].atoms[0])).aa_num,
	      (*(points[dom1][(store[j].p1[i])].atoms[0])).cid, (*(points[dom1][(store[j].p1[i])].atoms[0])).in,
	      (*(points[store[j].domain][(store[j].p2[i])].atoms[0])).aa, (*(points[store[j].domain][(store[j].p2[i])].atoms[0])).aa_num,
	      (*(points[store[j].domain][(store[j].p2[i])].atoms[0])).cid, (*(points[store[j].domain][(store[j].p2[i])].atoms[0])).in);
    }
    for(i=0; i<natoms[store[j].domain]; i++){
      x = ((float)(atoms[store[j].domain][i].co[0])/PRECIS);
      y = ((float)(atoms[store[j].domain][i].co[1])/PRECIS);
      z = ((float)(atoms[store[j].domain][i].co[2])/PRECIS);
      /*
	  x = ((float) (*(points[dom1][(store[j].p1[i])].atoms[k])).co[0])/PRECIS;
	  y = ((float) (*(points[dom1][(store[j].p1[i])].atoms[k])).co[1])/PRECIS;
	  z = ((float) (*(points[dom1][(store[j].p1[i])].atoms[k])).co[2])/PRECIS;
      */
      xn = store[j].R[0][0]*x + store[j].R[0][1]*y + store[j].R[0][2]*z + store[j].V[0];
      yn = store[j].R[1][0]*x + store[j].R[1][1]*y + store[j].R[1][2]*z + store[j].V[1];
      zn = store[j].R[2][0]*x + store[j].R[2][1]*y + store[j].R[2][2]*z + store[j].V[2];

      fprintf(par->out, "ATOM  %5d %4s%4s %1c%4d    %8.3f%8.3f%8.3f  1.00\n", n++, (atoms[store[j].domain][i]).type, (atoms[store[j].domain][i]).aa, chain[j], (atoms[store[j].domain][i]).aa_num, xn, yn, zn);
    }
  }
  return j;
}

int print_short(struct store_struc *store, int dom1, struct point **points, int *natoms, struct atom **atoms, struct domain_loc *domains, struct parameter *par){
  int i, j;

  for(j=0; j<par->found; j++){
    fprintf(par->out, "MOTIF # %i\n", j+1);
    fprintf(par->out, "%i ATOMS FITTED WITH E = %g (E(700) = %g; RMSD = %f)\n", store[j].n, store[j].P, store[j].Ps, store[j].rms);
    fprintf(par->out, "ATOM EQUIVALENCES FOR MOTIF FROM %s | %s \n", domains[dom1].id, domains[store[j].domain].id);
    for(i=0; i<store[j].n; i++){
      fprintf(par->out, "%s %i %c %c | %s %i %c %c\n", (*(points[dom1][(store[j].p1[i])].atoms[0])).aa, (*(points[dom1][(store[j].p1[i])].atoms[0])).aa_num,
	      (*(points[dom1][(store[j].p1[i])].atoms[0])).cid, (*(points[dom1][(store[j].p1[i])].atoms[0])).in,
	      (*(points[store[j].domain][(store[j].p2[i])].atoms[0])).aa, (*(points[store[j].domain][(store[j].p2[i])].atoms[0])).aa_num,
	      (*(points[store[j].domain][(store[j].p2[i])].atoms[0])).cid, (*(points[store[j].domain][(store[j].p2[i])].atoms[0])).in);
    }
  }
  return j;
}

int calc_rms(struct store_struc *store, struct point *points[2], struct parameter *par){
  int i, j, z, n, k, c, kgv;
  int *small;
  long int **atoms1;
  long int **atoms2;

  for(j=0; j<par->found; j++){
    small=(int *)malloc(store[j].n*sizeof(int));
    if(par->res_fit == 1){
      n=0;
      for(i=0; i<store[j].n; i++){
	if(points[0][(store[j].p1[i])].number <  points[1][(store[j].p2[i])].number){
	  small[i] = points[0][(store[j].p1[i])].number;
	}else{
	  small[i] =  points[1][(store[j].p2[i])].number;
	}
	n += small[i];
      }
    }else if(par->res_fit == 2){
      kgv = 1;
      for(i=0; i<store[j].n; i++){
	if(points[0][(store[j].p1[i])].number <  points[1][(store[j].p2[i])].number){
	  small[i] = points[0][(store[j].p1[i])].number;
	}else{
	  small[i] =  points[1][(store[j].p2[i])].number;
	}
	if(((int)(((float)kgv)/small[i])) != (((float)kgv)/small[i])){
	  kgv *= small[i];
	}
      }
      n = kgv*store[j].n;
    }else{
      n=store[j].n;
    }
    atoms1 = (long int **)malloc(n*sizeof(long int *));
    atoms2 = (long int **)malloc(n*sizeof(long int *));
    c=0;
    for(i=0; i<store[j].n; i++){
      k=0;
      do{
	z=0;
	do{
	  atoms2[c] = (long int *)malloc(3*sizeof(long int));
	  atoms2[c][0] = (*points[1][(store[j].p2[i])].atoms[k]).co[0];
	  atoms2[c][1] = (*points[1][(store[j].p2[i])].atoms[k]).co[1];
	  atoms2[c][2] = (*points[1][(store[j].p2[i])].atoms[k]).co[2];
	  atoms1[c] = (long int *)malloc(3*sizeof(long int));
	  atoms1[c][0] = (*points[0][(store[j].p1[i])].atoms[k]).co[0];
	  atoms1[c][1] = (*points[0][(store[j].p1[i])].atoms[k]).co[1];
	  atoms1[c][2] = (*points[0][(store[j].p1[i])].atoms[k]).co[2];
	  c++;
	  z++;
	}while((par->res_fit == 2) && (z<(kgv/small[i])));
	k++;
      }while((par->res_fit > 0) && (k<small[i]));
    }
    store[j].rms = RBR_imatfit(atoms1, atoms2, store[j].R, store[j].V, n, 1, PRECIS);
    for(i=0; i<n; i++){
      free(atoms1[i]);
      free(atoms2[i]);
    }
    free(small);
    free(atoms1);
    free(atoms2);
  }
}


int calc_permut(struct permut_struct *permute, struct store_struc *store, struct point *points[2], int res, int curr_motif, int **perm_list, int *perm_taken, int *current){
  int i;

  if(res == store[curr_motif].n){
    /*we filled in all !*/
    for(i=0; i<store[curr_motif].n; i++){
      perm_list[(*perm_taken)][i]=current[i];
    }
    (*perm_taken)++;
  }else{
    for(i=0; i<(permute[(points[0][((store[curr_motif]).p1[res])]).type].n); i++){
      current[res]=i;
      calc_permut(permute, store, points, (res+1), curr_motif, perm_list, perm_taken, current);
    }
  }
}



int calc_rms_perm(struct permut_struct *permute, struct store_struc *store, struct point *points[2], struct parameter *par){
  int i, j, z, n, k, red_k, c, p, kgv, npermut, perm_taken;
  int *small;
  int **perm_list;
  int *current; /*will hold the current permutation list during recursion*/
  long int **atoms1;
  long int **atoms2;
  /*these keep the results for diff. permutations*/
  float rms;
  float R[3][3];
  float V[3];


  /*alloc once, then only realloc during the loops*/
  if(((perm_list=(int **)malloc(sizeof(int *))) == NULL)
     || ((current=(int *)malloc(sizeof(int)))==NULL)
     || ((small=(int *)malloc(sizeof(int))) == NULL)
     || ((atoms1 = (long int **)malloc(sizeof(long int *))) == NULL)
     || ((atoms2 = (long int **)malloc(sizeof(long int *))) == NULL)){
    fprintf(stderr, "Memory overflow during permutations of atom fitting\n");
    exit(-1);
  }
  
  for(j=0; j<par->found; j++){

    /*determine smaller set and kgv for weighted fit*/
    if((small=(int *)realloc(small, store[j].n*sizeof(int)))==NULL){
      fprintf(stderr, "Memory overflow during permutations of atom fitting\n");
      exit(-1);
    }
    if(par->res_fit == 1){
      n=0;
      for(i=0; i<store[j].n; i++){
	if(points[0][(store[j].p1[i])].number <  points[1][(store[j].p2[i])].number){
	  small[i] = points[0][(store[j].p1[i])].number;
	}else{
	  small[i] =  points[1][(store[j].p2[i])].number;
	}
	n += small[i];
      }
    }else if(par->res_fit == 2){
      kgv = 1;
      for(i=0; i<store[j].n; i++){
	if(points[0][(store[j].p1[i])].number <  points[1][(store[j].p2[i])].number){
	  small[i] = points[0][(store[j].p1[i])].number;
	}else{
	  small[i] =  points[1][(store[j].p2[i])].number;
	}
	if(((int)(((float)kgv)/small[i])) != (((float)kgv)/small[i])){
	  kgv *= small[i];
	}
      }
      n = kgv*store[j].n;
    }else{
      n=store[j].n;
    }

    /*realloc / alloc memory for atoms*/
    if(((atoms1 = (long int **)realloc(atoms1, n*sizeof(long int *))) == NULL)
       || ((atoms2 = (long int **)realloc(atoms2, n*sizeof(long int *))) == NULL)){
      fprintf(stderr, "Memory overflow during permutations of atom fitting\n");
      exit(-1);
    }
    for(i=0; i<n; i++){
      if(((atoms1[i] = (long int *)malloc(3*sizeof(long int))) == NULL)
	 || ((atoms2[i] = (long int *)malloc(3*sizeof(long int))) == NULL)){
	fprintf(stderr, "Memory overflow during permutations of atom fitting\n");
	exit(-1);
      }
    }  

    /******************************************************/

    /*Now, create a list of all permutations that have to be tested
     *the first pattern stays, the second is moved through all possible
     *permutations, i.e. for each residue, all permut. !*/
    npermut = 1;
    for(i=0; i<store[j].n; i++){
      npermut*=permute[points[0][(store[j].p1[i])].type].n;
    }

    /*
    printf("Permutations %i\n", npermut);
    */

    /*reallocat perm_list[permutation][residue] and current*/
    if(((perm_list=(int **)realloc(perm_list, npermut*sizeof(int *))) == NULL)
       || ((current=(int *)realloc(current, (store[j].n)*sizeof(int)))==NULL)){
      fprintf(stderr, "Memory overflow during permutations of atom fitting\n");
      exit(-1);
    }
    for(i=0; i<npermut; i++){
      if((perm_list[i]=(int *)malloc(store[j].n*sizeof(int)))==NULL){
	fprintf(stderr, "Memory overflow during permutations of atom fitting\n");
	exit(-1);
      }
    }

    perm_taken=0;
    /*Now fill in perm_list recursively (i.e. put first res., then second, ...)*/
    calc_permut(permute, store, points, 0, j, perm_list, &perm_taken, current);

    /*
    printf("Permutations done %i\n", perm_taken);
    for(i=0; i<npermut; i++){
      for(k=0;k<store[j].n; k++){
	printf("%i, ", perm_list[i][k]);
      }
      printf("\n");
    }
    */

    /*set rms to large value*/
    store[j].rms = 100000;

    /*for all permutations*/
    for(p=0; p<npermut; p++){

      c=0;
      /*for all residues i*/
      for(i=0; i<store[j].n; i++){
	k=0;
	do{
	  z=0;
	  do{
	    /*now atom number k is redirected to the atom number given by the permutation*/
	    red_k = (permute[ (points[1][(store[j].p2[i])].type) ].p[ (perm_list[p][i]) ][k]);
	    /*                 |-------------------------------|      |---------------|   |
                                      Type of residue i                 state of res i    atom#
                                                                        in permut p
	    */

	    atoms2[c][0] = (*points[1][(store[j].p2[i])].atoms[red_k]).co[0];
	    atoms2[c][1] = (*points[1][(store[j].p2[i])].atoms[red_k]).co[1];
	    atoms2[c][2] = (*points[1][(store[j].p2[i])].atoms[red_k]).co[2];

	    atoms1[c][0] = (*points[0][(store[j].p1[i])].atoms[k]).co[0];
	    atoms1[c][1] = (*points[0][(store[j].p1[i])].atoms[k]).co[1];
	    atoms1[c][2] = (*points[0][(store[j].p1[i])].atoms[k]).co[2];
	    c++;
	    z++;
	  }while((par->res_fit == 2) && (z<(kgv/small[i])));
	  k++;
	}while((par->res_fit > 0) && (k<small[i]));
      }/*for all residues i*/
      /*      store[j].rms = RBR_imatfit(atoms1, atoms2, store[j].R, store[j].V, n, 1, PRECIS);*/

      rms = RBR_imatfit(atoms1, atoms2, R, V, n, 1, PRECIS);

      if(rms < store[j].rms){
	store[j].rms = rms;
	store[j].V[0] = V[0];
	store[j].V[1] = V[1];
	store[j].V[2] = V[2];

	store[j].R[0][0] = R[0][0];
	store[j].R[0][1] = R[0][1];
	store[j].R[0][2] = R[0][2];
	store[j].R[1][0] = R[1][0];
	store[j].R[1][1] = R[1][1];
	store[j].R[1][2] = R[1][2];
	store[j].R[2][0] = R[2][0];
	store[j].R[2][1] = R[2][1];
	store[j].R[2][2] = R[2][2];
      }
    }/*for all permutations p*/
    
    /*free all memory for permutations*/
    for(i=0; i<npermut; i++){
      free(perm_list[i]);
    }

    /*free memory for atoms*/
    for(i=0; i<n; i++){
      free(atoms1[i]);
      free(atoms2[i]);
    }
    
  }/*for all matches j*/

  free(perm_list);
  free(current);
  free(small);
  free(atoms1);
  free(atoms2);

}




int calc_P(struct store_struc *store, float *abund, int npoints[2], struct point *points[2], struct parameter *par){
  int i, j;
  int small;
  double A, B;
  double C, D;
  double Corr; /* correction for incl. matches (n choose m) */
  int N; /* number of points in smaller set */

  /* N is set to smaller set as we always split this into sev. "complete searches" */
  if(npoints[0] < npoints[1]){
    N = npoints[0];
  }else{
    N = npoints[1];
  }

  for(j=0; j<par->found; j++){
    if((par->pf == 0) || (par->pe == 0)){
      /*parameters not given --> calculate*/
      C = 1;
      D = 0;
      if(store[j].n>2){
	A = 2.678E9*powf(1.79E-3, ((double)store[j].n));
	B = 2.93*((double)store[j].n)-5.88;
	for(i=0; i<store[j].n; i++){
	  A *= abund[points[0][(store[j].p1[i])].type];
	  if(points[0][(store[j].p1[i])].number <  points[1][(store[j].p2[i])].number){
	    small = points[0][(store[j].p1[i])].number;
	  }else{
	    small =  points[1][(store[j].p2[i])].number;
	  }
	  if(small==2){
	    C *= 0.196;
	    D += 2;
	  }else if(small > 2){
	    C *= 0.094;
	    D += 3;
	  }
	}
      }else{
	A = 2.678E9*0.1277E-6;
	B = 0.97;
	for(i=0; i<store[j].n; i++){
	  A *= abund[points[0][(store[j].p1[i])].type];
	  if(points[0][(store[j].p1[i])].number <  points[1][(store[j].p2[i])].number){
	    small = points[0][(store[j].p1[i])].number;
	  }else{
	    small =  points[1][(store[j].p2[i])].number;
	  }
	  if(small==2){
	    C *= 0.196;
	    D += 2;
	  }else if(small > 2){
	    C *= 0.094;
	    D += 3;
	  }
	}
      }
      if(par->res_fit==0){
	C = 1;
	D = 0;
      }
    }else{
      A = par->pf;
      B = par->pe;
      C = 1;
      D = 0;
    }
    /* calculate Correction as N choose m=store[j].n 
     * split into 2 cases to prevent large numbers */
    Corr = (double)1;
    if(store[j].n < N/2){
      /* calc. as N*(N-1)*(N-2)*...*(N-m+1) / m! */
      for(i=N; i>(N-store[j].n); i--){
	Corr *= (double)i;
      }
      for(i=store[j].n; i>0; i--){
	Corr /= (double)i;
      }
    }else{
      /* calc. as N*(N-1)*(N-2)*...*(m+1) / (N-m)! */
      for(i=N; i>store[j].n; i--){
	Corr *= (double)i;
      }
      for(i=(N-store[j].n); i>0; i--){
	Corr /= (double)i;
      }
    }
    /* this is now changed to E value !! kept a copy of old....
    if((A*C*pow((double)store[j].rms, (B+D))) < (A*pow((double)store[j].rms, B))){
      store[j].P = 1-exp(-(Corr*A*C*pow((double)store[j].rms, (B+D))));
    }else{
      store[j].P = 1-exp(-(Corr*A*pow((double)store[j].rms, B)));
    }
    */
    /* uncorrected Ps */
    if((A*C*pow((double)store[j].rms, (B+D))) < (A*pow((double)store[j].rms, B))){
      store[j].Ps = A*C*pow((double)store[j].rms, (B+D));
    }else{
      store[j].Ps = A*pow((double)store[j].rms, B);
    }
    /*corrected P*/
    store[j].P = Corr*store[j].Ps;
    /*    printf("%g | %i | %i | %g | %g\n", Corr, N, store[j].n, store[j].P, store[j].Ps); */
  }
}

int search(struct point **points, struct store_struc **store, int **motif, char **match, char **radius, int npoints[2], int i_pos, int *j_array, long int ***dist, int act_level, struct parameter *par){
  int i, j, k;
  int found;
  int new_i_pos;
  int swap_help;
  int until_i;
  float lw, lv, a0, a1, a2, b0, b1, b2, v0, v1, v2, w0, w1, w2;


  act_level++;

  until_i = npoints[0]+act_level-par->min_atoms;
  if(act_level >= par->min_atoms) until_i = npoints[0]-1;
  for(i=i_pos; i<=until_i; i++){
      /* now checking whether new atom[i] is within the range of max_size. max_size should be the diameter
       * of the motif but is used as the radius. In the best case (atom[i] at the edge of a motif) it fits,
       * in the worst case, the found motif can be twice as big. Testing against the whole "base" does not 
       * eliminate the problem but reduces it */
    found = 1;
    k = act_level;
    do {
      k--;
      if(!radius[i][(motif[0][k])]) {
	found = 0;
	break;
      }
    } while (k>0);
    if(found == 0){
      continue;
    }

    for(j=act_level; j<npoints[1]; j++){
      if(!match[i][j_array[j]]){
	/*if((points[0][i].type) != (points[1][j_array[j]].type)){*/
	continue; /* if types don't match goto next j */
      }
      found = act_level+1;
      k = act_level;
      do {
	k--;
	if(((dist[0][i][(motif[0][k])] - dist[1][j_array[j]][(motif[1][k])]) > par->max_tolerance) ||
	   ((dist[0][i][(motif[0][k])] - dist[1][j_array[j]][(motif[1][k])]) < (-par->max_tolerance))){
	  found = 0;
	  break;
	}
      } while (k>0);

      if(found == 0){
	continue;
      }
      
      if(act_level == 3){
	/* the 4th atom was found. do chirality check here */
	/* cross product for 0->1 and 0->2 in set 1 */
	a0 = (float)((points[0][motif[0][1]].co[0])-(points[0][motif[0][0]].co[0]));
	a1 = (float)((points[0][motif[0][1]].co[1])-(points[0][motif[0][0]].co[1]));
	a2 = (float)((points[0][motif[0][1]].co[2])-(points[0][motif[0][0]].co[2]));
	b0 = (float)((points[0][motif[0][2]].co[0])-(points[0][motif[0][0]].co[0]));
	b1 = (float)((points[0][motif[0][2]].co[1])-(points[0][motif[0][0]].co[1]));
	b2 = (float)((points[0][motif[0][2]].co[2])-(points[0][motif[0][0]].co[2]));

	v0 = a1*b2-a2*b1;
	v1 = a2*b0-a0*b2;
	v2 = a0*b1-a1*b0;
	lv = sqrtf((v0 * v0) + (v1 * v1) + (v2 *v2));
      
	/* cross product for 0->1 and 0->2 in set 2 */
      	a0 = (float)((points[1][motif[1][1]].co[0])-(points[1][motif[1][0]].co[0]));
	a1 = (float)((points[1][motif[1][1]].co[1])-(points[1][motif[1][0]].co[1]));
	a2 = (float)((points[1][motif[1][1]].co[2])-(points[1][motif[1][0]].co[2]));
	b0 = (float)((points[1][motif[1][2]].co[0])-(points[1][motif[1][0]].co[0]));
	b1 = (float)((points[1][motif[1][2]].co[1])-(points[1][motif[1][0]].co[1]));
	b2 = (float)((points[1][motif[1][2]].co[2])-(points[1][motif[1][0]].co[2]));

	w0 = a1*b2-a2*b1;
	w1 = a2*b0-a0*b2;
	w2 = a0*b1-a1*b0;
	lw = sqrtf(w0*w0 + w1*w1 + w2*w2);
      
	/* product of the two scalar products of the two cross products is the squared distance from
	 * the plane (correct orientation) or the negative squared distance (wrong orientation)
	 * -> allowance is max_tolerance */
      	if(((v0/lv*((points[0][i].co[0])-(points[0][motif[0][0]].co[0])) + v1/lv*((points[0][i].co[1])-(points[0][motif[0][0]].co[1])) + v2/lv*((points[0][i].co[2])-(points[0][motif[0][0]].co[2])))*(w0/lw*((points[1][j_array[j]].co[0])-(points[1][motif[1][0]].co[0])) + w1/lw*((points[1][j_array[j]].co[1])-(points[1][motif[1][0]].co[1])) + w2/lw*((points[1][j_array[j]].co[2])-(points[1][motif[1][0]].co[2])))) < -((float)par->max_tolerance * (float)par->max_tolerance)){
	  continue;
	}
      }
      
      /* adding the (act_level+1) found atoms (represented by their numbers = indexes) to the motifs */
      motif[0][act_level] = i;
      motif[1][act_level] = j_array[j];

      /* Check whether motif is "good" enough to save and save eventually. */
      if(found >= par->min_atoms){
        if((((*store) = (struct store_struc *)realloc((*store), (par->found+1)*sizeof(struct store_struc))) == NULL) ||
          (((*store)[par->found].p1 = (int *)malloc(found*sizeof(int))) == NULL) ||
	   (((*store)[par->found].p2 = (int *)malloc(found*sizeof(int))) == NULL)){
	  fprintf(stderr, "Error in search(): Memory overflow while storing motif\n");
	  exit(-1);
	}

	(*store)[par->found].n = found;
	k = found;
	do {
	  k--;
	  (*store)[par->found].p1[k] = motif[0][k];
	  (*store)[par->found].p2[k] = motif[1][k];
	} while(k>0);

	(par->found)++;
      }

      /* initiate recursive call if max_depth is not reached */
      if(found < par->max_depth){
	/* setting start point for subsequent rounds to i and j respectively */
	new_i_pos = i+1;

	/* swapping contents of j_array: the atom that fits into the current motif is at j. now reorder j_array
	 * that the atom (-number) is at act_level. all atoms that have been jumped over, are now "in front" of the
	 * search-loop and can thus be found. */

	swap_help = j_array[j];
	j_array[j] = j_array[act_level];
	j_array[act_level] = swap_help;

	/* recursive call with the updated motif, actual positions pos are now i and j, act_level is already updated */
	search(points, store, motif, match, radius, npoints, new_i_pos, j_array, dist, act_level, par);
	  
	/* after recursive call returned, swap j_array back to its original form */
	j_array[act_level] = j_array[j];
	j_array[j] = swap_help;
      } /* END of recursive call */

    } /* END of for(j...): all atoms in set 2 have been tried. Now try other atom in set 1 (inc i) */
  } /* END of for(i...): all possibilities in this level have been tried. return to upper level */
  return 0;
}


int lin_search(struct point **points, struct store_struc **store, int **motif, char **match, char **radius, int npoints[2], int i_pos, long int ***dist, int act_level, struct parameter *par){
  int i, j, k;
  int found;
  int new_i_pos;
  int until_i;
  float lw, lv, a0, a1, a2, b0, b1, b2, v0, v1, v2, w0, w1, w2;


  act_level++;

  until_i = npoints[0]+act_level-par->min_atoms;
  if(act_level >= par->min_atoms) until_i = npoints[0]-1;
  for(i=i_pos; i<=until_i; i++){
      /* now checking whether new atom[i] is within the range of max_size. max_size should be the diameter
       * of the motif but is used as the radius. In the best case (atom[i] at the edge of a motif) it fits,
       * in the worst case, the found motif can be twice as big. Testing against the whole "base" does not 
       * eliminate the problem but reduces it */
    found = 1;
    k = act_level;
    do {
      k--;
      if(!radius[i][(motif[0][k])]) {
	found = 0;
	break;
      }
    } while (k>0);
    if(found == 0){
      continue;
    }

    for(j=act_level; j<npoints[1]; j++){
      if(!match[i][j]){
	/*if((points[0][i].type) != (points[1][j].type)){*/
	continue; /* if types don't match goto next j */
      }
      found = act_level+1;
      k = act_level;
      do {
	k--;
	if(((dist[0][i][(motif[0][k])] - dist[1][j][(motif[1][k])]) > par->max_tolerance) ||
	   ((dist[0][i][(motif[0][k])] - dist[1][j][(motif[1][k])]) < (-par->max_tolerance))){
	  found = 0;
	  break;
	}
      } while (k>0);

      if(found == 0){
	continue;
      }
      
      if(act_level == 3){
	/* the 4th atom was found. do chirality check here */
	/* cross product for 0->1 and 0->2 in set 1 */
	a0 = (float)((points[0][motif[0][1]].co[0])-(points[0][motif[0][0]].co[0]));
	a1 = (float)((points[0][motif[0][1]].co[1])-(points[0][motif[0][0]].co[1]));
	a2 = (float)((points[0][motif[0][1]].co[2])-(points[0][motif[0][0]].co[2]));
	b0 = (float)((points[0][motif[0][2]].co[0])-(points[0][motif[0][0]].co[0]));
	b1 = (float)((points[0][motif[0][2]].co[1])-(points[0][motif[0][0]].co[1]));
	b2 = (float)((points[0][motif[0][2]].co[2])-(points[0][motif[0][0]].co[2]));

	v0 = a1*b2-a2*b1;
	v1 = a2*b0-a0*b2;
	v2 = a0*b1-a1*b0;
	lv = sqrtf((v0 * v0) + (v1 * v1) + (v2 *v2));
      
	/* cross product for 0->1 and 0->2 in set 2 */
      	a0 = (float)((points[1][motif[1][1]].co[0])-(points[1][motif[1][0]].co[0]));
	a1 = (float)((points[1][motif[1][1]].co[1])-(points[1][motif[1][0]].co[1]));
	a2 = (float)((points[1][motif[1][1]].co[2])-(points[1][motif[1][0]].co[2]));
	b0 = (float)((points[1][motif[1][2]].co[0])-(points[1][motif[1][0]].co[0]));
	b1 = (float)((points[1][motif[1][2]].co[1])-(points[1][motif[1][0]].co[1]));
	b2 = (float)((points[1][motif[1][2]].co[2])-(points[1][motif[1][0]].co[2]));

	w0 = a1*b2-a2*b1;
	w1 = a2*b0-a0*b2;
	w2 = a0*b1-a1*b0;
	lw = sqrtf(w0*w0 + w1*w1 + w2*w2);
      
	/* product of the two scalar products of the two cross products is the squared distance from
	 * the plane (correct orientation) or the negative squared distance (wrong orientation)
	 * -> allowance is max_tolerance */
      	if(((v0/lv*((points[0][i].co[0])-(points[0][motif[0][0]].co[0])) + v1/lv*((points[0][i].co[1])-(points[0][motif[0][0]].co[1])) + v2/lv*((points[0][i].co[2])-(points[0][motif[0][0]].co[2])))*(w0/lw*((points[1][j].co[0])-(points[1][motif[1][0]].co[0])) + w1/lw*((points[1][j].co[1])-(points[1][motif[1][0]].co[1])) + w2/lw*((points[1][j].co[2])-(points[1][motif[1][0]].co[2])))) < -((float)par->max_tolerance * (float)par->max_tolerance)){
	  continue;
	}
      }
      
      /* adding the (act_level+1) found atoms (represented by their numbers = indexes) to the motifs */
      motif[0][act_level] = i;
      motif[1][act_level] = j;

      /* Check whether motif is "good" enough to save and save eventually. */
      if(found >= par->min_atoms){
        if((((*store) = (struct store_struc *)realloc((*store), (par->found+1)*sizeof(struct store_struc))) == NULL) ||
          (((*store)[par->found].p1 = (int *)malloc(found*sizeof(int))) == NULL) ||
	   (((*store)[par->found].p2 = (int *)malloc(found*sizeof(int))) == NULL)){
	  fprintf(stderr, "Error in search(): Memory overflow while storing motif\n");
	  exit(-1);
	}

	(*store)[par->found].n = found;
	k = found;
	do {
	  k--;
	  (*store)[par->found].p1[k] = motif[0][k];
	  (*store)[par->found].p2[k] = motif[1][k];
	} while(k>0);

	(par->found)++;
      }

      /* initiate recursive call if max_depth is not reached */
      if(found < par->max_depth){
	/* setting start point for subsequent rounds to i and j respectively */
	new_i_pos = i+1;

	/* recursive call with the updated motif, actual positions pos are now i and j, act_level is already updated */
	lin_search(points, store, motif, match, radius, npoints, new_i_pos, dist, act_level, par);

      } /* END of recursive call */

    } /* END of for(j...): all atoms in set 2 have been tried. Now try other atom in set 1 (inc i) */
  } /* END of for(i...): all possibilities in this level have been tried. return to upper level */
  return 0;
}


int start_search(struct point **points, int *npoints, long int ***dist, char **radius, char **match, int **motif, int *j_array, struct store_struc **store, struct parameter *par){
  int i, j, act_level;
  int i_pos;
  int swap_help;

  act_level = 0;
  par->found = 0;

  /* initializes j_array as 1:1 pointer 
   * if swapping/swapping back works, this could be done ONCE in main */
  i = npoints[1];
  do {
    i--;
    j_array[i] = i;
  } while(i>0);

  i = npoints[0];
  do {
    i--;
    j = npoints[1];
    do{
      j--;
      match[i][j] = ((points[0][i].type) == (points[1][j].type));
    } while (j>0);
  } while (i>0);

  for(i=0; i<=(npoints[0]-par->min_atoms); i++){ /* for the last few atoms <(n-(minatoms+1)): no motif of sufficient size can be found*/
    if(par->verbose){
      if((npoints[0]-par->min_atoms) != 0){
	printf("%i %% : %i motif(s) found\n", 100*i/(npoints[0]-par->min_atoms), par->found);
      }
    }
    if((par->found >= par->max_hits) && (par->max_hits > 0)){
      printf("Max. number of hits was reached\n");
      return 0;
    }
    for(j=0; j<npoints[1]; j++){
      if(match[i][j]){
	motif[0][0] = i;
	motif[1][0] = j;
	i_pos = i+1;

	swap_help = j_array[j];
	j_array[j] = j_array[act_level];
	j_array[act_level] = swap_help;

	search(points, store, motif, match, radius, npoints, i_pos, j_array, dist, act_level, par);

	j_array[act_level] = j_array[j];
	j_array[j] = swap_help;

      }
    }
  }

  return par->found;
}


int start_lin_search(struct point **points, int *npoints, long int ***dist, char **radius, char **match, int **motif, struct store_struc **store, struct parameter *par){
  int i, j, act_level;
  int i_pos;
  int swap_help;

  act_level = 0;
  par->found = 0;

  i = npoints[0];
  do {
    i--;
    j = npoints[1];
    do{
      j--;
      match[i][j] = ((points[0][i].type) == (points[1][j].type));
    } while (j>0);
  } while (i>0);

  for(i=0; i<=(npoints[0]-par->min_atoms); i++){ /* for the last few atoms <(n-(minatoms+1)): no motif of sufficient size can be found*/
    if(par->verbose){
      if((npoints[0]-par->min_atoms) != 0){
	printf("%i %% : %i motif(s) found\n", 100*i/(npoints[0]-par->min_atoms), par->found);
      }
    }
    if((par->found >= par->max_hits) && (par->max_hits > 0)){
      printf("Max. number of hits was reached\n");
      return 0;
    }
    for(j=0; j<npoints[1]; j++){
      if(match[i][j]){
	motif[0][0] = i;
	motif[1][0] = j;
	i_pos = i+1;

	lin_search(points, store, motif, match, radius, npoints, i_pos, dist, act_level, par);

      }
    }
  }

  return par->found;
}




int prepare_domain(struct point *points, int npoints, long int **dist, char **radius, long int max_size){
  int i, j;

  /* calculating pairwise distance and the binary size-allowance :
   * i runs from npoints-1 to 0; j runs from i to 0 (upper diagonal)
   * -> lower diagonal is filled by dist[j][i] = dist[i][j] */
  i = npoints;
  do {
    i--;
    j = i;
    do{
      dist[i][j] = (long int) (sqrtf(powf(((points[i]).co[0]-(points[j]).co[0]), 2) + powf(((points[i]).co[1]-(points[j]).co[1]), 2) + powf(((points[i]).co[2]-(points[j]).co[2]), 2)) + 0.5);
      radius[i][j] = (((dist[j][i] = dist[i][j]) <= max_size) || (max_size == 0));
      radius[j][i] = radius[i][j];
      j--;
    } while (j>=0);
  } while (i>0);

  return 1;
}



/* This function changes a character to lowercase. */
char RBR_utol(char c) {

	if (c>='A' && c<='Z')
	   return c - 'A' + 'a';
	else return c;
}


/* This function changes a character to uppercase. */
char RBR_ltou(c)

char c;
{
	if (c>='a' && c<='z')
	   return c - 'a' + 'A';
	else return c;
}

int testfile(char *file) {

	FILE *f;
	int flag;

	if((f=fopen(file,"r"))==NULL) {
		flag = 0;
	} else  {
		flag = 1;
		fclose(f);
	}
	return flag;
}


FILE *openfile(char *filename, char *type) {

	FILE *handle;
	char command[1000];

/*	printf("Attempting to open file %s\n",filename); */
	if(strcmp(type,"r")==0) { /* A read file */
		if(strcmp(&filename[strlen(filename)-2],".Z")==0) { /* UNIX compression */
/*		     sprintf(command,"zcat %s 2> /dev/null",filename);  */
		     sprintf(command,"zcat %s",filename);  
		     handle=popen(command,"r");
		} else if(strcmp(&filename[strlen(filename)-3],".gz")==0) { /* gzipped  */
/*		     sprintf(command,"gunzip -c %s 2> /dev/null",filename);  */
		     sprintf(command,"gunzip -c %s",filename); 
		     handle=popen(command,"r");
		} else { /* presume no compression */
		    handle=fopen(filename,"r");
		}
	} else {
	  	    handle=fopen(filename,"w");
	}
/*	printf("Opened file %s\n",filename); */
	return handle;
}

	
void closefile(FILE *handle, char *filename) {


	char b[1000];
	if((strcmp(&filename[strlen(filename)-2],".Z")==0) || 
           (strcmp(&filename[strlen(filename)-3],".gz")==0)) { 

	    /* This fgets prevents "Broken pipe" messages from appearing when running under Solaris,
	     *  and silly newlines to stderr when running under OSF - something about a wrong implementation
	     *  of pclose()? */
	    while(fgets(b,999,handle)!=NULL);
	    pclose(handle);
	} else { 
	    fclose(handle);
	}
/*	printf("Closed file %s\n",filename); */
}



 
/* Converts three letter amino acid code to one letter
 *  amino acid code
 * Note: CYS refers to cystine, CYH refers to cysteine */
 
char RBR_a3to1(char *a3){
        int i;
        char new;
 
        new='X';
        for(i=0;i<strlen(RBR_a3to1_acids1); i++) {
           if (strncmp(&RBR_a3to1_acids3[i*4],a3,3) == 0)
              new=(char)RBR_a3to1_acids1[i];
        }   /* end of for  */
 
        return(new);
} /* end of function */ 

void a1to3(char aa, char new[4]){
  int i;

  strcpy(new,"UNK");
  for(i=0; i<strlen(RBR_a3to1_acids1); i++){
    if(aa == RBR_a3to1_acids1[i]){
      strncpy(new, &(RBR_a3to1_acids3[i*4]), 3);
      new[3] = '\0';
      return;
    }
  }
}

int max_acc(char aa){
  int i;
  int acc;

  for(i=0; i<strlen(ACC_NAM); i++){
    if(aa == ACC_NAM[i]){
      sscanf(&(MAX_ACC[4*i]), "%d", &acc);
      return acc;     
    }
  }
  return -1;
}

/*
#define ABUND "8.19 1.64 5.79 5.99 3.98 7.96 2.33 5.42 6.04 8.39 2.03 4.66 4.59 3.71 4.62 6.33 6.15 7.00 1.54 3.65";
#define ABUND_NAM "ACDEFGHIKLMNPQRSTVWY";
*/

float abundance(char aa){
  int i;
  float ab;

  for(i=0; i<strlen(ABUND_NAM); i++){
    if(aa == ABUND_NAM[i]){
      sscanf(&(ABUND[5*i]), "%f", &ab);
      return ab;     
    }
  }
  return 1; /*if no aa or *, set abundance to 1 (neutral)*/
}


   
/* Counts the number of domain descriptors in a file ignoring comments */
int RBR_count_domain(FILE *IN){
  int N,col,comment;
  char c;

  N=0;
  comment=0;
  col=0;
  while((c=getc(IN))!=(char)EOF){
    if((col==0) && (c=='#' || c=='%')){
      comment=1;
    }
    if(col!=0 && !comment && c=='{'){
      N++;
    }
    col++;
    if(c=='\n') {
      col=0;
      comment=0;
    }
  }
  return N;
}


/* getfile: 
 *  opens a file containing a list of appropriate directories, suffixes and
 *  prefixes.  It then makes combinates of these with the supplied code
 *  to see if a file exists 
 *
 * Allocates memory for a string corresponding to the filename read in,
 *  then returns this string or '\0' if none is found 
 *
 * Modification March 1999 - now tries to expand into a "distr" style hierarchy */
char *RBR_getfile(char *code_safe, char *dirfile, int code_length, FILE *OUTPUT) {

  char *temp;
  char *dir;
  char *buff;
  char prefix[10],suffix[10];
  char *code;
  char *CODE;
  char *two_letter;
  char a;

  int i,j;
  int found;

  FILE *DIR;

  dir=(char*)malloc(1000*sizeof(char));
  buff=(char*)malloc(1000*sizeof(char));
  temp=(char*)malloc(1000*sizeof(char));
  CODE=(char*)malloc((strlen(code_safe)+1)*sizeof(char));
  code=(char*)malloc((strlen(code_safe)+1)*sizeof(char));
  two_letter = (char*)malloc(3*sizeof(char));

  /* copy code_safe into code to prevent the changing case from
   * carrying on through the program */

  a = code[code_length];
  code[code_length]='\0'; /* truncates the code if necessary */
  for(i=0; i<strlen(code_safe); ++i) code[i]=RBR_utol(code_safe[i]);
  for(i=0; i<strlen(code_safe); ++i) CODE[i]=RBR_ltou(code_safe[i]); /* upper case code */
  code[code_length]='\0';
  CODE[code_length]='\0';
  two_letter[0] = code[1]; two_letter[1] = code[2]; two_letter[2] = '\0'; /* For PDB distr style */

  if((DIR=fopen(dirfile,"r"))==NULL) {
    fprintf(stderr,"Error: no directories file found\n");
    temp[0]='\0';
    return temp;
  }
  found=0;
  while((fgets(buff,MAXLINE,DIR))!=NULL){

    sscanf(buff,"%s %s %s",dir,&prefix[0],&suffix[0]);
    if(prefix[0] == '_') { prefix[0] = '\0'; }
    if(suffix[0] == '_') { suffix[0] = '\0'; }

    /* Lower case, all files one directory */
    sprintf(temp,"%s/%s%s%s",dir,prefix,code,suffix);
    if(testfile(temp)) { found=1; break; }

    /* Lower case, files in 'distr' type directories */
    sprintf(temp,"%s/%s/%s%s%s",dir,two_letter,prefix,code,suffix);
    if(testfile(temp)) { found=1; break; }

    /* Upper case, all files one directory */
    sprintf(temp,"%s/%s%s%s",dir,prefix,CODE,suffix);
    if(testfile(temp)) { found=1; break; }

    /* Upper case, files in 'distr' type directories */
    sprintf(temp,"%s/%s/%s%s%s",dir,two_letter,prefix,CODE,suffix);
    if(testfile(temp)) { found=1; break; }

  }

  if(found==0) {
    temp[0]='\0';
  }
  fclose(DIR);
  free(dir);
  free(buff);
  free(code);
  free(CODE);
  free(two_letter);

  return temp;
}


int RBR_domdefine(struct domain_loc *domain, int *gottrans, char *env, int DSSP, FILE *INPUT, FILE *OUTPUT){
/* reads in the next domain descriptor from a supplied input file 
 * returns 0 if all is well, -1 if an error occurs, 1 if EOF occurs */

  int i,j,k;
  int nobjects;
  int comment;
  int pt;

  char c;
  char *descriptor;
  char *buff;
  char *temp;
  char *dirfile;
	
  FILE *TEST;

  buff=(char*)malloc(2000*sizeof(char));
  dirfile=(char*)malloc(500*sizeof(char));
  descriptor=(char*)malloc(2000*sizeof(char));
	
  if(DSSP) {
    sprintf(dirfile,"%s/dssp.directories",env);
  } else {
    sprintf(dirfile,"%s/pdb.directories",env);
  }

  /* Now then, the best way to do this is to skip the comments, and
   *  then just read in a buff starting after the last newline, and
   *  ending at the end brace */
	
  while((c=getc(INPUT))!=(char)EOF){
    if(c!='%' && c!='#') { /* not a comment */
      /* lets just read in the whole thing */
      i=0;
      buff[i++]=c;
      while((c=getc(INPUT))!=(char)EOF && c!='}'){
	buff[i++]=c;
      }
      if(c==(char)EOF) break;
      buff[i]=c;
      buff[i+1]='\0';
      break;
    } else { /* skip over comment */
      while((c=getc(INPUT))!=(char)EOF && c!='\n');
      if(c==(char)EOF) break;
    }
  }
  if(c==(char)EOF) { /* End of file, no domain, exit */
    free(buff); free(dirfile); free(descriptor);
    return 1;
  }
  /* if we have a domain, and are not at the end of the file, 
   *  we should read until the end of the last line, to 
   *  leave the file pointer ready for next time */
  while((c=getc(INPUT))!=(char)EOF && c!='\n');

  /* First read the file name */
  sscanf(buff,"%s",&domain[0].filename[0]); /* read the filename */
  pt=0;
  if((pt=skiptononspace(buff,pt))==-1) getdomain_error(buff);
  sscanf(&buff[pt],"%s",&domain[0].id[0]);	/* read the identifier */

  /* check to see whether the file exists, otherwise, look for a file
   *  with a similar ID */
  if((TEST=fopen(domain[0].filename,"r"))==NULL){
    /* look for the file */
    temp=RBR_getfile(domain[0].id,dirfile,4,OUTPUT); /* assume the first four characters are the id */
    if(temp[0]=='\0') {
      fprintf(OUTPUT,"file for %s not found, nor was any corrsponding file\n",domain[0].id);
      fprintf(OUTPUT,"  found in %s\n",dirfile);
      free(buff); free(dirfile); free(descriptor);
      return -1;
    } else {
      strcpy(&domain[0].filename[0],temp);
    }
    free(temp);
  } else {
    fclose(TEST);
  }
	   
  if((pt=skiptononspace(buff,pt))==-1) getdomain_error(buff);
  /* copy the bit between the braces into the string called descriptor */
  i=0;
  while(buff[pt]!='{' && buff[pt]!='\n' && buff[pt]!='\0') pt++;
  if(buff[pt]=='\n' || buff[pt]=='\0') getdomain_error(buff);
  pt++;
  if(buff[pt]=='\n' || buff[pt]=='\0') getdomain_error(buff);
  j=0; while(buff[pt]!='}' && buff[pt]!='\0'){
    if(buff[pt]=='\0') getdomain_error(buff);
    descriptor[j]=buff[pt];
    pt++;
    j++;
  }
  descriptor[j]='\0';

  /* allocation of memory, initially */
  domain[0].reverse=(int*)malloc(sizeof(int));
  domain[0].type=(int*)malloc(sizeof(int));
  domain[0].start=(struct brookn*)malloc(sizeof(struct brookn));
  domain[0].end=(struct brookn*)malloc(sizeof(struct brookn));
  domain[0].V=(float*)malloc(3*sizeof(float));
  domain[0].v=(float*)malloc(3*sizeof(float));
  domain[0].R=(float**)malloc(3*sizeof(float*));
  domain[0].r=(float**)malloc(3*sizeof(float*));
  for(i=0; i<3; ++i){
    domain[0].R[i]=(float*)malloc(3*sizeof(float));
    domain[0].r[i]=(float*)malloc(3*sizeof(float));
    for(j=0; j<3; ++j)
      if(i==j) domain[0].R[i][j]=domain[0].r[i][j]=1.0;
      else domain[0].R[i][j]=domain[0].r[i][j]=0.0;
      domain[0].V[i]=domain[0].v[i]=0.0;
  }

  nobjects=0;
  for(i=0; i<strlen(descriptor); ++i) descriptor[i]=RBR_ltou(descriptor[i]);
  pt=0;
  if(strlen(descriptor)==0) getdomain_error(buff);
  while(descriptor[pt]==' ' && descriptor[pt]!='\0') pt++;
  if(descriptor[pt]=='\0' || descriptor[pt]=='}') getdomain_error(buff);
  while(descriptor[pt]!='\0' && descriptor[pt]!='\n'  && pt!=-1) { /* read until end of string */
    if(strncmp(&descriptor[pt],"REVERSE",7)==0) { /* coordinates are to be reversed */
      domain[0].reverse[nobjects]=1;
      pt=skiptononspace(descriptor,pt);
    } else {
      domain[0].reverse[nobjects]=0;
      /* don't skip over the text if the word "REVERSE" isn't there */
    }
    if(strncmp(&descriptor[pt],"ALL",3)==0) {  /* want all the coordinates in the file */
      domain[0].type[nobjects]=1;
      domain[0].start[nobjects].cid=domain[0].start[nobjects].in=
	domain[0].end[nobjects].cid=domain[0].end[nobjects].in='?';
      domain[0].start[nobjects].n=domain[0].end[nobjects].n=0;
      pt=skiptononspace(descriptor,pt);
      nobjects++;
    } else if(strncmp(&descriptor[pt],"CHAIN",5)==0) { /* want specific chain only */
      domain[0].type[nobjects]=2;
      if((pt=skiptononspace(descriptor,pt))==-1) getdomain_error(buff); /* no chain given */
      domain[0].start[nobjects].cid=domain[0].end[nobjects].cid=descriptor[pt];
      domain[0].start[nobjects].in=domain[0].end[nobjects].in='?';
      domain[0].start[nobjects].n=domain[0].end[nobjects].n=0;
      pt=skiptononspace(descriptor,pt);
      nobjects++;
    } else { /* assume that otherwise a specific start and end will be provided */
      domain[0].type[nobjects]=3;
      /* cid 1 */
      if(descriptor[pt]=='_') domain[0].start[nobjects].cid=' ';
      else domain[0].start[nobjects].cid=descriptor[pt];
      if((pt=skiptononspace(descriptor,pt))==-1) getdomain_error(buff);
      /* n 1 */
      sscanf(&descriptor[pt],"%d",&domain[0].start[nobjects].n);
      if((pt=skiptononspace(descriptor,pt))==-1) getdomain_error(buff);
      /* ins 1 */
      if(descriptor[pt]=='_') domain[0].start[nobjects].in=' ';
      else domain[0].start[nobjects].in=descriptor[pt];
      if((pt=skiptononspace(descriptor,pt))==-1) getdomain_error(buff);
      /* skipping over 'to' */
      if(strncmp(&descriptor[pt],"TO",2)!=0) getdomain_error(buff);
      if((pt=skiptononspace(descriptor,pt))==-1) getdomain_error(buff);
      /* cid 2 */
      if(descriptor[pt]=='_') domain[0].end[nobjects].cid=' ';
      else domain[0].end[nobjects].cid=descriptor[pt];
      if((pt=skiptononspace(descriptor,pt))==-1) getdomain_error(buff);
      /* n 2 */
      sscanf(&descriptor[pt],"%d",&domain[0].end[nobjects].n);
      if((pt=skiptononspace(descriptor,pt))==-1) getdomain_error(buff);
      /* ins 2 */
      if(descriptor[pt]=='_') domain[0].end[nobjects].in=' ';
      else domain[0].end[nobjects].in=descriptor[pt];
      pt=skiptononspace(descriptor,pt);
      nobjects++;
    }
    if(pt!=-1 && descriptor[pt]=='\n') break;
    /* reallocing if necessary */
    if(pt!=-1 && strlen(&descriptor[pt])>0 && descriptor[pt]!='\0' && descriptor[pt]!='\n'){
      /*		  printf("Allocating memory for a new object !\n"); */
      domain[0].reverse=(int*)realloc(domain[0].reverse,(nobjects+1)*sizeof(int));
      domain[0].type=(int*)realloc(domain[0].type,(nobjects+1)*sizeof(int));
      domain[0].start=(struct brookn*)realloc(domain[0].start,(nobjects+1)*sizeof(struct brookn));
      domain[0].end=(struct brookn*)realloc(domain[0].end,(nobjects+1)*sizeof(struct brookn));
    }
    /* now either stop, or move onto the next descriptor */
  }
	 
  /* check to see whether there is a transformation */
  if(pt!=-1) { /* there is */
    while(descriptor[pt]!='\n' && descriptor[pt]!='\0') pt++;
    if(descriptor[pt]=='\0') getdomain_error(buff);
    (*gottrans)=1;
    if(sscanf(&descriptor[pt],"%f%f%f%f%f%f%f%f%f%f%f%f",
	      &domain[0].R[0][0],&domain[0].R[0][1],&domain[0].R[0][2],&domain[0].V[0],
	      &domain[0].R[1][0],&domain[0].R[1][1],&domain[0].R[1][2],&domain[0].V[1],
	      &domain[0].R[2][0],&domain[0].R[2][1],&domain[0].R[2][2],&domain[0].V[2])==(char)EOF)
      getdomain_error(buff);
  }
  domain[0].nobj=nobjects;
  free(dirfile);
  free(descriptor);
  free(buff);
  return 0;
}


int RBR_getdomain(FILE *IN, struct domain_loc *domains, int *ndomain,
                  int maxdomain, int *gottrans, char *env, int DSSP, FILE *OUTPUT){
  int i,j;
  int comment;
  int count,end,nobjects;

  char c;
  char *buff;


  count=0;
  end=0;
  (*gottrans)=0;
  while(!end){
    end=RBR_domdefine(&domains[count],&i,env,DSSP,IN,OUTPUT);
    if(i==1) (*gottrans)=1;
    if(end==-1){
      fprintf(stderr,"Error in domain specification file\n");
      return -1;
    }
    count+=(!end);
    if(count>maxdomain && end!=1){
      fprintf(stderr,"Error in getdomain(): maximum domain limit exceeded\n");
      return -1;
    }
    if(count==maxdomain) end=1;
  } /* end while(!end)... */
  (*ndomain)=count;

  /* check for duplication */
  for(i=0; i<(*ndomain); ++i)
    for(j=i+1; j<(*ndomain); ++j)
      if(strcmp(domains[i].id,domains[j].id)==0){
	fprintf(stderr,"Error in getdomain(): domain identifiers must not be the same\n");
	fprintf(OUTPUT,"       found two copies of %s, domains %d & %d\n",
		domains[i].id,i+1,j+1);
	return -1;
      }

  return 0;
}


int skiptononspace(char *string, int pointer){
  while(string[pointer]!=' ' && string[pointer]!='\0' && string[pointer]!='\n') pointer++;
  if(string[pointer]=='\0') return -1;
  while(string[pointer]==' ' && string[pointer]!='\0' && string[pointer]!='\n') pointer++;
  if(string[pointer]=='\0') return -1;
  return pointer;
}



void getdomain_error(char *buff) {
	fprintf(stderr,"error in domain descriptors\n");
	fprintf(stderr,"Last domain read:\n%s\n",buff);
	exit(-1);
}


int RBR_getatoms(struct domain_loc *domain, struct atom **atoms, int *natoms, char *pdb_files, int verbose, int file_type, struct parameter *par){

  int i,j,k;
  int pointer;
  int begin,end;
  int number;
  int atom_number;
  int onlastresidue;
  int nres;

  char *pdbfile;
  char buff[100];
  char lastres[20];
  char atomname[10];
  char tmpstring[100];
  char cid,in,alt;

  float x;

  struct residue *res;

  FILE *PDB;
	
  (*natoms) = 0;

  if((PDB=openfile(domain[0].filename,"r"))==NULL){
    pdbfile=RBR_getfile(domain[0].id,pdb_files,4,stdout);
    if(pdbfile[0]=='\0'){
      fprintf(stderr,"Error in readats() no PDB file found for %s\n",domain[0].id);
      return -1;
    }
    strcpy(domain[0].filename,pdbfile);
    free(pdbfile);
    if((PDB=openfile(domain[0].filename,"r"))==NULL){
      fprintf(stderr,"Error in readats() no PDB file found for %s\n",domain[0].id);
      return -1;
    }
  }

  if(verbose) printf("Filename is %s\n",domain[0].filename);

  *atoms = (struct atom *)malloc(sizeof(struct atom));

  for(i=0; i<domain[0].nobj; ++i){
    if(verbose) {
      printf("Domain %s object %4d type %4d\n",domain[0].id,i,domain[0].type[i]);
	   
      switch(domain[0].type[i]) {
      case 1: printf(" all residues"); break;
      case 2: printf(" chain %c",domain[0].start[i].cid); break;
      case 3: printf(" from %c %4d %c to %c %4d %c",
		     domain[0].start[i].cid,domain[0].start[i].n,domain[0].start[i].in,
		     domain[0].end[i].cid,domain[0].end[i].n,domain[0].end[i].in); break;
      }
      printf("\n");
    }

    closefile(PDB,domain[0].filename);
    PDB=openfile(domain[0].filename,"r");

    begin=0;  lastres[0]='\0'; onlastresidue=0;
    while(fgets(buff,99,PDB)!=NULL){
      if((strncmp(buff,"ENDMDL",6)==0 || strncmp(buff,"END   ",6)==0) && begin==1){
	break;
      }
      if((strncmp(buff,"ATOM  ",6)==0) || (par->heteroatoms && (strncmp(buff, "HETATM", 6) == 0))){
	/* get chain, number and insertion code */
	cid=buff[21]; sscanf(&buff[22],"%d",&number); in=buff[26]; alt=buff[16]; /* alternate position indicator */
	if(!begin && ((domain[0].type[i]==1) || (domain[0].start[i].cid==cid && domain[0].start[i].n==number &&
						 domain[0].start[i].in==in) || (domain[0].start[i].cid==cid && domain[0].type[i]==2))){
	  begin=1;
	  onlastresidue=0;
	}

	/* Test if this is a new residue */
	if((strncmp(lastres,&buff[17],11)!=0)){
	  if(begin){
	    if(onlastresidue==1) break;
	  }
	}

	strncpy(lastres,&buff[17],11); lastres[11]='\0';
	if(begin && (alt==' ' || alt=='A' || alt=='1') ) { /* only read one of multiple ATOM entries */
	  if((*atoms = (struct atom *)realloc(*atoms, (((*natoms)+1)*sizeof(struct atom)))) == NULL){
	    fprintf(stderr, "Memory overlflow while reading coordinates\n");
	  }
	  if(cid==' ') atoms[0][(*natoms)].cid='_';
	  else atoms[0][(*natoms)].cid=cid;

	  if(in==' ') atoms[0][(*natoms)].in='_';
	  else atoms[0][(*natoms)].in=in;

	  strncpy(&(atoms[0][(*natoms)].type[0]), &buff[13],3);
	  /* Rob did it with buff[12]:
	     strncpy(&(atoms[0][(*natoms)].type[0]), &buff[12],4);
	  */
	  atoms[0][(*natoms)].type[3] = '\0';
		

	  /*	  atoms[0][(*natoms)].aa = RBR_a3to1(&buff[17]);*/
	  strncpy(atoms[0][(*natoms)].aa, &buff[17],3);
	  atoms[0][(*natoms)].aa[3] = '\0';

	  atoms[0][(*natoms)].aa_num = number;

	  sscanf(&buff[7],"%d",&atom_number); /*evtl. put white space in control string to jump over leading white spaces*/
	  atoms[0][(*natoms)].num = atom_number;

	  if(file_type == 2){
	    strncpy(tmpstring, &buff[54], 8);
	    tmpstring[8] = '\0';
	    sscanf(tmpstring, "%f", &(atoms[0][(*natoms)].aac));
	    /* atoms[0][(*natoms)].aac = (int) (x+0.5);   rounds to int*/
	  }else{
	    atoms[0][(*natoms)].aac = -1; /* accessibility set to -1 */
	  }

	  atoms[0][(*natoms)].rac = -1;

	  for(j=0; j<3; ++j){
	    strncpy(&tmpstring[0],&buff[30+j*8],8);
	    tmpstring[8]='\0';
	    sscanf(tmpstring,"%f",&x);
	    atoms[0][(*natoms)].co[j]=(long int)(PRECIS*x);
	  }

	  (*natoms)++;
	  /*
	  if((*natoms)>MAX_SEQ_LEN) {
	    fprintf(stderr,"error: number of coordinates read surpasses memory limit\n");
	    return -1;
	  }
	  */
	} /* if (begin && alt....*/

	/* Need to fix this */
	if(begin && 
	   (domain[0].type[i]==3 && 
	   domain[0].end[i].cid==cid && domain[0].end[i].n==number && 
	   domain[0].end[i].in==in && domain[0].type[i]==3) ||
	   (domain[0].type[i]==2 && domain[0].start[i].cid!=cid) ||
	   (!((strncmp(buff,"ATOM  ",6)==0) || (par->heteroatoms && (strncmp(buff, "HETATM", 6) == 0))))){
	  onlastresidue=1;
	}
      } /* if(strncmp(buff, "ATOM  "... */
    } /* while(fgets...*/
  } /* for(i=0; i<domain[0].nobj;...) */

  closefile(PDB,domain[0].filename);
  if(!begin){
    fprintf(stderr,"error: begin of sequence not found in PDB file\n");
    (*natoms)=0;
    return -1;
  }
  return *natoms;
}

/*def. from main..
  char excl_file[200];
  char *excl[30];  
  int n_excl;      
*/


int get_excl(FILE *EX, int nexcl, char ***excl){
  int count;
  char buff[100];

  while(fgets(buff,99,EX)!=NULL){
    /* New domain: realloc. memory */
    if(((*excl = (char **)realloc(*excl, ((nexcl+1)*sizeof(char*)))) == NULL)
       || (((*excl)[nexcl] = (char *)malloc(30*sizeof(char))) == NULL)){
      return -1;
    }
    /* parse description line */
    count = 0;
    while((count < strlen(buff)) && (buff[count] != ' ') && (buff[count] != '\n') && (count < 30)) count++;
    strncpy((*excl)[nexcl], &(buff[0]), count);
    (*excl)[nexcl][count] = '\0';
    nexcl++;
  }
  fclose(EX);
  return nexcl;
}

int get_database(struct domain_loc **domains, struct atom ***atoms, int **natoms, int ndomains, int *ndom_file, FILE *DB, struct parameter *par){
  int j, char_count;
  float x;
  char tmpstring[10];
  char buff[100];
  *ndom_file = 0;

  /* read first line and check whether it starts with > */
  if(fgets(buff, 99, DB)==NULL){
    fprintf(stderr, "Error: database file corrupt\n");
  }else if(strncmp(buff, ">", 1) != 0){
    fprintf(stderr, "Error: wrong database format\n");
    exit(-1);
  }else{
    /* file starts with >, seems ok --> alloc. memory for the first round */
    if(((*domains = (struct domain_loc *)realloc(*domains, (ndomains+(*ndom_file)+1)*sizeof(struct domain_loc))) == NULL)
       || ((*atoms = (struct atom **)realloc(*atoms, (ndomains+(*ndom_file)+1)*sizeof(struct atom *))) == NULL)
       || (((*atoms)[ndomains+(*ndom_file)] = (struct atom *)malloc(sizeof(struct atom))) == NULL)
       || ((*natoms = (int *)realloc(*natoms, (ndomains+(*ndom_file)+1)*sizeof(int))) == NULL)){
      return -1;
    }
      (*natoms)[ndomains+(*ndom_file)] = 0;

      /* parse description line */
      char_count = 1;
      while((char_count < strlen(buff)) && (buff[char_count] != ' ') && (buff[char_count] != '\n') && (char_count < 30)) char_count++;
      strncpy((*domains)[ndomains+(*ndom_file)].id, &(buff[1]), (char_count-1));
      (*domains)[ndomains+(*ndom_file)].id[char_count-1] = '\0';
      strcpy((*domains)[ndomains+(*ndom_file)].filename, "UNK");
      (*domains)[ndomains+(*ndom_file)].nobj = 1;
      if(((*domains)[ndomains+(*ndom_file)].type=(int*)malloc(sizeof(int))) == NULL){
	return -1;
      }
      (*domains)[ndomains+(*ndom_file)].type[0] = 1;
  }

  while(fgets(buff,99,DB)!=NULL){
    /* check whether line starts with '>' */
    if(strncmp(buff, ">", 1) == 0){
      /* New domain: realloc. memory */
      (*ndom_file)++;
      if(((*domains = (struct domain_loc *)realloc(*domains, (ndomains+(*ndom_file)+1)*sizeof(struct domain_loc))) == NULL)
	 || ((*atoms = (struct atom **)realloc(*atoms, (ndomains+(*ndom_file)+1)*sizeof(struct atom *))) == NULL)
	 || (((*atoms)[ndomains+(*ndom_file)] = (struct atom *)malloc(sizeof(struct atom))) == NULL)
	 || ((*natoms = (int *)realloc(*natoms, (ndomains+(*ndom_file)+1)*sizeof(int))) == NULL)){
	return -1;
      }
      (*natoms)[ndomains+(*ndom_file)] = 0;

      /* parse description line */
      char_count = 1;
      while((char_count < strlen(buff)) && (buff[char_count] != ' ') && (buff[char_count] != '\n') && (char_count < 30)) char_count++;
      strncpy((*domains)[ndomains+(*ndom_file)].id, &(buff[1]), (char_count-1));
      (*domains)[ndomains+(*ndom_file)].id[char_count-1] = '\0';
      strcpy((*domains)[ndomains+(*ndom_file)].filename, "UNK");
      (*domains)[ndomains+(*ndom_file)].nobj = 1;
      if(((*domains)[ndomains+(*ndom_file)].type=(int*)malloc(sizeof(int))) == NULL){
	return -1;
      }
      (*domains)[ndomains+(*ndom_file)].type[0] = 1;

    }else if(((strncmp(buff,"ATOM  ",6)==0) || (par->heteroatoms && (strncmp(buff, "HETATM", 6) == 0)))
	     && ((buff[16] == ' ') || (buff[16] == 'A') || (buff[16] == '1'))){
      /* Its an atom with alternate pos. indicator 1, A or nothing -> alloc. mem. and store */
      if(((*atoms)[ndomains+(*ndom_file)] = (struct atom *)realloc((*atoms)[ndomains+(*ndom_file)], ((((*natoms)[ndomains+(*ndom_file)])+1)*sizeof(struct atom)))) == NULL){
	fprintf(stderr, "Memory overlflow while reading coordinates\n");
	return -1;
      }
      /* CID: chain identifier */
      if(buff[21] == ' '){
	(*atoms)[ndomains+(*ndom_file)][(*natoms)[ndomains+(*ndom_file)]].cid = '_';
      }else{
	(*atoms)[ndomains+(*ndom_file)][(*natoms)[ndomains+(*ndom_file)]].cid = buff[21];
      }
      /* IN: insertion code */
      if(buff[26] == ' '){
	(*atoms)[ndomains+(*ndom_file)][(*natoms)[ndomains+(*ndom_file)]].in = '_';
      }else{
	(*atoms)[ndomains+(*ndom_file)][(*natoms)[ndomains+(*ndom_file)]].in = buff[26];
      }
      /* AA-number */
      sscanf(&buff[22], "%d", &((*atoms)[ndomains+(*ndom_file)][(*natoms)[ndomains+(*ndom_file)]].aa_num));
      /*	sscanf(&buff[22],"%d",&number);
		atoms[0][(*natoms)].aa_num = number;*/
      /* Atom-type */
      strncpy(&((*atoms)[ndomains+(*ndom_file)][(*natoms)[ndomains+(*ndom_file)]].type[0]), &buff[13], 3);
      (*atoms)[ndomains+(*ndom_file)][(*natoms)[ndomains+(*ndom_file)]].type[3] = '\0';
      /*      strncpy(&(atoms[0][(*natoms)].type[0]), &buff[13],3);
	      Rob did it with buff[12]:
	         strncpy(&(atoms[0][(*natoms)].type[0]), &buff[12],4);
	      atoms[0][(*natoms)].type[3] = '\0'; */
      /* AA-type */
      strncpy(&((*atoms)[ndomains+(*ndom_file)][(*natoms)[ndomains+(*ndom_file)]].aa[0]), &buff[17], 3);
      (*atoms)[ndomains+(*ndom_file)][(*natoms)[ndomains+(*ndom_file)]].aa[3] = '\0';
      /*	  strncpy(atoms[0][(*natoms)].aa, &buff[17],3);
		  atoms[0][(*natoms)].aa[3] = '\0'; */
      /* Atom-number */
      sscanf(&buff[7],"%d", &((*atoms)[ndomains+(*ndom_file)][(*natoms)[ndomains+(*ndom_file)]].num));
      /*	  sscanf(&buff[7],"%d",&(atom_number);
		  atoms[0][(*natoms)].num = atom_number; */

      /*	alt=buff[16]; alternate position indicator */

      /* ACC. values to -1 */
      (*atoms)[ndomains+(*ndom_file)][(*natoms)[ndomains+(*ndom_file)]].aac = -1;
      (*atoms)[ndomains+(*ndom_file)][(*natoms)[ndomains+(*ndom_file)]].rac = -1;

      /* Coordinates */
      for(j=0; j<3; ++j){
	strncpy(&tmpstring[0],&buff[30+j*8],8);
	tmpstring[8]='\0';
	sscanf(tmpstring,"%f",&x);
	(*atoms)[ndomains+(*ndom_file)][(*natoms)[ndomains+(*ndom_file)]].co[j] = (long int)(PRECIS*x);
      }

      (*natoms)[ndomains+(*ndom_file)]++;
    }
  }
  fclose(DB);
  (*ndom_file)++;
}


int tokenize(char *string, char sep, char ***tokens){
  int i,j, n;

  if(((*tokens)=(char **)malloc(sizeof(char *))) == NULL){
    fprintf(stderr, "MEMORY OVERFLOW WHILE TOKENIZING\n");
    return -1;
  }
  /* number of tokens */
  n=0;
  for(i=0; ;i++){
    if(((*tokens)=(char **)realloc((*tokens), (n+1)*sizeof(char *))) == NULL){
      fprintf(stderr, "MEMORY OVERFLOW WHILE TOKENIZING\n");
      return -1;
    }
    /*alloc. mem. for token n maxsize of token is set to 100 */
    if(((*tokens)[n]=(char *)malloc(100*sizeof(char))) == NULL){
      fprintf(stderr, "MEMORY OVERFLOW WHILE TOKENIZING\n");
      return -1;
    }
    /* go through string until sep. char or end or too large */
    for(j=0; (string[i]!=sep) && (string[i]) && (string[i] != '\n') && (j<99); j++){
      (*tokens)[n][j] = string[i];
      i++;
    }
    /* terminate string */
    (*tokens)[n][j] = '\0';

    /* we got a new token -> inc. n*/
    n++;

    /* jump over sep. chars */
    while((string[i] == sep) && (string[i+1] == sep)) i++;
    /*
    if(((*tokens)[n]=(char *)malloc(100*sizeof(char))) == NULL){
      fprintf(stderr, "MEMORY OVERFLOW WHILE TOKENIZING\n");
      return -1;
    }
    */

    if((!string[i]) || (string[i]=='\n')){
      break;
    }
  }
  return n;
}


/*
struct permut_struct {
  int n;
  int a;
  int **p;
};
*/

int gettable(struct permut_struct **permut, struct table_struct **table, struct pseudo_struct **pseudo, char *file, int *ntable_def, int *npseudo_def, char *def_files, float **abund){
  FILE *DAT;
  char buff[100];
  char atom;
  char aa[4];
  char s_atom[100];
  char s_perm[100]; /*holds permutation info before parsing it*/
  int type;
  float aac, rac;
  int max_type;
  int i;
  int j;
  int ntokens, ntokens2;
  char **tokens, **tokens2;
  char *new_file;
  char id_rd; /*switch for identical redirect*/

  if((DAT=fopen(file, "r"))==NULL){
    if((new_file=(char *)malloc(200*sizeof(char))) == NULL){
      fprintf(stderr, "Error in gettable(): Memory overflow\n");
    }
    new_file=RBR_getfile(file, def_files,2,stdout);
    if(new_file[0]=='\0'){
      fprintf(stderr, "Error in gettable(): definition file %s not found / opened\n", file);
      return -1;
    }
    strcpy(file, new_file);
    free(new_file);
    if((DAT=fopen(file, "r"))==NULL){
      fprintf(stderr, "Error in gettable(): definition file %s not found / opened\n", file);
      return -1;
    }
  }

  if(((*table=(struct table_struct *)realloc(*table, sizeof(struct table_struct))) == NULL)
     || ((*pseudo=(struct pseudo_struct *)realloc(*pseudo, sizeof(struct pseudo_struct))) == NULL) 
     || ((*abund=(float *)realloc(*abund, sizeof(float))) == NULL)
     || ((*permut=(struct permut_struct *)realloc(*permut, sizeof(struct permut_struct))) == NULL)){
    fprintf(stderr, "Error in gettable(): Memory overflow\n");
    return -1;
  }

  (*ntable_def) = (*npseudo_def) = 0;
  max_type = -1;
  abund[0][0] = 0;

  while(fgets(buff, 99, DAT)!=NULL){
    if(((buff[0]>='A') && (buff[0]<='Y')) || (buff[0] == '*')){
      ntokens = tokenize(buff, ' ', &tokens);
      if((ntokens < 3) || (ntokens > 5)){
	fprintf(stderr, "Error in gettable(): wrong format in definition file %s\n", file);
	return -1;
      }
      if(strlen(tokens[0]) == 1){
	if(tokens[0][0] == '*'){
	  strncpy(aa,tokens[0], 3);
	  aa[3] = '\0';
	}else{
	  a1to3(tokens[0][0],&(aa[0]));
	}
      }
      else if(strlen(tokens[0]) == 3){
	strncpy(aa,tokens[0],3);
	aa[3] = '\0';
      }
      else{
	fprintf(stderr, "Error in gettable(): wrong format in definition file %s\n", file);
	return -1;
      }

      strcpy(s_atom, tokens[1]);
      
      if(sscanf(tokens[2], "%d", &type) == 0){
	fprintf(stderr, "Error in gettable(): wrong format in definition file %s\n", file);
	return -1;
      }

      if(ntokens > 3){
	if(sscanf(tokens[3], "%f", &rac) == 0){
	fprintf(stderr, "Error in gettable(): wrong format in definition file %s\n", file);
	  return -1;
	}
	if((rac<0) || (rac>100)){
	  rac = -1;
	}
	free(tokens[3]);
      }else {
	rac = -1;
      }

      if(ntokens > 4){
	if(sscanf(tokens[4], "%f", &aac) == 0){
	fprintf(stderr, "Error in gettable(): wrong format in definition file %s\n", file);
	  return -1;
	}
	if((aac<0) || (aac>300)){
	  aac = -1;
	}
	free(tokens[4]);
      }else {
	aac = -1;
      }

      free(tokens[0]);
      free(tokens[1]);
      free(tokens[2]);
      free(tokens);

      if(strlen(s_atom)<4){   /* this is a regular atom */
	if((*table=(struct table_struct *)realloc(*table, ((*ntable_def)+1)*sizeof(struct table_struct))) == NULL){
	  fprintf(stderr, "Error in gettable(): Memory overflow\n");
	  return -1;
	}

	strcpy(&(table[0][(*ntable_def)].aa[0]), aa);
	/*(table[0][(*ntable_def)]).atom = atom;*/
	strcpy(&(table[0][(*ntable_def)].sp_atom[0]), s_atom);
	(table[0][(*ntable_def)]).type = type;
	(table[0][(*ntable_def)]).aac = aac;
	(table[0][(*ntable_def)]).rac = rac;

	
	if(type>max_type){
	  if(type>max_type+1){
	    fprintf(stderr, "Error in gettable(): Types not ordered\n");
            return -1;
          }
	  (permut[0][type]).a = 1;
	  (permut[0][type]).n = 1;
	  if((((permut[0][type]).p = (int **)malloc(sizeof(int *))) == NULL)
	     || (((permut[0][type]).p[0] = (int *)malloc(sizeof(int))) == NULL)){
	    fprintf(stderr, "Error in gettable(): Memory overflow\n");
	    return -1;
	  }
	  permut[0][type].p[0][0] = 0;
	}

	if(type>max_type){
	  if(type>max_type+1){
	    fprintf(stderr, "Error in gettable(): Types not ordered\n");
	    return -1;
	  }
	  max_type++;
	  if((*abund=(float *)realloc(*abund, (max_type+2)*sizeof(float))) == NULL){
	    fprintf(stderr, "Error in gettable(): Memory overflow\n");
	    return -1;
	  }
	  abund[0][max_type]=abundance(RBR_a3to1(aa));

	  if((*permut=(struct permut_struct *)realloc(*permut, (max_type+2)*sizeof(struct permut_struct))) == NULL){
	    fprintf(stderr, "Error in gettable(): Memory overflow\n");
	    return -1;
	  }

	}else{
	  abund[0][max_type]+=abundance(RBR_a3to1(aa));
	}

	(*ntable_def)++;
      }

      else {  /* this is a pseudo-atom definition */
	if((*pseudo=(struct pseudo_struct *)realloc(*pseudo, ((*npseudo_def)+1)*sizeof(struct pseudo_struct))) == NULL){
	  fprintf(stderr, "Error in gettable(): Memory overflow\n");
	  return -1;
	}

	/* now parsing the atoms */
	if(((s_atom[0] != '(') && (s_atom[0] != '[') && (s_atom[0] != '{')) ||
	   ((s_atom[strlen(s_atom)-1] != ')') && (s_atom[strlen(s_atom)-1] != '}') && (s_atom[strlen(s_atom)-1] != ']'))){
	  fprintf(stderr, "Error in gettable(): wrong format in definition file %s\n", file);
	  return -1;
	}
	/*replace ) with \0 */
	s_atom[strlen(s_atom)-1] = '\0';

	/*first, separate atoms (token[0]) from permut (token[1])*/
	if((ntokens = tokenize(&s_atom[1], '|', &tokens)) > 2){
	  fprintf(stderr, "Error in gettable(): wrong format in definition file %s\n", file);
	  return -1;
	}
	/*put atom info in token[0] back to s_atom*/
	strcpy(s_atom, tokens[0]);
	free(tokens[0]);

	if(ntokens == 2){
	  strcpy(s_perm, tokens[1]);
	  free(tokens[1]);
	}else{
	  s_perm[0]='\0';
	}
	free(tokens);

	/*
	printf("%s\t%s\n",s_atom, s_perm);
	*/

	/*First parse the atoms*/
	if((ntokens = tokenize(s_atom, ',', &tokens)) > 15){
	  fprintf(stderr, "Error in gettable(): wrong format in definition file %s\n", file);
	  return -1;
	}
	for(i=0; i<ntokens; i++){
	  if(strlen(tokens[i])<4){
	  strcpy((pseudo[0][(*npseudo_def)]).atom[i], tokens[i]);
	  } else {
	    fprintf(stderr, "Error in gettable(): wrong format in definition file %s\n", file);
	    return -1;
	  }
	  free(tokens[i]);
	}
	free(tokens);

	(pseudo[0][(*npseudo_def)]).natoms = ntokens;

 	strcpy(&(pseudo[0][(*npseudo_def)].aa[0]), aa);

	(pseudo[0][(*npseudo_def)]).type = type;
	(pseudo[0][(*npseudo_def)]).aac = aac;
	(pseudo[0][(*npseudo_def)]).rac = rac;


	/*Now parse the permutation info
	 *this can be off, F (flipp), R (rot.), A (F+R) or a list of permutations
	 *We allow only one permutation per type -> check consistency !
	 *
	 *1st: is type new, then just go ahead.. */
	if(type>max_type){
	  if(type>max_type+1){
	    fprintf(stderr, "Error in gettable(): Types not ordered\n");
	    return -1;
	  }
	  if(s_perm[0]=='\0'){
	    /*no permutation only 1:1 redirect*/
	    (permut[0][type]).a = (pseudo[0][(*npseudo_def)]).natoms;
	    (permut[0][type]).n = 1;
	    if((((permut[0][type]).p = (int **)malloc(sizeof(int *))) == NULL)
	       || (((permut[0][type]).p[0] = (int *)malloc((permut[0][type]).a*sizeof(int))) == NULL)){
	      fprintf(stderr, "Error in gettable(): Memory overflow\n");
	      return -1;
	    }
	    for(i=0;i<(permut[0][type]).a; i++){
	      (permut[0][type]).p[0][i] = i;
	    }
	  }else if(s_perm[0]=='F'){
	    /*flipped*/
	    (permut[0][type]).a = (pseudo[0][(*npseudo_def)]).natoms;
	    (permut[0][type]).n = 2;
	    
	    if((((permut[0][type]).p = (int **)malloc(2*sizeof(int *))) == NULL)
	       || (((permut[0][type]).p[0] = (int *)malloc((permut[0][type]).a*sizeof(int))) == NULL)
	       || (((permut[0][type]).p[1] = (int *)malloc((permut[0][type]).a*sizeof(int))) == NULL)){
	      fprintf(stderr, "Error in gettable(): Memory overflow\n");
	      return -1;
	    }
	    for(i=0;i<(permut[0][type]).a; i++){
	      (permut[0][type]).p[0][i] = i;
	      (permut[0][type]).p[1][i] = ((permut[0][type]).a -i -1);
	    }
	  }else if(s_perm[0]=='R'){
	    /*rotation*/
	    (permut[0][type]).a = (pseudo[0][(*npseudo_def)]).natoms;
	    (permut[0][type]).n = (permut[0][type]).a;
	    if(((permut[0][type]).p = (int **)malloc((permut[0][type]).n*sizeof(int *))) == NULL){
	      fprintf(stderr, "Error in gettable(): Memory overflow\n");
	      return -1;
	    }
	    for(i=0;i<(permut[0][type]).n; i++){
	      if(((permut[0][type]).p[i] = (int *)malloc((permut[0][type]).a*sizeof(int))) == NULL){
		fprintf(stderr, "Error in gettable(): Memory overflow\n");
		return -1;
	      }
	    }
	    
	    for(i=0;i<(permut[0][type]).n; i++){
	      for(j=0;j<(permut[0][type]).a; j++){
		if((i+j) < (permut[0][type]).a){
		  (permut[0][type]).p[i][j] = i+j;
		}else{
		  (permut[0][type]).p[i][j] = i+j - (permut[0][type]).a;
	      }
	    }
	    }
	  }else if(s_perm[0]=='A'){
	    /*rotation AND flipped*/
	    (permut[0][type]).a = (pseudo[0][(*npseudo_def)]).natoms;
	    (permut[0][type]).n = 2*(permut[0][type]).a;
	    
	    if(((permut[0][type]).p = (int **)malloc((permut[0][type]).n*sizeof(int *))) == NULL){
	      fprintf(stderr, "Error in gettable(): Memory overflow\n");
	      return -1;
	    }
	    for(i=0;i<(permut[0][type]).n; i++){
	      if(((permut[0][type]).p[i] = (int *)malloc((permut[0][type]).a*sizeof(int))) == NULL){
		fprintf(stderr, "Error in gettable(): Memory overflow\n");
		return -1;
	      }
	    }
	    for(i=0;i<(permut[0][type]).a; i++){ /*it's actually n/2 */
	      for(j=0;j<(permut[0][type]).a; j++){
		if((i+j) < (permut[0][type]).a){
		  (permut[0][type]).p[i][j] = i+j;
		  (permut[0][type]).p[i+(permut[0][type]).a][j] = ((permut[0][type]).a - (i+j) -1);
		}else{
		  (permut[0][type]).p[i][j] = i+j - (permut[0][type]).a;
		  (permut[0][type]).p[i+(permut[0][type]).a][j] = ((permut[0][type]).a - (i+j - (permut[0][type]).a) -1);
		    /*(permut[0][type]).a - (i+j - (permut[0][type]).a);*/
		}
	      }
	    }
	  }else{
	    /*parse permutation string "0,1,2;1,2,0;2,0,1" etc.
	     *We don't know how many permutations are there
	     *and we have to add/check for the identical redirection !
	     *first : only the redirection, we realloc later */

	    (permut[0][type]).n = 1;
	    (permut[0][type]).a = (pseudo[0][(*npseudo_def)]).natoms;
	    if((((permut[0][type]).p = (int **)malloc(sizeof(int *))) == NULL)
	       || (((permut[0][type]).p[0] = (int *)malloc((permut[0][type]).a*sizeof(int))) == NULL)){
	      fprintf(stderr, "Error in gettable(): Memory overflow\n");
	      return -1;
	    }
	    for(i=0;i<(permut[0][type]).a; i++){
	      (permut[0][type]).p[0][i] = i;
	    }
	      
	    /*first: number of permutations : parse for ';'*/
	    ntokens = tokenize(s_perm, ';', &tokens);
	    
	    for(i=0; i<ntokens; i++){
	      /*	      printf("Token: %s\n", tokens[i]);*/
	      /*now, get atom positions : parse for ',' */
	      if((ntokens2 = tokenize(tokens[i], ',', &tokens2)) != (permut[0][type]).a){
		/*number of atom positions must be numer of atoms !*/
		fprintf(stderr, "Error in gettable(): wrong format in definition file %s\n", file);
		return -1;
	      }
	      /*
	      printf("ntoken2: %i\n", ntokens2);
	      */
	      /*check whether it's just identical redirect and check number of atoms*/
	      id_rd = '\1';
	      for(j=0; j<ntokens2; j++){
		/*		printf("Token2: %s\n", tokens2[j]);*/
		if(atoi(tokens2[j]) >=(permut[0][type]).a){
		  fprintf(stderr, "Error in gettable(): wrong format in definition file %s\n", file);
		  return -1;
		}
		if(atoi(tokens2[j]) != j){
		  id_rd = '\0';
		}
	      }
	      if(!id_rd){
		/*new type of permutation
		 *incr. n, realloc p and alloc memory only for the newest (n-1)*/
		((permut[0][type]).n)++;
		if((((permut[0][type]).p = (int **)realloc((permut[0][type]).p, (permut[0][type]).n*sizeof(int *))) == NULL)
		   || (((permut[0][type]).p[((permut[0][type]).n)-1] = (int *)malloc((permut[0][type]).a*sizeof(int))) == NULL)){
		  fprintf(stderr, "Error in gettable(): Memory overflow\n");
		  return -1;
		}
		for(j=0; j<ntokens2; j++){
		  (permut[0][type]).p[((permut[0][type]).n)-1][j] = atoi(tokens2[j]);
		}
	      }
	      for(j=0; j<ntokens2; j++){
		free(tokens2[j]);
	      }
	      free(tokens2);

	      free(tokens[i]);
	    }
	    free(tokens);
	  }/*manual permutation definition*/

	  /*print permutations to check
	  printf("Type: %i\n", type);
	  for(i=0; i<permut[0][type].n; i++){
	    for(j=0; j<permut[0][type].a; j++){
	      printf("%i\t", permut[0][type].p[i][j]);
	    }
	    printf("\n");
	  }
	  */

	} /* if(type>maxtype...)*/
	

	if(type>max_type){
	  if(type>max_type+1){
	    fprintf(stderr, "Error in gettable(): Types not ordered\n");
	    return -1;
	  }
	  max_type++;
	  if((*abund=(float *)realloc(*abund, (max_type+2)*sizeof(float))) == NULL){
	    fprintf(stderr, "Error in gettable(): Memory overflow\n");
	    return -1;
	  }
	  abund[0][max_type]=abundance(RBR_a3to1(aa));

	  if((*permut=(struct permut_struct *)realloc(*permut, (max_type+2)*sizeof(struct permut_struct))) == NULL){
	    fprintf(stderr, "Error in gettable(): Memory overflow\n");
	    return -1;
	  }


	}else{
	  abund[0][max_type]+=abundance(RBR_a3to1(aa));
	}

	(*npseudo_def)++;
      }
    }
  }

  fclose(DAT);
 
  max_type++;

  return max_type;
}

int get_points(struct point **points, struct atom *atoms, struct table_struct *table, struct pseudo_struct *pseudos, int n, int ntable_def, int npseudo_def, struct parameter *par){
  int i, j, k, l;
  int old_aa_num;
  int npoints;
  int *natoms;
  struct atom ***atom_buff;

  /* eventually : the first point has already been allocated in main 
   * maybe this caused efence message ??
   * take it out or try realloc ??
  if((*points=(struct point *)malloc(sizeof(struct point))) == NULL){
    fprintf(stderr, "MEMORY OVERFLOW WHILE CREATING POINTS\n");
    return -1;
  }
  */
  if(((natoms=(int *)malloc((npseudo_def+1)*sizeof(int))) == NULL)
     || ((atom_buff=(struct atom ***)malloc((npseudo_def+1)*sizeof(struct atom **))) == NULL)){
    fprintf(stderr, "MEMORY OVERFLOW WHILE CREATING POINTS\n");
    return -1;
  }

  for(i=0; i<npseudo_def; i++){
    natoms[i] = 0;
    if((atom_buff[i]=(struct atom **)malloc(100*sizeof(struct atom *))) == NULL){
      fprintf(stderr, "MEMORY OVERFLOW WHILE CREATING POINTS\n");
      return -1;
    }
  }

  npoints=0;
  old_aa_num = atoms[0].aa_num;

  for(i=0; i<n; i++){
    if(((par->abs_surf<=0) || ((par->abs_surf>0) && (par->abs_surf<=atoms[i].aac))) && ((par->rel_surf<=0) || ((par->rel_surf>0) && (par->rel_surf<=atoms[i].rac)))){
      /* both requirements have to be either off or satisfied */
      for(j=0; j<ntable_def; j++){
	if((table[j].aa[0] == '*') || (strcmp(table[j].aa,atoms[i].aa) == 0)){
	  /*assumes that table contains the smaller string (e.g. "CA" instead of "CA ") I found no alternative */
	  if((table[j].sp_atom[0] == '*') || (strncmp(table[j].sp_atom, atoms[i].type, strlen(table[j].sp_atom))==0)){
	    /*(par off) OR (defined value satisfied) OR (definition off) */
	    if(((par->abs_surf != 0) || ((table[j].aac>0) && (table[j].aac<=atoms[i].aac)) || (table[j].aac<=0))
	       && ((par->rel_surf != 0) || ((table[j].rac>0) && (table[j].rac<=atoms[i].rac)) || (table[j].rac<=0))){
	      if((*points=(struct point *)realloc(*points, (npoints+1)*sizeof(struct point)))==NULL){
		fprintf(stderr, "MEMORY OVERFLOW while creating POINTS\n");
		return -1;
	      }
	      points[0][npoints].type = table[j].type;
	      points[0][npoints].number = 1;
	      if((points[0][npoints].atoms=(struct atom **)malloc(sizeof(struct atom *)))==NULL){
		fprintf(stderr, "MEMORY OVERFLOW while creating POINTS\n");
		return -1;
	      }
	      points[0][npoints].atoms[0] = &atoms[i];
	      for(k=0; k<3; k++){
		points[0][npoints].co[k] = atoms[i].co[k];
	      }
	      npoints++;
	      break;
	    }
	  }
	}
    } /* for(j... */

    /* creation of pseudo-atoms */
      if(atoms[i].aa_num != old_aa_num){
      /*new residue, do calculation for the old one*/
	for(j=0; j<npseudo_def; j++){
	  if(natoms[j] == pseudos[j].natoms){
	    /*enough atoms, create pseudo atom*/
	    if((*points=(struct point *)realloc(*points, (npoints+1)*sizeof(struct point)))==NULL){
	      fprintf(stderr, "MEMORY OVERFLOW while creating POINTS\n");
	      return -1;
	    }
	    points[0][npoints].number = pseudos[j].natoms;
	    points[0][npoints].type = pseudos[j].type;
	    if((points[0][npoints].atoms=(struct atom **)malloc(pseudos[j].natoms*sizeof(struct atom *)))==NULL){
	      fprintf(stderr, "MEMORY OVERFLOW while creating POINTS\n");
	      return -1;
	    }
	    for(l=0; l<3; l++){
	      points[0][npoints].co[l] = 0;
	    }
	    for(k=0; k<natoms[j]; k++){
	      points[0][npoints].atoms[k] = atom_buff[j][k];
	      for(l=0; l<3; l++){
		points[0][npoints].co[l] += (long int) (((*atom_buff[j][k]).co[l])/natoms[j]);
	      }
	  }
	    npoints++;
	  }
      } /*end of for(j... checking for found points */
	for(j=0; j<npseudo_def; j++){
	  natoms[j] = 0;
	}
	old_aa_num = atoms[i].aa_num;
      } /* end of what is done for new residues */
      else{
	/*same residue, check and store*/
	for(j=0; j<npseudo_def; j++){
	  /* j goes through the "lines" of pseudo_def */
	  if((pseudos[j].aa[0] == '*') || (strcmp(pseudos[j].aa, atoms[i].aa) == 0)){
	    /* aa requirement ok*/
	    for(k=0; k<pseudos[j].natoms; k++){
	      if(strncmp(pseudos[j].atom[k], atoms[i].type, strlen(pseudos[j].atom[k]))==0){
		/* atom requirement ok*/
		/*test surface : (par off) OR (defined value satisfied) OR (definition off) */
		if(((par->abs_surf != 0) || ((pseudos[j].aac>0) && (pseudos[j].aac<=atoms[i].aac)) || (pseudos[j].aac<=0))
		   && ((par->rel_surf != 0) || ((pseudos[j].rac>0) && (pseudos[j].rac<=atoms[i].rac)) || (pseudos[j].rac<=0))){
		  /*surface requirement ok*/
		  atom_buff[j][k] = &atoms[i]; /* using k ensures that the atoms are stored in the same order as
							* in the definition file */
		  natoms[j]++;
		}
	      }
	    }
	  }
	}
      } /*end else*/
    }
  }
  /*the last res. has not been checked for pseudoatoms*/
  for(j=0; j<npseudo_def; j++){
    if(natoms[j] == pseudos[j].natoms){
      /*enough atoms, create pseudo atom*/
      if((*points=(struct point *)realloc(*points, (npoints+1)*sizeof(struct point)))==NULL){
	fprintf(stderr, "MEMORY OVERFLOW while creating POINTS\n");
	return -1;
      }
      points[0][npoints].number = pseudos[j].natoms;
      points[0][npoints].type = pseudos[j].type;
      if((points[0][npoints].atoms=(struct atom **)malloc(pseudos[j].natoms*sizeof(struct atom *)))==NULL){
	fprintf(stderr, "MEMORY OVERFLOW while creating POINTS\n");
	return -1;
      }
      for(l=0; l<3; l++){
	points[0][npoints].co[l] = 0;
      }
      for(k=0; k<natoms[j]; k++){
	points[0][npoints].atoms[k] = atom_buff[j][k];
	for(l=0; l<3; l++){
	  points[0][npoints].co[l] += (long int) (((*atom_buff[j][k]).co[l])/natoms[j]);
	}
      }
      npoints++;
    }
  } /*end of for(j... checking for found points */

  free(natoms);
  for(i=0; i<npseudo_def; i++){
    free(atom_buff[i]);
  }
  free(atom_buff);

  return npoints;
}

int read_dssp(struct atom *atoms, int natoms, char *dssp_file){
  int count, i, j, max_a;
  char buff[1000];
  char aa;
  struct surf *acc;
  FILE *DSSP;

  if((DSSP = fopen(dssp_file, "r")) == NULL){
    fprintf(stderr, "Error in read_dssp() : dssp-file %s not found\n", dssp_file);
    return -1;
  }
  if((acc = (struct surf *)malloc(sizeof(struct surf))) == NULL){
    fprintf(stderr, "Error in read_dssp() : memory overflow\n");
    return -1;
  }
  while((fgets(buff, 999, DSSP) != 0) && (strncmp(buff, "  #  RESIDUE", 12) != 0));
  count = 0;
  while(fgets(buff, 999, DSSP) != 0){
    if((strlen(buff)<40) || (buff[13] == '!')){
      continue;
    }
    if((acc = (struct surf *)realloc(acc, (count+1)*sizeof(struct surf))) == NULL){
      fprintf(stderr, "Error in read_dssp() : memory overflow\n");
      return -1;
    }
    sscanf(&buff[7], "%d", &acc[count].num);
    aa = buff[13];
    if((aa >= 'a') && (aa <= 'z')){
      aa = 'C';
    }
    if((max_a = max_acc(aa)) == -1){
      fprintf(stderr, "Error in dssp-file %s\n", dssp_file);
      return -1;
    }
    sscanf(&buff[35], "%f", &acc[count].aac);
    acc[count].rac = 100 * acc[count].aac / max_a;

    a1to3(aa, &(acc[count].aa[0]));

    acc[count].cid = buff[11];
    if(acc[count].cid == ' ') {
      acc[count].cid = '_';
    }
    acc[count].in = buff[10];
    if(acc[count].in == ' '){
      acc[count].in = '_';
    }

    count++;
  }
  fclose(DSSP);

  for(i=0; i<natoms; i++){
    for(j=0; j<count; j++){
      if((atoms[i].aa_num == acc[j].num) && (atoms[i].cid == acc[j].cid) && (atoms[i].in == acc[j].in)){
	atoms[i].aac = acc[j].aac;
	atoms[i].rac = acc[j].rac;
	break;
      }
    }
    if(j == count){
      fprintf(stderr, "Error in read_dssp() : no acc. value found for %s %i %c %c\n", atoms[i].aa, atoms[i].aa_num, atoms[i].cid, atoms[i].in);
    }
  }
  free(acc);
  return count;
}

int read_rsa(struct atom *atoms, int natoms, char *dssp_file){
  int count, i, j;
  char buff[1000];
  struct surf *acc;
  FILE *DSSP;

  if((DSSP = fopen(dssp_file, "r")) == NULL){
    fprintf(stderr, "Error in read_naccess : naccess-rsa-file %s not found\n", dssp_file);
    return -1;
  }
  if((acc = (struct surf *)malloc(sizeof(struct surf))) == NULL){
    fprintf(stderr, "Error in read_dssp : memory overflow\n");
    return -1;
  }
  while((fgets(buff, 999, DSSP) != 0) && (strncmp(buff, "  #  RESIDUE", 12) != 0));
  count = 0;
  while((fgets(buff, 999, DSSP) != 0) && (strlen(buff)>37)){
    if(strncmp(buff, "RES ", 4) != 0){
      continue;
    }
    if((acc = (struct surf *)realloc(acc, (count+1)*sizeof(struct surf))) == NULL){
      fprintf(stderr, "Error in read_naccess : memory overflow\n");
      return -1;
    }
    /*
REM RES _ NUM      All-atoms   Total-Side   Main-Chain    Non-polar    All polar 
REM                ABS   REL    ABS   REL    ABS   REL    ABS   REL    ABS   REL 
RES ASN   123    68.36  47.5  68.36  64.3   0.00   0.0  15.22  32.9  53.14  54.4 
RES ALA    27    76.22  70.6  55.01  79.3  21.21  55.0  56.95  79.8  19.27  52.7 
RES ASN     5   120.50  83.7  91.71  86.3  28.79  76.4  38.29  82.8  82.22  84.1   
012345678901234567890123456789012345678901234567890123456789
    */
    sscanf(&buff[9], "%d", &acc[count].num);
    strncpy(acc[count].aa, &buff[4], 3);
    acc[count].aa[3] = '\0';

    sscanf(&buff[16], "%f", &acc[count].aac);
    sscanf(&buff[23], "%f", &acc[count].rac);

    acc[count].cid = buff[8];
    if(acc[count].cid == ' ') {
      acc[count].cid = '_';
    }
    /* I have no clue about the insertion code here !!! */
    acc[count].in = buff[13];
    if(acc[count].in == ' '){
      acc[count].in = '_';
    }

    count++;
  }
  fclose(DSSP);

  for(i=0; i<natoms; i++){
    for(j=0; j<count; j++){
      if((atoms[i].aa_num == acc[j].num) && (atoms[i].cid == acc[j].cid) && (atoms[i].in == acc[j].in)){
	atoms[i].aac = acc[j].aac;
	atoms[i].rac = acc[j].rac;
	break;
      }
    }
    if(j == count){
      fprintf(stderr, "Error in read_naccess() : not all residues found\n");
      return -1;
    }
  }
  free(acc);
  return 1;
}

int read_asa(struct atom *atoms, int natoms, char *dssp_file){
  int count, i, j;
  char buff[1000];
  struct surf *acc;
  FILE *DSSP;
  char cid, alt, in;
  int number;
  char tmpstring[10];
  float f;

  if((DSSP = fopen(dssp_file, "r")) == NULL){
    fprintf(stderr, "Error in read_naccess : naccess-rsa-file %s not found\n", dssp_file);
    return -1;
  }
  if((acc = (struct surf *)malloc(sizeof(struct surf))) == NULL){
    fprintf(stderr, "Error in read_dssp : memory overflow\n");
    return -1;
  }
  while((fgets(buff, 999, DSSP) != 0) && (strncmp(buff, "  #  RESIDUE", 12) != 0));
  count = 0;
  while((fgets(buff, 999, DSSP) != 0) && (strlen(buff)>37)){
    if(strncmp(buff, "RES ", 4) != 0){
      continue;
    }
    if((acc = (struct surf *)realloc(acc, (count+1)*sizeof(struct surf))) == NULL){
      fprintf(stderr, "Error in read_naccess : memory overflow\n");
      return -1;
    }
    /*
012345678901234567890123456789012345678901234567890123456789
ASA:
ATOM     61  CG  ASN     8       0.444  16.671  -3.119   3.909  1.76
ATOM     62  OD1 ASN     8      -0.686  17.233  -3.227  18.407  1.40
ATOM     80  CB  ALA    10      -4.661  16.539  -2.091  36.102  1.87
ATOM   1428  N   ASP   181       1.636   8.602  30.565   0.000  1.65
ATOM   1429  CA  ASP   181       2.602   9.177  31.500   1.503  1.87 
ATOM   2495  CD1 ILE I   7     -18.146 -15.801   7.313  23.698  1.87
PDB:
ATOM   1477  ND1 HIS   186        .406  19.614  29.742  1.00 13.27      4CPA1616 
ATOM   2715  CA  TYR I  37     -13.229  -3.379  13.624  1.00  6.00      4CPAB274
    */

    if(buff[16] == ' ' || buff[16] == 'A' || buff[16] == '1'){

      sscanf(&buff[7], "%d", &acc[count].num); /* atom number */
      strncpy(acc[count].aa, &buff[13], 3); /*atom type */
      acc[count].aa[3] = '\0';
      acc[count].cid = buff[21];
      if(acc[count].cid == ' ') {
	acc[count].cid = '_';
      }
      acc[count].in = buff[26];
      if(acc[count].in == ' '){
	acc[count].in = '_';
      }
      strncpy(tmpstring, &buff[54], 8);
      tmpstring[8] = '\0';
      sscanf(tmpstring, "%f", &(acc[count].aac));
      /* acc[count].aac = (int) (f+0.5);   rounds to int*/

      count++;
    }
  }

  fclose(DSSP);

  for(i=0; i<natoms; i++){
    for(j=0; j<count; j++){
      if((atoms[i].num == acc[j].num) && (atoms[i].cid == acc[j].cid) && (atoms[i].in == acc[j].in) && (strcmp(atoms[i].type, acc[j].aa) == 0)){
	atoms[i].aac = acc[j].aac;
	break;
      }
    }
    if(j == count){
      fprintf(stderr, "Error in read_asa() : no acc. value found for %s %i %c %c\n", atoms[i].aa, atoms[i].aa_num, atoms[i].cid, atoms[i].in);
    }
  }
  free(acc);
  return 1;
}

int rms_filter(struct store_struc *store, struct parameter *par){
  register int i, j;
  register struct store_struc swap_store;

  if(par->max_rms == -1){
    return;
  }
  j=0;
  for(i=0; i<par->found; i++){
    if(store[i].rms <= par->max_rms){
      /* we swap the array (j is incremented after swapping)
       * only assigning store[j] = store[i] without 
       * store[i] = store[j] results in 2 copies store[i] and prevents
       * freeing memory in main !! 
       * (same holds for all other filters ! */
      swap_store = store[j];
      store[j++] = store[i];
      store[i] = swap_store;
    }
  }
  par->found = j;
  return j;
}


int P_filter(struct store_struc *store, struct parameter *par){
  register int i, j;
  register struct store_struc swap_store;

  j=0;
  for(i=0; i<par->found; i++){
    if(store[i].P <= par->max_P){
      swap_store = store[j];
      store[j++] = store[i];
      store[i] = swap_store;
    }
  }
  par->found = j;
  return j;
}

int Ps_filter(struct store_struc *store, struct parameter *par){
  register int i, j;
  register struct store_struc swap_store;

  j=0;
  for(i=0; i<par->found; i++){
    if(store[i].Ps <= par->max_Ps){
      swap_store = store[j];
      store[j++] = store[i];
      store[i] = swap_store;
    }
  }
  par->found = j;
  return j;
}


void qs_P(struct store_struc *store, int left, int right){
  register int i, j;
  register double P;
  register struct store_struc swap;

  i = left; j = right;
  P = store[(left+right)/2].P;

  do{
    while((store[i].P < P) && (i < right)) i++;
    while((store[j].P > P) && (j > left)) j--;

    if(i <= j){
      swap = store[i];
      store[i] = store[j];
      store[j] = swap;
      i++; j--;
    }
  }while(i<=j);

  if(left < j) qs_P(store, left, j);
  if(i < right) qs_P(store, i, right);
}

void sort_P(struct store_struc *store, int n){
  qs_P(store, 0, n-1);
}

void qs_size(struct store_struc *store, int left, int right){
  register int i, j, n;
  register struct store_struc swap;

  i = left; j = right;
  n = store[(left+right)/2].n;

  do{
    while((store[i].n > n) && (i < right)) i++;
    while((store[j].n < n) && (j > left)) j--;

    if(i <= j){
      swap = store[i];
      store[i] = store[j];
      store[j] = swap;
      i++; j--;
    }
  }while(i<=j);

  if(left < j) qs_size(store, left, j);
  if(i < right) qs_size(store, i, right);
}

void sort_size(struct store_struc *store, int n){
  qs_size(store, 0, n-1);
}


int LR_filter(struct store_struc *store, struct point **points, struct parameter *par){
  /*this filter removes all patterns that do not have at least one residue-residue
   *distance of par->long_range. This is measured on the first dataset (0) only.
   *If residues have different chain ids, long distance OK is assumed. */

  int i, j, k, l;
  char *rm;
  struct store_struc swap_store;

  /*all 'removed'*/
  rm = (char *)malloc(par->found*sizeof(char));
  for(i=0; i<par->found; i++){
    rm[i] = '\1';
  }

  /*for each motif i...*/
  for(i=0; i<par->found; i++){
    /*now pairwise all residues k and l within that motif...*/
    for(k=0; k<store[i].n-1; k++){
      for(l=k+1; l<store[i].n; l++){
	if((((*(points[0][(store[i].p1[k])].atoms[0])).aa_num - (*(points[0][(store[i].p1[l])].atoms[0])).aa_num) >= par->long_range)
	   || (((*(points[0][(store[i].p1[k])].atoms[0])).aa_num - (*(points[0][(store[i].p1[l])].atoms[0])).aa_num) <= (- par->long_range))
	   || ((*(points[0][(store[i].p1[k])].atoms[0])).cid != (*(points[0][(store[i].p1[l])].atoms[0])).cid)){
	  rm[i] = '\0';
	  break;
	}
      }
      if(!rm[i]){
	break;
      }
    }
  }

  j=0;
  for(i=0; i<par->found; i++){
    if(!rm[i]){
      swap_store = store[j];
      store[j++] = store[i];
      store[i] = swap_store;
    }
  }

  free(rm);

  par->found = j;
  return j;
}







int rm_submotifs(struct store_struc *store, struct parameter *par){
  int i, j, k, l, min, n;
  char *rm;
  int s_count;
  float s_rms;
  struct store_struc swap_store;

  /*sorting from small -> big*/
  sort_size(store, par->found);

  /*motifs are sorted big --> small (j is always smaller or equal)*/
  rm = (char *)malloc(par->found*sizeof(char));
  for(i=0; i<par->found; i++){
    rm[i] = '\0';
  }

  for(i=0; i<par->found; i++){
    if(rm[i]){
      continue; /*already excluded*/
    }
    for(j=i+1; j<par->found; j++){
      if((store[i].n == store[j].n) || (rm[j])){
	continue; /* if they have the same size, they are not submotifs of each other or already excluded */
      }
      n = 0;
      for(k=0; k<store[i].n; k++){
	for(l=0; l<store[j].n; l++){
	  if((store[i].p1[k] == store[j].p1[l]) && (store[i].p2[k] == store[j].p2[l])){
	    n++;
	  }
	}
      }
      if(n == store[j].n){ /* if number of identical pairs equals size of smaller motif, rm smaller motif */
	rm[j] = '\1';
      }
    }
  }

  j=0;
  for(i=0; i<par->found; i++){
    if(!rm[i]){
      swap_store = store[j];
      store[j++] = store[i];
      store[i] = swap_store;
    }
  }

  free(rm);

  par->found = j;
  return j;
}


int rm_submotifs_smart(struct store_struc *store, struct parameter *par){
  int i, j, k, l, min, n;
  char *rm;
  int s_count;
  float s_rms;
  struct store_struc swap_store;

  /*sorting from small -> big*/
  sort_size(store, par->found);

  /*motifs are sorted big --> small (j is always smaller or equal)*/
  rm = (char *)malloc(par->found*sizeof(char));
  for(i=0; i<par->found; i++){
    rm[i] = '\0';
  }

  for(i=0; i<par->found; i++){
    if(rm[i]){
      continue; /*already excluded*/
    }
    for(j=i+1; j<par->found; j++){
      if((store[i].n == store[j].n) || (rm[j])){
	continue; /* if they have the same size, they are not submotifs of each other or already excluded */
      }
      n = 0;
      for(k=0; k<store[i].n; k++){
	for(l=0; l<store[j].n; l++){
	  if((store[i].p1[k] == store[j].p1[l]) && (store[i].p2[k] == store[j].p2[l])){
	    n++;
	  }
	}
      }
      if((n == store[j].n) && (store[j].Ps > store[i].Ps)){ /* if number of identical pairs equals size of smaller motif AND P is worse, rm smaller motif */
	/* could be changed to > 10*store[i].Ps... */
	rm[j] = '\1';
      }
    }
  }

  j=0;
  for(i=0; i<par->found; i++){
    if(!rm[i]){
      swap_store = store[j];
      store[j++] = store[i];
      store[i] = swap_store;
    }
  }

  free(rm);

  par->found = j;
  return j;
}






int rm_permutations(struct store_struc *store, struct parameter *par){
  int i, j, k, l, worse, n1, n2;
  char *rm;
  struct store_struc swap_store;

  rm = (char *)malloc(par->found*sizeof(char));
  for(i=0; i<par->found; i++){
    rm[i] = '\0';
  }

  for(i=0; i<par->found; i++){
    for(j=i+1; j<par->found; j++){
      if(store[i].n != store[j].n){
	continue; /* if they have different sizes, they cannot be permutations of each other */
      }else if(store[i].rms >= store[j].rms){ /* min contains number of motif with worse rmsd */
	worse = i;
      }else{
	worse = j;
      }
      n1 = 0;
      n2=0;
      for(k=0; k<store[i].n; k++){
	for(l=0; l<store[j].n; l++){
	  if(store[i].p1[k] == store[j].p1[l]){
	    n1++;
	  }
	  if(store[i].p2[k] == store[j].p2[l]){
	    n2++;
	  }
	}
      }
      if((n1 == store[worse].n) && (n2 == store[worse].n)){ /* if both have same residues rm worse motif */
	rm[worse] = '\1';
      }
    }
  }

  j=0;
  for(i=0; i<par->found; i++){
    if(!rm[i]){
      swap_store = store[j];
      store[j++] = store[i];
      store[i] = swap_store;
    }
  }

  free(rm);

  par->found = j;
  return j;
}


int red_clust(struct sc *clust, int n, float cutoff){
  int i, j, k, l;
  struct sc swap;
  char incl;

  /* reverse array */
  for(i=0; i<((int)n/2); i++){
    swap = clust[i];
    clust[i] = clust[n-1-i];
    clust[n-1-i] = swap;
  }
  
  j=0;
  for(i=0; i<n; i++){
    if(clust[i].score >= cutoff){
      /* clust is good enough, now check wheather included in clusters that
       * have been stored already (store is from 0 to j !!)
       * you have to check only one (the first one) of the new cluster but against
       * all members of all other clusters */
      incl = '\0';
      for(k=0; k<j; k++){
	for(l=0; l<clust[k].n; l++){
	  if(clust[i].members[0] == clust[k].members[l]){
	    incl = '\1';
	    break;
	  }
	}
	if(incl) break;
      }
      if(!incl){
	clust[j] = clust[i];
	j++;
      }
    }
  }
  return j;
}

int red_motifs(struct sc *clust, int n, struct store_struc *store, struct parameter *par){
  int i, j, best;
  char *keep;
  double min;
  struct store_struc swap_store;

  keep = (char *)malloc(par->found*sizeof(char));
  for(i=0; i<par->found; i++){
    keep[i] = '\1';
  }

  for(i=0; i<n; i++){
    min = 10000;
    for(j=0; j<clust[i].n; j++){
      keep[clust[i].members[j]] = '\0'; /* all members of clust are NOT kept */
      if(store[clust[i].members[j]].P<min){
	min = store[clust[i].members[j]].P;
	best = clust[i].members[j];
      }
    }
    keep[best] = '\1'; /* best of group is again set to KEEP */
  }

  j=0;
  for(i=0; i<par->found; i++){
    if(keep[i]){
      swap_store = store[j];
      store[j++] = store[i];
      store[i] = swap_store;
    }
  }

  free(keep);

  par->found = j;
  return j;
}

/*
void sort_motifs(struct store_struc *store, struct parameter *par){
  int i, j;
  struct store_struc swap;

  for(i=(par->found-1); i>0; i--){
    for(j=0; j<i; j++){
      if(store[j].P>store[j+1].P){
	swap = store[j];
	store[j] = store[j+1];
	store[j+1] = swap;
      }
    }
  }
}
*/



int calc_ident(float **ident, struct store_struc *store, struct parameter *par){
  int i, j, k, l, n, min;

  for(i=0; i<(par->found-1); i++){
    for(j=i+1; j<par->found; j++){
      if(store[i].n <= store[j].n){
	min = store[i].n;
      }else{
	min = store[j].n;
      }
      n = 0;
      for(k=0; k<store[i].n; k++){
	for(l=0; l<store[j].n; l++){
	  if((store[i].p1[k] == store[j].p1[l]) && (store[i].p2[k] == store[j].p2[l])){
	    n++;
	  }
	}
      }
      if(par->group_filter == 1){
	ident[i][j-i-1] = (float) n / min;
      }else{
	ident[i][j-i-1] = (float) n;
      }
    }
  }
}


void help(int what){
  printf("\n");
  if(what == 0){
    printf("      -r                         RESIDUE based search (DEFAULT MODE)\n");
    printf("                                  Coordinate-file : pdb-format; Domain-file points to pdb-file\n");
    printf("                                  Residue accessibility : DSSP; Default definition file : sc.def\n");
    printf("      -a                         ATOM based search\n");
    printf("                                  Coordinate-file : NACCESS asa-format; Domain-file points to asa-file\n");
    printf("                                  Residue accessibility : NACCESS; Default definition file : atoms.def\n");
    printf("      -d     <domain file>       domain-file format from STAMP package (1 or 2 domains)\n");
    printf("      -c     <coordinates file>  pdb- / asa-format: all residues / atoms loaded (= one domain)\n");
    printf("      -ma    <integer>           minimum number of points required for match\n");
    printf("                                  DEFAULT: 5 / 10\n");
    printf("      -mt    <float>             maximum distance tolerance allowed during search\n");
    printf("                                  DEFAULT: 3 / 1\n");
    printf("      -ms    <float>             maximum size of motifs to be considered\n");
    printf("                                  DEFAULT: 15\n");
    printf("      -md    <integer>           maximum search depth\n");
    printf("                                  DEFAULT: 20\n");
    printf("      -mr    <float>             maximum rms allowed for match\n");
    printf("                                  DEFAULT: 3 / 1\n");
    printf("      -me    <float>             maximum E-value allowed for match\n");
    printf("                                  DEFAULT: 10\n");
    printf("      -mes   <float>             maximum standard E-value (E700) allowed for match\n");
    printf("                                  DEFAULT: off\n");
    printf("      -gfp   <float>             combine motifs with <float> % identical atoms (group_filter_percent)\n");
    printf("                                  DEFAULT: off\n");
    printf("      -gfa   <float>             combine motifs with <float> identical atoms (group_filter_absolut)\n");
    printf("                                  DEFAULT: off\n");
    printf("      -all   [<float>]           smaller set has to be found to <float>=[0..1] % OR [else] entirely for match\n");
    printf("                                  DEFAULT: off\n");
    printf("      -rs    [<int>]             minimum relative surface accessibility ([<int>] %)\n");
    printf("                                 (without [<int>]: value from definition file is considered)\n");
    printf("                                  DEFAULT: off\n");
    printf("      -as    [<int>]             minimum absolute surface accessibility ([<int>] %)\n");
    printf("                                 (without [<int>]: value from definition file is considered)\n");
    printf("                                  DEFAULT: off / cutoff from atoms.def\n");
    printf("      -v                         verbose\n");
    printf("      -o     <file>              redirect output to file\n");
    printf("      -of    <string>            output format: short/0 OR stamp/trafo/1 OR pdb/struct/2\n");
    printf("                                  DEFAULT: short\n");
    printf("      -ha                        load hetero-atoms from pdb file\n");
    printf("      -def   <definition file>   file containing definitions for considering and matching atoms\n");
    printf("      -rb    <integer>           report n best matches only (off : 0)\n");
    printf("      -rbe   <integer>           report n best matches per database entry (off : 0)\n");
    printf("      -vp                        report permutations\n");
    printf("      -wf                        RMSD calculation on multiple atoms per residue (equal weight for all residues)\n");
    printf("      -rf                        RMSD calculation on multiple atoms per residue\n");
    printf("      -mh    <integer>           maximum number of hits during search (off : 0)\n");
    printf("      -pf    <float>             fact. parameter for statistics\n");
    printf("      -ef    <float>             exp. parameter for staticsics\n");
  } else {
    printf("      -d     <domain file>\n");
    printf("      -c     <coordinates file (pdb/asa)>\n");
    printf("      -help\n");
  }
  printf("\n");
}



int main(int argc, char *argv[]){
  /******************** Declaring Variables ********************/
  struct parameter *par;
  struct table_struct *table;
  struct pseudo_struct *pseudo;
  struct permut_struct *permut; /* holds infos about different atom equivalencing */
  struct store_struc *store; /* holds results during one round of searches before filters */
  struct store_struc *final_res; /* keeps filtered results */

  int nfiles; /* number of files (0 - (nfiles-1))*/
  int *filetype; /* filetype for all nfiles */
  char **filename; /* filename for all files */

  int ndomains; /* number of domains */
  struct domain_loc *domains; /* all ndomains domains */

  struct domain_loc *curr_domains; /* current pair of domains */

  int *natoms; /* number of atoms for each domain */
  struct atom **atoms; /* all atoms for all domains */

  int *curr_natoms; /* number of atoms for current pair of domains */
  struct atom **curr_atoms; /* all atoms for current domains */

  int *npoints; /* number of points for each domain */
  struct point **points; /* all points for all domains */

  int *curr_npoints; /* number of points for current pair of domains */
  struct point **curr_points; /* all points for current pair of domains */

  long int ***dist; /* distances for each of the 2 sets to be compared [0/1] */
  char ***radius; /* binary array for size requirement for each set*/
  char **match; /* binary array for type-matching requirement only one (across 2 sets)*/
  int **motif; /* 2D array (0/1) that holds numbers of equivalenced points DURING the search */
  int *j_array; /* array of ints that redirects access to the points during the search
		 * this allows "swapping" while always counting further */

  int max_points; /* largest number of points */
  int big_dom; /* index of largest domain */
  int ndom_file; /* number of domains in current file */
  int naccess; /* number of acc. values read in */

  int matches_found; /* total number of matches in search (== alloc. store) */
  int matches_kept; /* number of matches kept after filters */

  float **ident;

  float *abund;

  int ntypes;
  int ntables;
  int npseudos;
  int nmatches; /* number of matches in current round (inner loop) */
  int report; /* number of matches reported (min(par->found, par->max_rep_each) */

  int gottrans;
  char *env;
  int i,j, k, l, m,  h_count;
  FILE *DOM;
  struct parameter param;
  char pdb_files[200];
  char dssp_files[200];  
  char asa_files[200];
  char def_files[200];
  char database_files[200];

  char *temp;

  struct tm *time_ptr;
  time_t lt;

  float f;
  int swap_help;
  char char_swap;
  void *array_swap;
  struct domain_loc domain_swap;
  struct point point_swap;
  char swapped;

  struct sc *groups;
  int nclusters;

  /* added 04.03.2003: provide a file with DB-ids to exclude from search */
  FILE *EX; /*file*/
  char **excl;   /* array of ids to exclude, domain_id is defined as char id[30];*/
  int nexcl;         /* number of ids to exclude */



  /******************** Initializing Parameters ********************/

   if(argc<2){
     help(1);
     exit(-1);
   }

   if(((par = (struct parameter*)malloc(sizeof(struct parameter))) == NULL)
      || ((filetype = (int *)malloc(sizeof(int))) == NULL)
      || ((filename = (char **)malloc(sizeof(char *))) == NULL)){
     exit(-1);
   }

   nfiles = 0;

   /* Getting Mode */
   par->mode = 0; /*default Mode is 0 (residue based)*/
   for(i=1; i<argc; ++i){
     if(argv[i][0]!='-'){
       continue;
     }
     if(strcmp(&argv[i][1],"a")==0){
       par->mode = 1;
     }
   }

   /* Setting up Defaults */
   if(par->mode == 0){
     par->max_tolerance = 3*PRECIS;
     par->max_size = 15*PRECIS;
     par->max_depth = 20;
     par->max_hits = 0;
     par->match_abs = 3;
     par->max_rms = 3; 
     strcpy(par->def_file, "sc.def"); /* side-chains */
     par->out = stdout;
     par->out_format = 0;
     par->verbose = 0;
     par->abs_surf = -1; /* off */
     par->rel_surf = -1; /* off */
     par->heteroatoms = '\0';
     par->group_filter = 0;
     par->min_group_sim = 3;
     par->cluster_type = 2; /*1: single, 2: complete, 3: means*/
     par->res_fit = 0; 
   }else{
     par->max_tolerance = 1*PRECIS;
     par->max_size = 15*PRECIS;
     par->max_depth = 20;
     par->max_hits = 0;
     par->match_abs = 10;
     par->max_rms = 1;
     strcpy(par->def_file, "atoms.def"); /* side-chains */
     par->out = stdout;
     par->out_format = 0;
     par->verbose = 0;
     par->abs_surf = 0; /* as specified in def-file */
     par->rel_surf = -1; /* off */
     par->heteroatoms = '\0';
     par->group_filter = 0;
     par->min_group_sim = 3;
     par->cluster_type = 2; /*1: single, 2: complete, 3: means*/
     par->res_fit = 0;
   }
   par->view_permut = '\0';
   par->view_subs = '\0';
   par->max_rep = 0;
   par->max_rep_each = 0;
   par->pf = 0;
   par->pe = 0;
   par->max_P = 10;
   par->max_Ps = -1;
   par->AAA = '\0';
   par->linear = '\0';
   par->match_rel = -1; /* disabled */
   par->long_range = -1; /* disabled */

   nexcl = 0; /*no domain excluded*/

  /******************** Parsing Command Line ********************/

   for(i=1; i<argc; ++i) {
     if(argv[i][0]!='-'){
       help(1);
       exit(-1);
     }
     if((strcmp(&argv[i][1],"d")==0) || (strcmp(&argv[i][1], "dom") == 0)) {
       /* DOMAIN file */
        if(((i+1)>=argc) || (strlen(argv[i+1]) > 999)) exit(-1);
	if(((filetype = (int *)realloc(filetype, (nfiles+1)*sizeof(int))) == NULL)
	   || ((filename = (char **)realloc(filename, (nfiles+1)*sizeof(char *))) == NULL)
	   || ((filename[nfiles] = (char *)malloc(1000*sizeof(char))) == NULL)){
	  exit(-1);
	}
	filetype[nfiles] = 0;
	strcpy(&filename[nfiles][0],argv[i+1]);
	nfiles++;
        i++;
      } else if((strcmp(&argv[i][1],"c")==0) || (strcmp(&argv[i][1], "coo") == 0)) {
       /* COORDINATE (PDB or ASA) file */
        if(((i+1)>=argc) || (strlen(argv[i+1]) > 999)) exit(-1);
	if(((filetype = (int *)realloc(filetype, (nfiles+1)*sizeof(int))) == NULL)
	   || ((filename = (char **)realloc(filename, (nfiles+1)*sizeof(char *))) == NULL)
	   || ((filename[nfiles] = (char *)malloc(1000*sizeof(char))) == NULL)){
	  exit(-1);
	}
	filetype[nfiles] = 1;
	strcpy(&filename[nfiles][0],argv[i+1]);
	nfiles++;
        i++;
      } else if((strcmp(&argv[i][1],"db")==0) || (strcmp(&argv[i][1], "dab") == 0)) {
       /* DATABASE file */
        if(((i+1)>=argc) || (strlen(argv[i+1]) > 999)) exit(-1);
	if(((filetype = (int *)realloc(filetype, (nfiles+1)*sizeof(int))) == NULL)
	   || ((filename = (char **)realloc(filename, (nfiles+1)*sizeof(char *))) == NULL)
	   || ((filename[nfiles] = (char *)malloc(1000*sizeof(char))) == NULL)){
	  exit(-1);
	}
	filetype[nfiles] = 2;
	strcpy(&filename[nfiles][0],argv[i+1]);
	nfiles++;
        i++;
      } else if(strcmp(&argv[i][1],"def")==0){
        if((i+1)>=argc) exit(-1);
        strcpy(&par->def_file[0],argv[i+1]);
        i++;
      } else if((strcmp(&argv[i][1],"rf")==0) || (strcmp(&argv[i][1],"residue_fit")==0)){
	par->res_fit = 1;
      } else if((strcmp(&argv[i][1],"wf")==0) || (strcmp(&argv[i][1],"weighted_fit")==0)){
	par->res_fit = 2;
      } else if(strcmp(&argv[i][1],"help")==0) {
	help(0);
	exit(0);
      } else if((strcmp(&argv[i][1],"ma")==0) || (strcmp(&argv[i][1],"min_atoms")==0)) {  
        if((i+1)>=argc) exit(-1);
        sscanf(argv[i+1],"%i",&par->match_abs);
        i++;
      } else if((strcmp(&argv[i][1],"lr")==0) || (strcmp(&argv[i][1],"long_range")==0)) {  
        if((i+1)>=argc) exit(-1);
        sscanf(argv[i+1],"%i",&par->long_range);
        i++;
      } else if((strcmp(&argv[i][1],"ct")==0) || (strcmp(&argv[i][1],"cluster_type")==0)) {  
        if((i+1)>=argc) exit(-1);
        sscanf(argv[i+1],"%i",&par->cluster_type);
	if((par->cluster_type < 1) || (par->cluster_type > 3)){
	  par->cluster_type = 2;
	}
        i++;
      } else if((strcmp(&argv[i][1],"rs")==0) || (strcmp(&argv[i][1],"rel_surf")==0)) {  
        if(((i+1)>=argc) || (argv[i+1][0] == '-')){
	  par->rel_surf = 0;
	}
	else {
	  sscanf(argv[i+1],"%f",&(par->rel_surf));
	  if(par->rel_surf == 0){
	    par->rel_surf = -1;
	  }
	  i++;
	}
      } else if((strcmp(&argv[i][1],"as")==0) || (strcmp(&argv[i][1],"abs_surf")==0)) {  
        if(((i+1)>=argc) || (argv[i+1][0] == '-')){
	  par->abs_surf = 0;
	}
	else {
	  sscanf(argv[i+1],"%f",&(par->abs_surf));
	  if(par->abs_surf == 0){
	    par->abs_surf = -1;
	  }
	  i++;
	}
      } else if((strcmp(&argv[i][1],"mt")==0) || (strcmp(&argv[i][1],"max_tolerance")==0)) {  
        if((i+1)>=argc) exit(-1);
        sscanf(argv[i+1],"%f",&f);
        par->max_tolerance = (long int)(PRECIS * f);
        i++;
      } else if((strcmp(&argv[i][1],"ms")==0) || (strcmp(&argv[i][1],"max_size")==0)) {  
        if((i+1)>=argc) exit(-1);
        sscanf(argv[i+1],"%f",&f);
	if(f != -1){
	  par->max_size = (int)(PRECIS * f);
	}else{
	  par->max_size = 0;
	}
	i++;
      } else if((strcmp(&argv[i][1],"md")==0) || (strcmp(&argv[i][1],"max_depth")==0)) {  
        if((i+1)>=argc) exit(-1);
        sscanf(argv[i+1],"%i",&par->max_depth);
        i++;
      } else if((strcmp(&argv[i][1],"mh")==0) || (strcmp(&argv[i][1],"max_hits")==0)) {  
        if((i+1)>=argc) exit(-1);
        sscanf(argv[i+1],"%i",&par->max_hits);
        i++;
      } else if((strcmp(&argv[i][1],"rb")==0) || (strcmp(&argv[i][1],"report_best")==0)) {  
        if((i+1)>=argc) exit(-1);
        sscanf(argv[i+1],"%i",&par->max_rep);
        i++;
      } else if((strcmp(&argv[i][1],"rbe")==0) || (strcmp(&argv[i][1],"report_each")==0)) {  
        if((i+1)>=argc) exit(-1);
        sscanf(argv[i+1],"%i",&par->max_rep_each);
        i++;
      } else if((strcmp(&argv[i][1],"mr")==0) || (strcmp(&argv[i][1],"max_rms")==0)) {  
        if((i+1)>=argc) exit(-1);
        sscanf(argv[i+1],"%f",&par->max_rms);
        i++;
      } else if((strcmp(&argv[i][1],"me")==0) || (strcmp(&argv[i][1],"mE")==0) || (strcmp(&argv[i][1],"max_E")==0)) {  
        if((i+1)>=argc) exit(-1);
        sscanf(argv[i+1],"%f",&par->max_P);
        i++;
      } else if((strcmp(&argv[i][1],"mes")==0) || (strcmp(&argv[i][1],"mEs")==0) || (strcmp(&argv[i][1],"max_Es")==0)) {  
        if((i+1)>=argc) exit(-1);
        sscanf(argv[i+1],"%f",&par->max_Ps);
        i++;
      } else if(strcmp(&argv[i][1],"pf")==0){
        if((i+1)>=argc) exit(-1);
        sscanf(argv[i+1],"%f",&par->pf);
        i++;
      } else if(strcmp(&argv[i][1],"pe")==0){
        if((i+1)>=argc) exit(-1);
        sscanf(argv[i+1],"%f",&par->pe);
        i++;
      } else if(strcmp(&argv[i][1],"all")==0) {
        if(((i+1)>=argc) || (argv[i+1][0] == '-')){
	  par->match_rel = 1;
	}
	else {
	  sscanf(argv[i+1],"%f",&(par->match_rel));
	  if((par->match_rel <= 0) || (par->match_rel >=1)){
	    par->match_rel = 1;
	  }
	  i++;
	}
      } else if((strcmp(&argv[i][1],"aaa")==0) || (strcmp(&argv[i][1], "AAA")==0)) {
	par->AAA = '\1';
      } else if(strcmp(&argv[i][1],"v")==0 || strcmp(&argv[i][1],"V")==0) {
        par->verbose = 1;  
      } else if(strcmp(&argv[i][1],"vp")==0 || strcmp(&argv[i][1],"permut")==0) {
	par->view_permut = '\1';
      } else if(strcmp(&argv[i][1],"vs")==0 || strcmp(&argv[i][1],"subs")==0) {
	par->view_subs = '\1';
      } else if(strcmp(&argv[i][1],"lin")==0 || strcmp(&argv[i][1],"linear")==0) {
	par->linear = '\1';
      } else if(strcmp(&argv[i][1],"o")==0 || strcmp(&argv[i][1],"O")==0) {
        if((i+1)>=argc) exit(-1);
	if((par->out = fopen(argv[i+1], "a")) == NULL){
	  fprintf(stderr, "Error: file %s for redirected output not opened\n", argv[i+1]);
	  exit(-1);
	}
        i++;
      } else if(strcmp(&argv[i][1],"ex")==0 || strcmp(&argv[i][1],"excl")==0) {
        if((i+1)>=argc) exit(-1);
	if((EX = fopen(argv[i+1], "r")) == NULL){
	  fprintf(stderr, "Error: file %s for domain exclusion not found / opened\n", argv[i+1]);
	  exit(-1);
	}
	if((excl = (char **)malloc(sizeof(char*))) == NULL){
	  fprintf(stderr, "Error: Memory overflow\n");
	  exit;
	}
	nexcl = get_excl(EX, nexcl, &excl);
	printf("\nExclude %i domains from search\n", nexcl);
        i++;
      } else if(strcmp(&argv[i][1],"of")==0 || strcmp(&argv[i][1],"out_format")==0) {
        if((i+1)>=argc) exit(-1);
	if((strcmp(&argv[i+1][0],"pdb")==0) || (strcmp(&argv[i+1][0],"struct")==0) || (strcmp(&argv[i+1][0],"2")==0)){
	  par->out_format = 2;
	} else if((strcmp(&argv[i+1][0],"stamp")==0) || (strcmp(&argv[i+1][0],"trafo")==0) || (strcmp(&argv[i+1][0],"1")==0)){
	  par->out_format = 1;
	} else if((strcmp(&argv[i+1][0],"line")==0) || (strcmp(&argv[i+1][0],"l")==0) || (strcmp(&argv[i+1][0],"3")==0)){
	  par->out_format = 3;
	}
	i++;
      } else if(strcmp(&argv[i][1],"ha")==0 || strcmp(&argv[i][1],"heteroatoms")==0) {
	par->heteroatoms = '\1';
      } else if((strcmp(&argv[i][1],"gfp")==0) || (strcmp(&argv[i][1],"group_percent")==0)) {  
        if((i+1)>=argc) exit(-1);
        sscanf(argv[i+1],"%f",&par->min_group_sim);
	if(par->min_group_sim > 0){
	  par->group_filter = 1;
	  par->min_group_sim = par->min_group_sim / 100;
	}else{
	  par->group_filter = 0;
	}
        i++;
      } else if((strcmp(&argv[i][1],"gfa")==0) || (strcmp(&argv[i][1],"group_absolut")==0)) {  
        if((i+1)>=argc) exit(-1);
        sscanf(argv[i+1],"%f",&par->min_group_sim);
	if(par->min_group_sim > 0){
	  par->group_filter = 2;
	}else{
	  par->group_filter = 0;
	}
        i++;
      } else if(strcmp(&argv[i][1],"a")==0 || strcmp(&argv[i][1],"r")==0) {
	/*the modi */
      } else {
	help(1);
	exit(-1);
      }
   } /* end of for(i...*/



  /******************** Loading Definition file ********************/

   /* number of domains ok, now reading in definitions */
   if(((table = (struct table_struct *)malloc(sizeof(struct table_struct))) == NULL)
      || ((pseudo = (struct pseudo_struct *)malloc(sizeof(struct pseudo_struct))) == NULL)
      || ((permut = (struct permut_struct *)malloc(sizeof(struct permut_struct))) == NULL)
      || ((abund = (float *)malloc(sizeof(float))) == NULL)){
     exit(-1);
   }

   /*int gettable(struct permut_struct **permut, struct table_struct **table, struct pseudo_struct **pseudo, char *file, int *ntable_def, int *npseudo_def, char *def_files, float **abund){*/
  if((ntypes = gettable(&permut, &table, &pseudo, par->def_file, &ntables, &npseudos, def_files, &abund)) == -1){
    fprintf(stderr, "Error in gettable()\n");
    exit(-1);
  }
  printf("\nRead %i atom and %i pseudo-atom definition(s) corresponding to %i different types from %s\n\n", ntables, npseudos, ntypes, par->def_file);


  /******************** Loading Domains, Atoms; Defining Points ********************/

  /* alloc. memory for first domain, atoms[ndomains][], natoms[ndomains], points[ndomains][], npoints[ndomains] */
  if(((domains = (struct domain_loc*)malloc(sizeof(struct domain_loc))) == NULL)
     || ((atoms = (struct atom**)malloc(sizeof(struct atom *))) == NULL)
     || ((natoms = (int *)malloc(sizeof(int))) == NULL)
     || ((points = (struct point **)malloc(sizeof(struct point *))) == NULL)
     || ((npoints = (int *)malloc(sizeof(int))) == NULL)){
    exit(-1);
  }

  /* initialize ndomains, max_points and pointer to biggest dom. (big_dom) */
  ndomains = 0;
  max_points = 0;
  big_dom = 0;

  for(i=0; i<nfiles; i++){
    switch(filetype[i]){
       case 0 :
	 /********** DOMAIN file **********/

	 if((DOM=fopen(filename[i],"r"))==NULL){
	   fprintf(stderr, "Error: domain-file %s not found / opened\n", filename[i]);
	   exit(-1);
	 }
	 /* Getting number of domain-definitions in that file */
	 ndom_file = RBR_count_domain(DOM);

	 if(ndom_file == 0){
	   /* no domains in file --> probably wrong format */
	   fprintf(stderr, "No domains found in %s\n", filename[i]);
	   exit(-1);
	 }

	 /* realloc. memory for the new domains */
	 if(((domains = (struct domain_loc*)realloc(domains, (ndomains+ndom_file+1)*sizeof(struct domain_loc))) == NULL)
	    || ((atoms = (struct atom**)realloc(atoms, (ndomains+ndom_file+1)*sizeof(struct atom *))) == NULL)
	    || ((natoms = (int *)realloc(natoms, (ndomains+ndom_file+1)*sizeof(int))) == NULL)
	    || ((points = (struct point **)realloc(points, (ndomains+ndom_file+1)*sizeof(struct point *))) == NULL)
	    || ((npoints = (int *)realloc(npoints, (ndomains+ndom_file+1)*sizeof(int))) == NULL)){
	   exit(-1);
	 }

	 /* now getting max. ndom_file domains from that file */
	 rewind(DOM);
	 if(RBR_getdomain(DOM,&(domains[ndomains]),&ndom_file,ndom_file,&gottrans,env,0,stdout)==-1) exit(-1);
	 fclose(DOM);

	 printf("\nRead %d domain(s) from file %s\n",ndom_file, filename[i]);

	 /**** now getting coordinates,accessibility and points for all ndom_file domains ****/
	 for(j=ndomains; j<(ndomains+ndom_file); j++){

	   /* alloc. memory for the first atom and point entry */
	   if(((atoms[j]=(struct atom*)malloc(sizeof(struct atom))) == NULL)
	      || ((points[j]=(struct point *)malloc(sizeof(struct point))) == NULL)){
	     exit(-1);
	   }

	   /* read coord. */
	   printf("\nReading coordinates for domain %s\n",domains[j].id);
	   if(par->mode == 0){
	     if(RBR_getatoms(&(domains[j]), &(atoms[j]), &(natoms[j]), pdb_files, 0, 1, par)==-1){
	       fprintf(stderr, "Error in getallcoords() for domain %s\n",domains[j].id);
	       exit(-1);
	     }
	     printf("Read %i atoms\n", natoms[j]);
	   }else{
	     if(RBR_getatoms(&(domains[j]), &(atoms[j]), &(natoms[j]), asa_files, 0, 2, par)==-1){
	       fprintf(stderr, "Error in getallcoords() for domain %s\n",domains[j].id);
	       exit(-1);
	     }
	     printf("Read %i atoms and atom-accessibility values\n", natoms[j]);
	   }

	   /* read accessibility if necessary */
	   if(((par->abs_surf != -1) || (par->rel_surf != -1)) && (par->mode == 0)){
	     printf("\nReading accessibility (dssp) for domain %s\n",domains[j].id);
	     temp = RBR_getfile(domains[j].id, dssp_files, 4, stdout);
	     if((naccess=read_dssp(atoms[j], natoms[j], temp)) == -1){
	       fprintf(stderr, "Error in read_dssp() for %s\n", temp);
	       exit(-1);
	     }
	     printf("Read accessibility for %i residues from file %s\n", naccess, temp);
	   }
	  
	   /* now get points (first point is alloc. above */
	   if((npoints[j]=get_points(&(points[j]), atoms[j], table, pseudo, natoms[j], ntables, npseudos, par)) < 0){
	     fprintf(stderr, "Error in get_points for %s\n", domains[j].id);
	     exit(-1);
	   }
	   if(npoints[j] > max_points){
	     max_points = npoints[j];
	     big_dom = j;
	   }
	   printf("Created %i datapoints\n", npoints[j]);
	 }

	 ndomains += ndom_file;
	 break;

       case 1 :
	 /********** COORD. file -> corresponds to one domain **********/

	 /* realloc. memory for the new domain */
	 if(((domains = (struct domain_loc*)realloc(domains, (ndomains+1)*sizeof(struct domain_loc))) == NULL)
	    || ((atoms = (struct atom**)realloc(atoms, (ndomains+1)*sizeof(struct atom *))) == NULL)
	    || ((natoms = (int *)realloc(natoms, (ndomains+1)*sizeof(int))) == NULL)
	    || ((points = (struct point **)realloc(points, (ndomains+1)*sizeof(struct point *))) == NULL)
	    || ((npoints = (int *)realloc(npoints, (ndomains+1)*sizeof(int))) == NULL)){
	   exit(-1);
	 }

	 /* creating domain entry */
	 strcpy(domains[ndomains].filename, filename[i]);
	 h_count = 0;
	 while((h_count < strlen(filename[i])) && (filename[i][h_count] != '.')) h_count++;
	 strncpy(domains[ndomains].id, filename[i],h_count);
	 domains[ndomains].id[h_count] = '\0';
	 domains[ndomains].nobj = 1;
	 if((domains[ndomains].type=(int*)malloc(sizeof(int))) == NULL){
	   exit(-1);
	 }
	 domains[ndomains].type[0] = 1;
	 printf("Defined coordinate-file %s as domain %s\n", filename[i], domains[ndomains].id);

	 /* alloc. memory for the first atom/point entry */
	 if(((atoms[ndomains]=(struct atom*)malloc(sizeof(struct atom))) == NULL)
	    || ((points[ndomains]=(struct point *)malloc(sizeof(struct point))) == NULL)){
	   exit(-1);
	 }

	 /* read coord. */
	 printf("\nReading coordinates for domain %s\n",domains[ndomains].id);
	 if(par->mode == 0){
	   if(RBR_getatoms(&(domains[ndomains]), &(atoms[ndomains]), &(natoms[ndomains]), pdb_files, 0, 1, par)==-1){
	     fprintf(stderr, "Error in getallcoords() for domain %s\n",domains[ndomains].id);
	     exit(-1);
	   }
	   printf("Read %i atoms\n", natoms[ndomains]);
	 }else{
	   if(RBR_getatoms(&(domains[ndomains]), &(atoms[ndomains]), &(natoms[ndomains]), asa_files, 0, 2, par)==-1){
	     fprintf(stderr, "Error in getallcoords() for domain %s\n",domains[ndomains].id);
	     exit(-1);
	   }
	   printf("Read %i atoms and atom-accessibility values\n", natoms[ndomains]);
	 }

	 /* read accessibility if necessary */
	 if(((par->abs_surf != -1) || (par->rel_surf != -1)) && (par->mode == 0)){
	   printf("\nReading accessibility (dssp) for domain %s\n",domains[ndomains].id);
	   temp = RBR_getfile(domains[ndomains].id, dssp_files, 4, stdout);
	   if((naccess=read_dssp(atoms[ndomains], natoms[ndomains], temp)) == -1){
	     fprintf(stderr, "Error in read_dssp() for %s\n", temp);
	     exit(-1);
	   }
	   printf("Read accessibility for %i residues from file %s\n", naccess, temp);
	 }
	  
	 /* get points, mem. for first is alloc. above */
	 if((npoints[ndomains]=get_points(&(points[ndomains]), atoms[ndomains], table, pseudo, natoms[ndomains], ntables, npseudos, par)) < 0){
	   fprintf(stderr, "Error in get_points for %s\n", domains[ndomains].id);
	   exit(-1);
	 }
	 if(npoints[ndomains] > max_points){
	   max_points = npoints[ndomains];
	   big_dom = ndomains;
	 }
	 printf("Created %i datapoints\n", npoints[ndomains]);
 
	 ndomains++;
	 break;

       case 2 :
	 /********** Database file **********/
	 if((DOM=fopen(filename[i],"r"))==NULL){
	   temp = RBR_getfile(domains[ndomains].id, database_files, 4, stdout);
	   if((DOM=fopen(temp, "r"))==NULL){
	     fprintf(stderr, "Error: domain-file %s not found / opened\n", filename[i]);
	     exit(-1);
	   }
	 }

	 ndom_file = 0;

	 get_database(&domains, &atoms, &natoms, ndomains, &ndom_file, DOM, par);

	 if(ndom_file == 0){
	   /* no domain read --> wrong file format */
	   fprintf(stderr, "Wrong format for database file %s\n", filename[i]);
	   exit(-1);
	 }

	 printf("\nRead %i domains from database file %s\n", ndom_file, filename[i]);

	 /* realloc. memory for points and npoints (atoms/domains was done in get_database) */
	 if(((points = (struct point **)realloc(points, (ndomains+ndom_file+1)*sizeof(struct point *))) == NULL)
	    || ((npoints = (int *)realloc(npoints, (ndomains+ndom_file+1)*sizeof(int))) == NULL)){
	   exit(-1);
	 }
	 for(j=ndomains; j<(ndomains+ndom_file); j++){
	   /* alloc. memory for first point */
	   if((points[j]=(struct point *)malloc(sizeof(struct point))) == NULL){
	     exit(-1);
	   }
	   if((npoints[j]=get_points(&(points[j]), atoms[j], table, pseudo, natoms[j], ntables, npseudos, par)) < 0){
	     fprintf(stderr, "Error in get_points for %s\n", domains[j].id);
	     exit(-1);
	   }
	   if(npoints[j] > max_points){
	     max_points = npoints[j];
	     big_dom = j;
	   }
	 }
	 ndomains += ndom_file;
	 break;
    }
  }


  /******************** Alloc. memory for dist[0/1][][], radius[][], match[][] ********************
   * curr_xxx hold the current pair with [0] = query = domains[i]
   * all of them are allocated for the largest domain with max_points points
   * this prevents freeing and reallocating memory for each comp.

   * found motifs are stored in motif. motif is always the current motif that is still
   * expanded. upon finishing, it is stored in ***store
   * j_array holds the ID-numbers of atom set 2. This allows swapping of "atoms" while still counting from j=0 to j<npoints[1]
     in search(). j_array is initialized as a 1:1 pointer redirection (j_array[n] = n)
     * motifs and j_array are also moved to main() to prevent freeing and realloc. */

  if(((dist = (long int ***)malloc(2*sizeof(long int **))) == NULL)
     || ((dist[0] = (long int **)malloc(max_points*sizeof(long int *))) == NULL)
     || ((dist[1] = (long int **)malloc(max_points*sizeof(long int *))) == NULL)
     || ((radius = (char ***)malloc(2*sizeof(char **))) == NULL)
     || ((radius[0] = (char **)malloc(max_points*sizeof(char *))) == NULL)
     || ((radius[1] = (char **)malloc(max_points*sizeof(char *))) == NULL)
     || ((match = (char **)malloc(max_points*sizeof(char *))) == NULL)
     || ((store = (struct store_struc *)malloc(sizeof(struct store_struc))) == NULL)
     || ((final_res = (struct store_struc *)malloc(sizeof(struct store_struc))) == NULL)
     || ((curr_domains = (struct domain_loc *)malloc(2*sizeof(struct domain_loc))) == NULL)
     || ((curr_natoms = (int *)malloc(2*sizeof(int))) == NULL)
     || ((curr_atoms = (struct atom **)malloc(2*sizeof(struct atom *))) == NULL)
     || ((curr_npoints = (int *)malloc(2*sizeof(int))) == NULL)
     || ((curr_points = (struct point **)malloc(2*sizeof(struct points *))) == NULL)
     || ((motif = (int **)malloc(2*sizeof(int *))) == NULL)
     || ((motif[0] = (int *)malloc(max_points*sizeof(int))) == NULL)
     || ((motif[1] = (int *)malloc(max_points*sizeof(int))) == NULL)
     || ((j_array = (int *)malloc(max_points*sizeof(int))) == NULL)){
    exit(-1);
  }

  i = max_points;
  do {
    i--;
    if(((dist[0][i] = (long int *)malloc(max_points*sizeof(long int))) == NULL)
       || ((dist[1][i] = (long int *)malloc(max_points*sizeof(long int))) == NULL)
       || ((radius[0][i] = (char *)malloc(max_points*sizeof(char))) == NULL)
       || ((radius[1][i] = (char *)malloc(max_points*sizeof(char))) == NULL)
       || ((match[i] = (char *)malloc(max_points*sizeof(char))) == NULL)){
      exit(-1);
    }
  } while(i > 0);


  /******************** Start outer loop: Prepare QUERY ********************/

  /* initialize number of found/kept matches to 0
   * -> in subsequent searches they will be come from the previous search */

  if(par->match_rel <=0){
    par->min_atoms = par->match_abs;
  }

  matches_found = 0;
  matches_kept = 0;

  i = 0;
  do {

    /* domains with 0/1 point are skipped */    
    if(npoints[i] < 2){
      continue;
    }

    curr_domains[0] = domains[i];
    curr_natoms[0] = natoms[i];
    curr_atoms[0] = atoms[i];
    curr_npoints[0] = npoints[i];
    curr_points[0] = points[i];

    prepare_domain(curr_points[0], curr_npoints[0], dist[0], radius[0], par->max_size);
    
    nmatches = 0; /* initialize number of matches for inner loop */
    /******************** Start inner loop: Prepare DB-entry ********************/
    j=i+1;
    do {

      /* domains with 0/1 point are skipped - as matches_found is not changed, just continue */
      if(npoints[j] < 2){
	continue;
      }

      /*domains to be excluded (ids in excl[][] are skipped*/
      for(k=0; k<nexcl; k++){
	if(strcmp(domains[j].id, excl[k]) == 0){
	  /*this should be excluded*/
	  k=-1;
	  break;
	}
      }
      if(k == -1){
	continue;
      }

      /* frees memory from previous searches and sets store to 1 again */
      for(k=0; k < matches_found; k++){
	free(store[k].p1);
	free(store[k].p2);
      }
      if((store = (struct store_struc *)realloc(store, sizeof(struct store_struc))) == NULL){
	exit(-1);
      }

      matches_found = 0;
      matches_kept = 0;

      curr_domains[1] = domains[j];
      curr_natoms[1] = natoms[j];
      curr_atoms[1] = atoms[j];
      curr_npoints[1] = npoints[j];
      curr_points[1] = points[j];

      prepare_domain(curr_points[1], curr_npoints[1], dist[1], radius[1], par->max_size);

      /* For the search, the smaller set must be 0 */
      swapped=0;

      if(npoints[0]>npoints[1]){

	domain_swap = curr_domains[0];
	curr_domains[0] = curr_domains[1];
	curr_domains[1] = domain_swap;

	swap_help = curr_natoms[0];
	curr_natoms[0] = curr_natoms[1];
	curr_natoms[1] = swap_help;

	array_swap = curr_atoms[0];
	curr_atoms[0] = curr_atoms[1];
	curr_atoms[1] = array_swap;

	swap_help = curr_npoints[0];
	curr_npoints[0] = curr_npoints[1];
	curr_npoints[1] = swap_help;

	array_swap = curr_points[0];
	curr_points[0] = curr_points[1];
	curr_points[1] = array_swap;

	array_swap = dist[0];
	dist[0] = dist[1];
	dist[1] = array_swap;

	array_swap = radius[0];
	radius[0] = radius[1];
	radius[1] = array_swap;

	swapped=1;
  }
  /* now, the smaller set is always 0 */

      /*
  if(!(par->match_all)){
    for(i=0; i<npoints[0]; i++){
      do{
	j = (int)(npoints[0]*(((float)rand())/RAND_MAX));
      }while((j >= npoints[0]) || (j < 0));
      point_swap = points[0][i];
      points[0][i] = points[0][j];
      points[0][j] = point_swap;
    }
    for(k=0; k<npoints[0]; k++){
      do{
	j = (int)(npoints[0]*(((float)rand())/RAND_MAX));
      }while(j >= npoints[0]);
      do{
	i = (int)(npoints[0]*(((float)rand())/RAND_MAX));
      }while(i >= npoints[0]);
      point_swap = points[0][i];
      points[0][i] = points[0][j];
      points[0][j] = point_swap;
    }
  }
      */

      if(par->match_rel > 0){
	par->min_atoms = (int) (par->match_rel * (float)curr_npoints[0] + .5);
	if((par->min_atoms < par->match_abs) || (par->min_atoms > curr_npoints[0])){
	  par->min_atoms = curr_npoints[0];
	}
	par->max_size = 0;
      }

      if(par->min_atoms>par->max_depth){
	par->max_depth = par->min_atoms + 1;
      }

      if(par->verbose == 1){
	printf("\nStarting search %i : %s, (%i points) - %i : %s, (%i points)...\n", i, domains[i].id, npoints[i], j, domains[j].id, npoints[j]);
	printf("Parameters used :\n");
	printf("Distance tolerance : %f\n",(float)par->max_tolerance/PRECIS);
	if(par->max_size != 0){
	  printf("Maximum motif-size : %f\n", (float)par->max_size/PRECIS);
	}else{
	  printf("Maximum motif-size : off\n");
	}
	printf("Maximum search depth : %i\n", par->max_depth);
	printf("Maximum number of hits : %i\n", par->max_hits);
	printf("Minimum number of atoms for motif : %i\n", par->min_atoms);
	if(par->rel_surf == -1){
	  printf("Minimum relative surface accessibility : off\n");
	}else if(par->rel_surf == 0){
	  printf("Minimum relative surface accessibility taken from definition file\n");
	}else{
	  printf("Minimum relative surface accessibility : %f %\n", par->rel_surf);
	}
	if(par->abs_surf == -1){
	  printf("Minimum absolute surface accessibility : off\n");
	}else if(par->abs_surf == 0){
	  printf("Minimum absolute surface accessibility taken from definition file\n");
	}else{
	  printf("Minimum absolute surface accessibility : %f\n", par->abs_surf);
	}
	if(par->max_rms != -1){
	  printf("Maximum rms deviation : %f\n", par->max_rms);
	}else{
	  printf("Maximum rms deviation : off\n");
	}
	if(par->group_filter == 1){
	  printf("Combine motifs with >= %f % identity\n", par->min_group_sim * 100);
	}else if(par->group_filter == 2){
	  printf("Combine motifs with >= %i shared atoms\n", (int) par->min_group_sim);
	}else{
	  printf("Combine similar motifs : off\n");
	}
	printf("\n\n");
      }

      if(par->linear){
	matches_found = start_lin_search(curr_points, curr_npoints, dist, radius[0], match, motif, &store, par);
      }else{
	matches_found = start_search(curr_points, curr_npoints, dist, radius[0], match, motif, j_array, &store, par);
      }

      if(swapped){
	/* swap all arrays back */

	domain_swap = curr_domains[0];
	curr_domains[0] = curr_domains[1];
	curr_domains[1] = domain_swap;

	swap_help = curr_natoms[0];
	curr_natoms[0] = curr_natoms[1];
	curr_natoms[1] = swap_help;

	array_swap = curr_atoms[0];
	curr_atoms[0] = curr_atoms[1];
	curr_atoms[1] = array_swap;

	swap_help = curr_npoints[0];
	curr_npoints[0] = curr_npoints[1];
	curr_npoints[1] = swap_help;

	array_swap = curr_points[0];
	curr_points[0] = curr_points[1];
	curr_points[1] = array_swap;

	array_swap = dist[0];
	dist[0] = dist[1];
	dist[1] = array_swap;

	array_swap = radius[0];
	radius[0] = radius[1];
	radius[1] = array_swap;

	for(k=0; k<par->found; k++){
	  array_swap = store[k].p1;
	  store[k].p1 = store[k].p2;
	  store[k].p2 = array_swap;
	}
      }
      /* now all arrays are back */


      if(matches_found < 1){
	/* not matches found */
	continue;
      }

      if(par->verbose){
	printf("Found %i motifs\n", par->found);
      }

      /* RMSD calculation and filter */
      /*      calc_rms(store, curr_points, par);*/
      /*int calc_rms_perm(struct permut_struct *permute, struct store_struc *store, struct point *points[2], struct parameter *par)*/
      calc_rms_perm(permut, store, curr_points, par);
      if(par->max_rms != -1){
	if(rms_filter(store, par) < 1){
	  continue;
	}
	if(par->verbose){
	  printf("%i motifs after rms-filter\n", par->found);
	}
      }

      /* P-value calculation and filter */
      /* int calc_P(struct store_struc *store, float *abund, int npoints[2], struct point *points[2], struct parameter *par){ */
      calc_P(store, abund, curr_npoints, curr_points, par);
      if(par->max_P != -1){
	if(P_filter(store, par) < 1){
	  continue;
	}
	if(par->verbose){
	  printf("%i motifs after E-value filter\n", par->found);
	}
      }

      if(par->max_Ps != -1){
	if(Ps_filter(store, par) < 1){
	  continue;
	}
	if(par->verbose){
	  printf("%i motifs after stand. E700 filter\n", par->found);
	}
      }

      /* Submotifs (= motifs totally contained in other motifs are removed
       * this does not occur when match_all */
      if((par->view_subs) && (par->match_rel < 1)){
	if(rm_submotifs_smart(store, par) < 1){
	  continue;
	}
	if(par->verbose){
	  printf("%i motifs after submotifs are removed\n", par->found);
	}
      }

      /* if long range filter */
      if(par->long_range > -1){
	/*int LR_filter(struct store_struc *store, struct point **points, struct parameter *par){*/
	if(LR_filter(store, curr_points, par) < 1){
	  continue;
	}
      }

      /* Permutations (=matches with same AAs but different order) are removed */
      if(!(par->view_permut)){
	if(rm_permutations(store, par) < 1){
	  continue;
	}
	if(par->verbose){
	  printf("%i motifs after permutations are removed\n", par->found);
	}
      }

      /* Clusters matches into groups according to common AAs */
      /* Memory management not properly done !!! */
      if((par->group_filter != 0) && (par->found > 1)){
	if((ident = (float **)malloc(par->found*sizeof(float *))) == NULL){
	  exit(-1);
	}
	for(i=0; i<par->found; i++){
	  if((ident[i] =  (float *)malloc(par->found*sizeof(float))) == NULL){
	    exit(-1);
	  }
	}

	calc_ident(ident, store, par);

	groups = GJnoc(ident, par->found, par->cluster_type);

	red_motifs(groups, red_clust(groups, par->found, par->min_group_sim), store, par);

	if(par->verbose){
	  printf("%i motifs after group_filter\n", par->found);
	}
      }


      /* sort according to P-value */
      sort_P(store, par->found);

      /* now copying report=min(max_rep,par->found) filtered results to final_res */
      if((par->max_rep_each > 0) && (par->max_rep_each < par->found)){
	report = par->max_rep_each;
      }else{
	report = par->found;
      }

      if((final_res = (struct store_struc *)realloc(final_res, (nmatches+report)*sizeof(struct store_struc))) == NULL){
	exit(-1);
      }
      for(k=0; k<report; k++){
	if(((final_res[nmatches+k].p1 = (int *)malloc(store[k].n*sizeof(int))) == NULL)
	   || ((final_res[nmatches+k].p2 = (int *)malloc(store[k].n*sizeof(int))) == NULL)){
	  exit(-1);
	}
	l=0;
	do{
	  final_res[nmatches+k].p1[l] = store[k].p1[l];
	  final_res[nmatches+k].p2[l] = store[k].p2[l];
	  l++;
	}while(l<store[k].n);

	final_res[nmatches+k].domain = j; /* current domain during inner loop */
	final_res[nmatches+k].n = store[k].n;
	final_res[nmatches+k].P = store[k].P;
	final_res[nmatches+k].Ps = store[k].Ps;
	final_res[nmatches+k].rms = store[k].rms;

	l = 0;
	do {
	  m = 0;
	  do {
	    final_res[nmatches+k].R[l][m] = store[k].R[l][m];
	    m++;
	  } while(m<3);
	  final_res[nmatches+k].V[l] = store[k].V[l];
	  l++;
	}while(l<3);
      }

      nmatches += report;
     
      /*
      k = print_struct_old(store, curr_points, curr_natoms, curr_atoms, curr_domains, par);
      printf("\n\nFound and filtered %i motifs comparing (%s | %s)\n", par->found, curr_domains[0].id, curr_domains[1].id);
      printf("\nWrote %i motifs\n",k );
      */

    } while(++j<ndomains);

    /* not matches found in the whole loop */
    if(nmatches < 1){
      printf(">No matches found for %s\n", domains[i].id);
      continue;
    }

    printf(">%i matches found for %s\n", nmatches, domains[i].id);

    /* nmatches found: sort acc. to P-value and print */
    sort_P(final_res, nmatches);
    par->found = nmatches;

    switch(par->out_format){
    case 0:
      k = print_short(final_res, i, points, natoms, atoms, domains, par);
      break;
    case 1:
      k = print_domain(final_res, i, points, natoms, atoms, domains, par);
      break;
    case 2:
      k = print_pdb(final_res, i, points, natoms, atoms, domains, par);
      break;
    case 3:
      k = print_line(final_res, i, points, natoms, atoms, domains, par);
      break;
    }
    /*
    k = print_short(final_res, i, points, natoms, atoms, domains, par);
    */

    /* now freeing memory and reallocating to one */
    for(k=0; k < nmatches; k++){
      free(final_res[k].p1);
      free(final_res[k].p2);
    }
    if((final_res = (struct store_struc *)realloc(final_res, sizeof(struct store_struc))) == NULL){
      exit(-1);
    }

  } while(par->AAA && (++i<(ndomains-1)));

  printf("\nElapsed time: %u secs.\n", clock()/CLOCKS_PER_SEC);

  if(par->out != stdout){
    fclose(par->out);
  }
  exit(1);
}
